/*------------------------------------------------------------------------------
 * MDK Middleware - Component ::Network
 * Copyright (c) 2004-2016 ARM Germany GmbH. All rights reserved.
 *------------------------------------------------------------------------------
 * Name:    SMTP_Client.c
 * Purpose: SMTP Client example
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include "cmsis_os.h"                   /* CMSIS RTOS definitions             */
#include "rl_net.h"                     /* Network definitions                */
#include "sam.h"                        // Device header
#include "Board_LED.h"                  // ::Board Support:LED
//-------- <<< Use Configuration Wizard in Context Menu >>> -----------------

//   <s.40>Remote IP Address
//   =======================
//   <i> IPv4 or IPv6 Address in text representation
//   <i>Default: "192.168.0.100" or "fe80::1c30:6cff:fea2:455e"
#define IP_ADDR        "fe80::1c30:6cff:fea2:455e"

//------------- <<< end of configuration section >>> -----------------------

extern void _SetupMemoryRegion( void );

netSMTPc_Event smtp_event;
osThreadId send_id;

static void BlinkLed (void const *arg);
static void SendMail (void const *arg);

osThreadDef(BlinkLed, osPriorityNormal, 1, 0);
osThreadDef(SendMail, osPriorityNormal, 1, 0);
                         
/// Inform application about SMTP status
void smtp_inform (netSMTPc_Event event) {
  smtp_event = event;
  // Notify send thread of SMTP status
  osSignalSet (send_id, 0x0001);
}

/*----------------------------------------------------------------------------
  Thread 'SendMail': Send an email
 *---------------------------------------------------------------------------*/
static void SendMail (void const *arg) {
  NET_ADDR smtp_srv;
  netStatus retv;

  send_id = osThreadGetId ();

  /* Setup server parameters */
  if (netIP_aton (IP_ADDR, NET_ADDR_IP4, smtp_srv.addr)) {
    /* IP_ADDR specifies IPv4 address */
    smtp_srv.addr_type = NET_ADDR_IP4;
  }
  else if (netIP_aton (IP_ADDR, NET_ADDR_IP6, smtp_srv.addr)) {
    /* IP_ADDR specified IPv6 address */
    smtp_srv.addr_type = NET_ADDR_IP6;
  }
  smtp_srv.port = 0;
  while(1) {
    // After 5 seconds start sending email
    osDelay (5000);
    printf("Sending email...\n");

    // Start SMTP client, use default SMTP port
    retv = netSMTPc_Connect (&smtp_srv);
    if (retv != netOK) {
      // Failed to start SMTP process
      printf("Connect failed\n");
      continue;
    }
    osSignalWait (0x0001, osWaitForever);
    // Check the completion code
    switch (smtp_event) {
      case netSMTPc_EventSuccess:
        // Success, terminate thread
        printf("Success, email sent\n");
        osThreadTerminate (send_id);
        break;
      case netSMTPc_EventTimeout:
        // Timeout, try again
        printf("Server timeout\n");
      break;
      case netSMTPc_EventError:
        // Error, try again
        printf("Operation failed\n");
        break;
    }
  }
}

/*----------------------------------------------------------------------------
  Thread 'BlinkLed': Blink the LEDs on an eval board
 *---------------------------------------------------------------------------*/
static void BlinkLed (void const *arg) {
  const uint8_t led_val[] = { 0x01, 0x02 };
  int cnt = 0;

  while(1) {
    // Every 200 ms
    LED_SetOut (led_val[cnt]);
    if (++cnt >= sizeof(led_val)) {
      cnt = 0;
    }
    osDelay (200);
  }
}


/*----------------------------------------------------------------------------
  Main Thread 'main': Run Network
 *---------------------------------------------------------------------------*/
int main (void) {
  /* Disable watchdog */
  WDT->WDT_MR = WDT_MR_WDDIS; 
  
  /* Setup MPU */
  _SetupMemoryRegion();
  
  /* Update system clock */
  SystemCoreClockUpdate();
  
  /* Enable cache */
  SCB_EnableDCache(); 
  SCB_EnableICache();   
    
  /* Leds initialization */
  LED_Initialize();
 
  /* Network initialization */
  netInitialize ();

  /* Create Threads */  
  osThreadCreate (osThread(BlinkLed), NULL);
  osThreadCreate (osThread(SendMail), NULL);

  while(1) {
    osSignalWait (0, osWaitForever);
  }
}
