/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2018 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        28. March 2018
 * $Revision:    V1.11
 *
 * Driver:       Driver_CAN0/1
 * Configured:   via RTE_Device.h configuration file
 * Project:      CMSIS CAN Driver for Atmel SAM-ESV7 (with CAN FD support)
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                 Value   CAN Interface
 *   ---------------------                 -----   -------------
 *   Connect to hardware via Driver_CAN# = 0       use CAN0
 *   Connect to hardware via Driver_CAN# = 1       use CAN1
 * --------------------------------------------------------------------------
 * Defines used for driver configuration (at compile time):
 *
 *   CAN_CLOCK_TOLERANCE:       defines maximum allowed clock tolerance in 1/1024 steps
 *     - default value:         15  (approx. 1.5 %)
 *   CAN_NON_ISO_EN:            used to enable non-ISO CAN FD mode
 *     - default value          0 (disabled)
 *
 *   CAN0 controller configuration:
 *   CAN0_RX_FIFO0_ELEM_NUM:    number of receive FIFO0 elements    (default=64,  max=64)
 *   CAN0_RX_FIFO1_ELEM_NUM:    number of receive FIFO1 elements    (default=64,  max=64)
 *   CAN0_TX_FIFO_ELEM_NUM:     number of transmit FIFO elements    (default=8,   max=32)
 *   CAN0_RX_BUF_NUM:           maximum number of receive buffers   (default=64,  max=64)
 *   CAN0_TX_BUF_NUM:           maximum number of transmit buffers  (default=24,  max=32)
 *   CAN0_FILTER_STD_NUM:       maximum number of standard elements (default=128, max=128)
 *   CAN0_FILTER_EXT_NUM:       maximum number of extended elements (default=64,  max=64)
 *   CAN0_TRANSMITTER_DELAY_COMPENSATION_EN: transmitter delay compensation enable
 *                             (0 = disabled
 *                              1 = enabled, TDCV value is measured by HW)
 *   CAN0_TRANSMITTER_DELAY_OFFSET: offset from measured delay to secondary sample point of a bit
 *                             (specified in CAN clock periods
 *   CAN0_TRANSMITTER_DELAY_FILTER: minimum value for the secondary sample point
 *                             (specified in CAN clock periods, and available from chip Revision B)
 *
 *   CAN1 controller configuration:
 *   CAN1_RX_FIFO0_ELEM_NUM:    number of receive FIFO0 elements    (default=64,  max=64)
 *   CAN1_RX_FIFO1_ELEM_NUM:    number of receive FIFO1 elements    (default=64,  max=64)
 *   CAN1_TX_FIFO_ELEM_NUM:     number of transmit FIFO elements    (default=8,   max=32)
 *   CAN1_RX_BUF_NUM:           maximum number of receive buffers   (default=64,  max=64)
 *   CAN1_TX_BUF_NUM:           maximum number of transmit buffers  (default=24,  max=32)
 *   CAN1_FILTER_STD_NUM:       maximum number of standard elements (default=128, max=128)
 *   CAN1_FILTER_EXT_NUM:       maximum number of extended elements (default=64,  max=64)
 *   CAN1_TRANSMITTER_DELAY_COMPENSATION_EN: transmitter delay compensation enable
 *                             (0 = disabled
 *                              1 = enabled, TDCV value is measured by HW)
 *   CAN1_TRANSMITTER_DELAY_OFFSET: offset from measured delay to secondary sample point of a bit
 *                             (specified in CAN clock periods
 *   CAN1_TRANSMITTER_DELAY_FILTER: minimum value for the secondary sample point
 *                             (specified in CAN clock periods, and available from chip Revision B)
 *
 * Notes:
 *   - Tx FIFO and Tx Buffers transmit concepts are used (Tx Queue is not used)
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.10
 *    Corrected Control function if ARM_CAN_SET_FD_MODE requested while 
 *    CAN controller is in INIT state
 *  Version 1.10
 *    Corrected abort message send functionality
 *  Version 1.9
 *    Corrected SetBitrate function
 *  Version 1.8
 *    Corrected Transceiver Delay Compensation functionality
 *  Version 1.7
 *    Updated for Q21B chip revision using different header
 *  Version 1.6
 *    Added support for chip revision D/E
 *    Added capability to use non-ISO CAN FD mode (only chip revisions D/E/...)
 *    CAN clock changed to use UPLL as suggested in documentation (UPLL/6 = 80 MHz)
 *    Corrected ObjectSetFilter function for filter removal
 *  Version 1.5
 *    Corrected standard ID handling (now supports 11 bits of ID, was 10)
 *  Version 1.4
 *    Corrected CAN0 configuration structure
 *  Version 1.3
 *    Added Rx/Tx FIFO support (enable mask and range filtering)
 *  Version 1.2
 *    Corrected transmit and receive for relative object index different than 0
 *  Version 1.1
 *    Corrected functionality when NULL pointer is provided for one or both 
 *    signal callbacks in Initialize function call
 *  Version 1.0
 *    Initial release
 */

#include "CAN_SAM-ESV7.h"

// Externally overridable configuration definitions

// Maximum allowed clock tolerance in 1/1024 steps
#ifndef CAN_CLOCK_TOLERANCE
#define CAN_CLOCK_TOLERANCE             (15U)   // 15/1024 approx. 1.5 %
#endif

// Use non-ISO CAN FD
#ifndef CAN_NON_ISO_EN
#define CAN_NON_ISO_EN                  (0U)    // non-ISO disabled by default
#endif


// CAN0 controller configuration ------

#ifndef CAN0_RX_FIFO0_ELEM_NUM
#define CAN0_RX_FIFO0_ELEM_NUM          (64U)
#endif
#if    (CAN0_RX_FIFO0_ELEM_NUM > 64U)
#error  Too many Receive FIFO0 Elements defined for CAN0, maximum number of Receive FIFO0 Elements is 64 !!!
#endif

#ifndef CAN0_RX_FIFO1_ELEM_NUM
#define CAN0_RX_FIFO1_ELEM_NUM          (64U)
#endif
#if    (CAN0_RX_FIFO1_ELEM_NUM > 64U)
#error  Too many Receive FIFO1 Elements defined for CAN0, maximum number of Receive FIFO1 Elements is 64 !!!
#endif

#ifndef CAN0_RX_BUF_NUM
#define CAN0_RX_BUF_NUM                 (64U)
#endif
#if    (CAN0_RX_BUF_NUM > 64U)
#error  Too many Receive Buffers defined for CAN0, maximum number of Receive Buffers is 64 !!!
#endif

#ifndef CAN0_TX_FIFO_ELEM_NUM
#define CAN0_TX_FIFO_ELEM_NUM           (8U)
#endif
#if    (CAN0_TX_FIFO_ELEM_NUM > 32U)
#error  Too many Transmit FIFO Elements defined for CAN0, maximum number of Transmit FIFO Elements is 32 !!!
#endif

#ifndef CAN0_TX_BUF_NUM
#define CAN0_TX_BUF_NUM                 (24U)
#endif
#if    (CAN0_TX_BUF_NUM > 32U)
#error  Too many Transmit Buffers defined for CAN0, maximum number of Transmit Buffers is 32 !!!
#endif

#if    ((CAN0_TX_FIFO_ELEM_NUM + CAN0_TX_BUF_NUM) > 32U)
#error  Too many Transmit Elements and Buffers defined for CAN0, maximum number of Transmit Elements and Buffers together is 32 !!!
#endif

#ifndef CAN0_FILTER_STD_NUM
#define CAN0_FILTER_STD_NUM             (128U)
#endif
#if    (CAN0_FILTER_STD_NUM > 128U)
#error  Standard Filter Elements defined for CAN0, maximum number of Standard Filter Elements is 128 !!!
#endif

#ifndef CAN0_FILTER_EXT_NUM
#define CAN0_FILTER_EXT_NUM             (64U)
#endif
#if    (CAN0_FILTER_EXT_NUM > 64U)
#error  Extended Filter Elements defined for CAN0, maximum number of Extended Filter Elements is 64 !!!
#endif

#ifndef CAN0_TRANSMITTER_DELAY_COMPENSATION_EN
#define CAN0_TRANSMITTER_DELAY_COMPENSATION_EN  (0U)
#endif

#ifndef CAN0_TRANSMITTER_DELAY_OFFSET
#define CAN0_TRANSMITTER_DELAY_OFFSET           (0U)
#endif

#ifndef CAN0_TRANSMITTER_DELAY_FILTER
#define CAN0_TRANSMITTER_DELAY_FILTER           (0U)
#endif
#if   ((CAN0_TRANSMITTER_DELAY_FILTER != 0U) &&  (CHIP_REV_A == 1))
#warning  Transmitter Delay Compensation Filter setting is only available from chip Revision B !!!
#endif

#define CAN0_RX_FIFO_EN                ((CAN0_RX_FIFO0_ELEM_NUM != 0U) || (CAN0_RX_FIFO1_ELEM_NUM != 0U))
#define CAN0_RX_BUF_EN                  (CAN0_RX_BUF_NUM        != 0U)
#define CAN0_TX_FIFO_EN                 (CAN0_TX_FIFO_ELEM_NUM  != 0U)
#define CAN0_TX_BUF_EN                  (CAN0_TX_BUF_NUM        != 0U)
#define CAN0_RX_FIFO_OBJ_NUM           ((CAN0_RX_FIFO0_ELEM_NUM != 0U) + (CAN0_RX_FIFO1_ELEM_NUM != 0U))
#define CAN0_RX_BUF_OBJ_NUM             (CAN0_RX_BUF_NUM)
#define CAN0_RX_TOT_OBJ_NUM             (CAN0_RX_FIFO_OBJ_NUM + CAN0_RX_BUF_OBJ_NUM)
#define CAN0_TX_FIFO_OBJ_NUM           ((CAN0_TX_FIFO_ELEM_NUM  != 0U))
#define CAN0_TX_BUF_OBJ_NUM             (CAN0_TX_BUF_NUM)
#define CAN0_TX_TOT_OBJ_NUM             (CAN0_TX_FIFO_OBJ_NUM + CAN0_TX_BUF_OBJ_NUM)
#define CAN0_TOT_OBJ_NUM                (CAN0_RX_TOT_OBJ_NUM  + CAN0_TX_TOT_OBJ_NUM)
#define CAN0_TOT_ELEM_NUM               (CAN0_RX_FIFO0_ELEM_NUM + CAN0_RX_FIFO1_ELEM_NUM + CAN0_RX_BUF_NUM + CAN0_TX_BUF_NUM)

#define CAN0_RAM_OFS_STD_ID             (0U)
#define CAN0_RAM_OFS_EXT_ID             (CAN0_RAM_OFS_STD_ID        +   CAN0_FILTER_STD_NUM)
#define CAN0_RAM_OFS_RX_FIFO0           (CAN0_RAM_OFS_EXT_ID        +  (CAN0_FILTER_EXT_NUM * 2U))
#define CAN0_RAM_OFS_RX_FIFO1           (CAN0_RAM_OFS_RX_FIFO0      +  (CAN0_RX_FIFO0_ELEM_NUM * 18U))
#define CAN0_RAM_OFS_RX_BUFFER          (CAN0_RAM_OFS_RX_FIFO1      +  (CAN0_RX_FIFO1_ELEM_NUM * 18U))
#define CAN0_RAM_OFS_TX_EVENT_FIFO      (CAN0_RAM_OFS_RX_BUFFER     +  (CAN0_RX_BUF_NUM * 18U))
#define CAN0_RAM_OFS_TX_BUFFER          (CAN0_RAM_OFS_TX_EVENT_FIFO +  (CAN0_TX_FIFO_ELEM_NUM * 2U))
#define CAN0_TOT_RAM                    (CAN0_RAM_OFS_TX_BUFFER     + ((CAN0_TX_FIFO_ELEM_NUM + CAN0_RX_BUF_OBJ_NUM) * 18U))


// CAN1 controller configuration ------

#ifndef CAN1_RX_FIFO0_ELEM_NUM
#define CAN1_RX_FIFO0_ELEM_NUM          (64U)
#endif
#if    (CAN1_RX_FIFO0_ELEM_NUM > 64U)
#error  Too many Receive FIFO0 Elements defined for CAN1, maximum number of Receive FIFO0 Elements is 64 !!!
#endif

#ifndef CAN1_RX_FIFO1_ELEM_NUM
#define CAN1_RX_FIFO1_ELEM_NUM          (64U)
#endif
#if    (CAN1_RX_FIFO1_ELEM_NUM > 64U)
#error  Too many Receive FIFO1 Elements defined for CAN1, maximum number of Receive FIFO1 Elements is 64 !!!
#endif

#ifndef CAN1_RX_BUF_NUM
#define CAN1_RX_BUF_NUM                 (64U)
#endif
#if    (CAN1_RX_BUF_NUM > 64U)
#error  Too many Receive Buffers defined for CAN1, maximum number of Receive Buffers is 64 !!!
#endif

#ifndef CAN1_TX_FIFO_ELEM_NUM
#define CAN1_TX_FIFO_ELEM_NUM           (8U)
#endif
#if    (CAN1_TX_FIFO_ELEM_NUM > 32U)
#error  Too many Transmit FIFO Elements defined for CAN1, maximum number of Transmit FIFO Elements is 32 !!!
#endif

#ifndef CAN1_TX_BUF_NUM
#define CAN1_TX_BUF_NUM                 (24U)
#endif
#if    (CAN1_TX_BUF_NUM > 32U)
#error  Too many Transmit Buffers defined for CAN1, maximum number of Transmit Buffers is 32 !!!
#endif

#if    ((CAN1_TX_FIFO_ELEM_NUM + CAN1_TX_BUF_NUM) > 32U)
#error  Too many Transmit Elements and Buffers defined for CAN1, maximum number of Transmit Elements and Buffers together is 32 !!!
#endif

#ifndef CAN1_FILTER_STD_NUM
#define CAN1_FILTER_STD_NUM             (128U)
#endif
#if    (CAN1_FILTER_STD_NUM > 128U)
#error  Standard Filter Elements defined for CAN1, maximum number of Standard Filter Elements is 128 !!!
#endif

#ifndef CAN1_FILTER_EXT_NUM
#define CAN1_FILTER_EXT_NUM             (64U)
#endif
#if    (CAN1_FILTER_EXT_NUM > 64U)
#error  Extended Filter Elements defined for CAN1, maximum number of Extended Filter Elements is 64 !!!
#endif

#ifndef CAN1_TRANSMITTER_DELAY_COMPENSATION_EN
#define CAN1_TRANSMITTER_DELAY_COMPENSATION_EN  (0U)
#endif

#ifndef CAN1_TRANSMITTER_DELAY_OFFSET
#define CAN1_TRANSMITTER_DELAY_OFFSET           (0U)
#endif

#ifndef CAN1_TRANSMITTER_DELAY_FILTER
#define CAN1_TRANSMITTER_DELAY_FILTER           (0U)
#endif
#if   ((CAN1_TRANSMITTER_DELAY_FILTER != 0U) &&  (CHIP_REV_A == 1))
#warning  Transmitter Delay Compensation Filter setting is only available from chip Revision B !!!
#endif

#define CAN1_RX_FIFO_EN                ((CAN1_RX_FIFO0_ELEM_NUM != 0U) || (CAN1_RX_FIFO1_ELEM_NUM != 0U))
#define CAN1_RX_BUF_EN                  (CAN1_RX_BUF_NUM        != 0U)
#define CAN1_TX_FIFO_EN                 (CAN1_TX_FIFO_ELEM_NUM  != 0U)
#define CAN1_TX_BUF_EN                  (CAN1_TX_BUF_NUM        != 0U)
#define CAN1_RX_FIFO_OBJ_NUM           ((CAN1_RX_FIFO0_ELEM_NUM != 0U) + (CAN1_RX_FIFO1_ELEM_NUM != 0U))
#define CAN1_RX_BUF_OBJ_NUM             (CAN1_RX_BUF_NUM)
#define CAN1_RX_TOT_OBJ_NUM             (CAN1_RX_FIFO_OBJ_NUM + CAN1_RX_BUF_OBJ_NUM)
#define CAN1_TX_FIFO_OBJ_NUM           ((CAN1_TX_FIFO_ELEM_NUM  != 0U))
#define CAN1_TX_BUF_OBJ_NUM             (CAN1_TX_BUF_NUM)
#define CAN1_TX_TOT_OBJ_NUM             (CAN1_TX_FIFO_OBJ_NUM + CAN1_TX_BUF_OBJ_NUM)
#define CAN1_TOT_OBJ_NUM                (CAN1_RX_TOT_OBJ_NUM  + CAN1_TX_TOT_OBJ_NUM)
#define CAN1_TOT_ELEM_NUM               (CAN1_RX_FIFO0_ELEM_NUM + CAN1_RX_FIFO1_ELEM_NUM + CAN1_RX_BUF_NUM + CAN1_TX_BUF_NUM)

#define CAN1_RAM_OFS_STD_ID             (0U)
#define CAN1_RAM_OFS_EXT_ID             (CAN1_RAM_OFS_STD_ID        +   CAN1_FILTER_STD_NUM)
#define CAN1_RAM_OFS_RX_FIFO0           (CAN1_RAM_OFS_EXT_ID        +  (CAN1_FILTER_EXT_NUM * 2U))
#define CAN1_RAM_OFS_RX_FIFO1           (CAN1_RAM_OFS_RX_FIFO0      +  (CAN1_RX_FIFO0_ELEM_NUM * 18U))
#define CAN1_RAM_OFS_RX_BUFFER          (CAN1_RAM_OFS_RX_FIFO1      +  (CAN1_RX_FIFO1_ELEM_NUM * 18U))
#define CAN1_RAM_OFS_TX_EVENT_FIFO      (CAN1_RAM_OFS_RX_BUFFER     +  (CAN1_RX_BUF_NUM * 18U))
#define CAN1_RAM_OFS_TX_BUFFER          (CAN1_RAM_OFS_TX_EVENT_FIFO +  (CAN1_TX_FIFO_ELEM_NUM * 2U))
#define CAN1_TOT_RAM                    (CAN1_RAM_OFS_TX_BUFFER     + ((CAN1_TX_FIFO_ELEM_NUM + CAN1_RX_BUF_OBJ_NUM) * 18U))


// CAN Driver ******************************************************************

#define ARM_CAN_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,11) // CAN driver version

// Driver Version
static const ARM_DRIVER_VERSION can_driver_version = { ARM_CAN_API_VERSION, ARM_CAN_DRV_VERSION };

// Driver Capabilities
static const ARM_CAN_CAPABILITIES can_driver_capabilities[2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
  {                             // CAN0 driver capabilities
    CAN0_TOT_OBJ_NUM,           // Number of CAN Objects available
    1U,                         // Supports reentrant calls to ARM_CAN_MessageSend, ARM_CAN_MessageRead, ARM_CAN_ObjectConfigure and abort message sending used by ARM_CAN_Control.
    1U,                         // Supports CAN with flexible data-rate mode (CAN_FD)
    1U,                         // Supports restricted operation mode
    1U,                         // Supports bus monitoring mode
    1U,                         // Supports internal loopback mode
    1U                          // Supports external loopback mode
  }, 
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
  {                             // CAN1 driver capabilities
    CAN1_TOT_OBJ_NUM,           // Number of CAN Objects available
    1U,                         // Supports reentrant calls to ARM_CAN_MessageSend, ARM_CAN_MessageRead, ARM_CAN_ObjectConfigure and abort message sending used by ARM_CAN_Control.
    1U,                         // Supports CAN with flexible data-rate mode (CAN_FD)
    1U,                         // Supports restricted operation mode
    1U,                         // Supports bus monitoring mode
    1U,                         // Supports internal loopback mode
    1U,                         // Supports external loopback mode
  }
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
};

// Object Capabilities
static const ARM_CAN_OBJ_CAPABILITIES can_object_capabilities_rx_fifo0[2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
  {                             // CAN0 Rx FIFO0 object capabilities
    0U,                         // Object does not support transmission
    1U,                         // Object supports reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    1U,                         // Object allows assignment of multiple filters to it
    1U,                         // Object supports exact identifier filtering
    1U,                         // Object supports range identifier filtering
    1U,                         // Object supports mask identifier filtering
    CAN0_RX_FIFO0_ELEM_NUM      // Object can buffer CAN0_RX_FIFO0_ELEM_NUM configured number of messages
  },
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
  {                             // CAN1 Rx FIFO0 object capabilities
    0U,                         // Object does not support transmission
    1U,                         // Object supports reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    1U,                         // Object allows assignment of multiple filters to it
    1U,                         // Object supports exact identifier filtering
    1U,                         // Object supports range identifier filtering
    1U,                         // Object supports mask identifier filtering
    CAN1_RX_FIFO0_ELEM_NUM      // Object can buffer CAN1_RX_FIFO0_ELEM_NUM configured number of messages
  }
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U }
#endif
};
static const ARM_CAN_OBJ_CAPABILITIES can_object_capabilities_rx_fifo1[2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
  {                             // CAN0 Rx FIFO1 object capabilities
    0U,                         // Object does not support transmission
    1U,                         // Object supports reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    1U,                         // Object allows assignment of multiple filters to it
    1U,                         // Object supports exact identifier filtering
    1U,                         // Object supports range identifier filtering
    1U,                         // Object supports mask identifier filtering
    CAN0_RX_FIFO1_ELEM_NUM      // Object can buffer CAN0_RX_FIFO1_ELEM_NUM configured number of messages
  },
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
  {                             // CAN1 Rx FIFO1 object capabilities
    0U,                         // Object does not support transmission
    1U,                         // Object supports reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    1U,                         // Object allows assignment of multiple filters to it
    1U,                         // Object supports exact identifier filtering
    1U,                         // Object supports range identifier filtering
    1U,                         // Object supports mask identifier filtering
    CAN1_RX_FIFO1_ELEM_NUM      // Object can buffer CAN1_RX_FIFO1_ELEM_NUM configured number of messages
  }
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U }
#endif
};
static const ARM_CAN_OBJ_CAPABILITIES can_object_capabilities_rx_buffer[2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
  {                             // CAN0 Rx Buffer object capabilities
    0U,                         // Object does not support transmission
    1U,                         // Object supports reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    1U,                         // Object allows assignment of multiple filters to it
    1U,                         // Object supports exact identifier filtering
    0U,                         // Object does not support range identifier filtering
    0U,                         // Object does not support mask identifier filtering
    1U                          // Object can buffer 1 messages
  },
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
  {                             // CAN1 Rx Buffer object capabilities
    0U,                         // Object does not support transmission
    1U,                         // Object supports reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    1U,                         // Object allows assignment of multiple filters to it
    1U,                         // Object supports exact identifier filtering
    0U,                         // Object does not support range identifier filtering
    0U,                         // Object does not support mask identifier filtering
    1U                          // Object can buffer 1 messages
  }
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U }
#endif
};
static const ARM_CAN_OBJ_CAPABILITIES can_object_capabilities_tx_fifo[2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
  {                             // CAN0 Tx FIFO object capabilities
    1U,                         // Object supports transmission
    0U,                         // Object does not support reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    0U,                         // Object does not allow assignment of multiple filters to it
    0U,                         // Object does not support exact identifier filtering
    0U,                         // Object does not support range identifier filtering
    0U,                         // Object does not support mask identifier filtering
    CAN0_TX_FIFO_ELEM_NUM       // Object can buffer CAN0_TX_FIFO_ELEM_NUM configured number of messages
  },
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
  {                             // CAN1 Tx FIFO object capabilities
    1U,                         // Object supports transmission
    0U,                         // Object does not support reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    0U,                         // Object does not allow assignment of multiple filters to it
    0U,                         // Object does not support exact identifier filtering
    0U,                         // Object does not support range identifier filtering
    0U,                         // Object does not support mask identifier filtering
    CAN1_TX_FIFO_ELEM_NUM       // Object can buffer CAN1_TX_FIFO_ELEM_NUM configured number of messages
  },
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
};
static const ARM_CAN_OBJ_CAPABILITIES can_object_capabilities_tx_buffer[2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
  {                             // CAN0 Tx Buffer object capabilities
    1U,                         // Object supports transmission
    0U,                         // Object does not support reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    0U,                         // Object does not allow assignment of multiple filters to it
    0U,                         // Object does not support exact identifier filtering
    0U,                         // Object does not support range identifier filtering
    0U,                         // Object does not support mask identifier filtering
    1U                          // Object can buffer 1 message
  },
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
  {                             // CAN1 Tx Buffer object capabilities
    1U,                         // Object supports transmission
    0U,                         // Object does not support reception
    0U,                         // Object does not support RTR reception and automatic Data transmission
    0U,                         // Object does not support RTR transmission and automatic Data reception
    0U,                         // Object does not allow assignment of multiple filters to it
    0U,                         // Object does not support exact identifier filtering
    0U,                         // Object does not support range identifier filtering
    0U,                         // Object does not support mask identifier filtering
    1U                          // Object can buffer 1 message
  }
#else
  { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U }
#endif
};

typedef struct _CAN_DRV_CONFIG {// Compile time configuration of driver structure
  uint32_t FILTER_STD_NUM;      // Number of Standard IDs
  uint32_t FILTER_EXT_NUM;      // Number of Extended IDs
  uint32_t RX_FIFO0_ELEM_NUM;   // Number of Rx FIFO0 elements
  uint32_t RX_FIFO1_ELEM_NUM;   // Number of Rx FIFO1 elements
  uint32_t TX_FIFO_ELEM_NUM;    // Number of Tx FIFO elements
  uint32_t RX_FIFO_OBJ_NUM;     // Number of Rx FIFO objects
  uint32_t RX_FIFO_OBJ_IDX_MAX; // Maximum index of Rx FIFO object (0=no max index, 1..n=index+1)
  uint32_t RX_BUF_OBJ_NUM;      // Number of Rx buffer objects
  uint32_t RX_BUF_OBJ_IDX_MAX;  // Maximum index of Rx buffer object (0=no max index, 1..n=index+1)
  uint32_t RX_TOT_OBJ_NUM;      // Number of all Rx FIFO and buffer objects
  uint32_t TX_FIFO_OBJ_NUM;     // Number of Tx FIFO objects
  uint32_t TX_FIFO_OBJ_IDX_MAX; // Maximum index of Tx FIFO object (0=no max index, 1..n=index+1)
  uint32_t TX_BUF_OBJ_NUM;      // Number of Tx buffer objects
  uint32_t TX_BUF_OBJ_IDX_MAX;  // Maximum index of Tx buffer object (0=no max index, 1..n=index+1)
  uint32_t TX_TOT_OBJ_NUM;      // Number of all Tx FIFO and buffer objects
  uint32_t TOT_OBJ_NUM;         // Number of all Rx + Tx objects
} CAN_DRV_CONFIG_t;

typedef struct _CAN_RAM_PTR {   // RAM pointers to areas of filters and buffers structure
  uint32_t *PTR_STD_ID;         // Pointer to Standard IDs RAM
  uint32_t *PTR_EXT_ID;         // Pointer to Extended IDs RAM
  uint32_t *PTR_RX_FIFO0;       // Pointer to Rx FIFO0 RAM
  uint32_t *PTR_RX_FIFO1;       // Pointer to Rx FIFO1 RAM
  uint32_t *PTR_RX_BUFFER;      // Pointer to Rx buffers RAM
  uint32_t *PTR_TX_EVENT_FIFO;  // Pointer to Tx Event FIFO RAM
  uint32_t *PTR_TX_BUFFER;      // Pointer to Tx buffers RAM
} CAN_RAM_PTR_t;

#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
// Statically allocated RAM for filters and buffers for CAN0 controller
static uint32_t         CAN0_RAM[CAN0_TOT_RAM];
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
// Statically allocated RAM for filters and buffers for CAN1 controller
static uint32_t         CAN1_RAM[CAN1_TOT_RAM];
#endif

// Encodings of DLC and LEC to speed up their usage
const  uint8_t          CAN_DLC_ENCODE [16] =  { 0U, 1U, 2U, 3U, 4U, 5U, 6U, 7U, 8U, 12U, 16U, 20U, 24U, 32U, 48U, 64U };
const  uint8_t          CAN_LEC_ENCODE [ 8] =  { ARM_CAN_LEC_NO_ERROR,  ARM_CAN_LEC_STUFF_ERROR, ARM_CAN_LEC_FORM_ERROR, ARM_CAN_LEC_ACK_ERROR,
                                                 ARM_CAN_LEC_BIT_ERROR, ARM_CAN_LEC_BIT_ERROR  , ARM_CAN_LEC_CRC_ERROR,  ARM_CAN_LEC_NO_ERROR };

// Compile time configuration of drivers
const  CAN_DRV_CONFIG_t CAN_DRV_CONFIG [ 2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
                                               { CAN0_FILTER_STD_NUM,
                                                 CAN0_FILTER_EXT_NUM,
                                                 CAN0_RX_FIFO0_ELEM_NUM,
                                                 CAN0_RX_FIFO1_ELEM_NUM,
                                                 CAN0_TX_FIFO_ELEM_NUM,
                                                 CAN0_RX_FIFO_OBJ_NUM,
                                               ( CAN0_RX_FIFO_OBJ_NUM *(CAN0_RX_FIFO_OBJ_NUM != 0U)),
                                                 CAN0_RX_BUF_OBJ_NUM,
                                               ((CAN0_RX_FIFO_OBJ_NUM + CAN0_RX_BUF_OBJ_NUM)*(CAN0_RX_BUF_OBJ_NUM != 0U)),
                                                (CAN0_RX_FIFO_OBJ_NUM + CAN0_RX_BUF_OBJ_NUM),
                                                 CAN0_TX_FIFO_OBJ_NUM,
                                               ((CAN0_RX_FIFO_OBJ_NUM + CAN0_RX_BUF_OBJ_NUM + CAN0_TX_FIFO_OBJ_NUM)*(CAN0_TX_FIFO_OBJ_NUM != 0U)),
                                                 CAN0_TX_BUF_OBJ_NUM,
                                               ((CAN0_RX_FIFO_OBJ_NUM + CAN0_RX_BUF_OBJ_NUM + CAN0_TX_FIFO_OBJ_NUM + CAN0_TX_BUF_OBJ_NUM)*(CAN0_TX_BUF_OBJ_NUM != 0U)),
                                                (CAN0_TX_FIFO_OBJ_NUM + CAN0_TX_BUF_OBJ_NUM),
                                                 CAN0_TOT_OBJ_NUM },
#else
                                               { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
                                               { CAN1_FILTER_STD_NUM,
                                                 CAN1_FILTER_EXT_NUM,
                                                 CAN1_RX_FIFO0_ELEM_NUM,
                                                 CAN1_RX_FIFO1_ELEM_NUM,
                                                 CAN1_TX_FIFO_ELEM_NUM,
                                                 CAN1_RX_FIFO_OBJ_NUM,
                                               ( CAN1_RX_FIFO_OBJ_NUM *(CAN1_RX_FIFO_OBJ_NUM != 0U)),
                                                 CAN1_RX_BUF_OBJ_NUM,
                                               ((CAN1_RX_FIFO_OBJ_NUM + CAN1_RX_BUF_OBJ_NUM)*(CAN1_RX_BUF_OBJ_NUM != 0U)),
                                                (CAN1_RX_FIFO_OBJ_NUM + CAN1_RX_BUF_OBJ_NUM),
                                                 CAN1_TX_FIFO_OBJ_NUM,
                                               ((CAN1_RX_FIFO_OBJ_NUM + CAN1_RX_BUF_OBJ_NUM + CAN1_TX_FIFO_OBJ_NUM)*(CAN1_TX_FIFO_OBJ_NUM != 0U)),
                                                 CAN1_TX_BUF_OBJ_NUM,
                                               ((CAN1_RX_FIFO_OBJ_NUM + CAN1_RX_BUF_OBJ_NUM + CAN1_TX_FIFO_OBJ_NUM + CAN1_TX_BUF_OBJ_NUM)*(CAN1_TX_BUF_OBJ_NUM != 0U)),
                                                (CAN1_TX_FIFO_OBJ_NUM + CAN1_TX_BUF_OBJ_NUM),
                                                 CAN1_TOT_OBJ_NUM },
#else
                                               { 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U }
#endif
                                              };

// Compile time RAM pointers to areas of filters and buffers
const  CAN_RAM_PTR_t    CAN_RAM_PTR    [ 2] = {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
                                               { &CAN0_RAM[CAN0_RAM_OFS_STD_ID],
                                                 &CAN0_RAM[CAN0_RAM_OFS_EXT_ID],
                                                 &CAN0_RAM[CAN0_RAM_OFS_RX_FIFO0],
                                                 &CAN0_RAM[CAN0_RAM_OFS_RX_FIFO1],
                                                 &CAN0_RAM[CAN0_RAM_OFS_RX_BUFFER],
                                                 &CAN0_RAM[CAN0_RAM_OFS_TX_EVENT_FIFO],
                                                 &CAN0_RAM[CAN0_RAM_OFS_TX_BUFFER] },
#else
                                               { NULL, NULL, NULL, NULL, NULL, NULL, NULL },
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
                                               { &CAN1_RAM[CAN1_RAM_OFS_STD_ID],
                                                 &CAN1_RAM[CAN1_RAM_OFS_EXT_ID],
                                                 &CAN1_RAM[CAN1_RAM_OFS_RX_FIFO0],
                                                 &CAN1_RAM[CAN1_RAM_OFS_RX_FIFO1],
                                                 &CAN1_RAM[CAN1_RAM_OFS_RX_BUFFER],
                                                 &CAN1_RAM[CAN1_RAM_OFS_TX_EVENT_FIFO],
                                                 &CAN1_RAM[CAN1_RAM_OFS_TX_BUFFER] }
#else
                                               { NULL, NULL, NULL, NULL, NULL, NULL, NULL }
#endif
                                              };

// Const variables for simplifying access to CAN instances
static Mcan * const    ptr_CANx[2]    = { MCAN0,      MCAN1 };
static const uint32_t  ID_MCANx[2]    = { ID_MCAN0,   ID_MCAN1 };
static const IRQn_Type MCANx_IRQn[2]  = { MCAN0_IRQn, MCAN1_IRQn };
static const Pin       CAN_PIN [4][5] = { RTE_CAN0_RX_PIN, RTE_CAN0_TX_PIN, RTE_CAN1_RX_PIN, RTE_CAN1_TX_PIN };


// Runtime variables
static uint8_t                     can_driver_powered    [2]              = { 0U, 0U };
static uint8_t                     can_driver_initialized[2]              = { 0U, 0U };
static uint8_t                     can_driver_fd_mode    [2]              = { 0U, 0U };
static uint8_t                     can_obj_cfg_access    [CAN_CTRL_NUM];
static uint32_t                    can_obj_rx_cfg        [CAN_CTRL_NUM][2];
static uint32_t                    can_obj_tx_cfg        [CAN_CTRL_NUM];
static uint32_t                    can_tx_obj_active     [CAN_CTRL_NUM];
static uint8_t                     can_filter_std_num    [CAN_CTRL_NUM];
static uint8_t                     can_filter_ext_num    [CAN_CTRL_NUM];
static ARM_CAN_SignalUnitEvent_t   CAN_SignalUnitEvent   [CAN_CTRL_NUM];
static ARM_CAN_SignalObjectEvent_t CAN_SignalObjectEvent [CAN_CTRL_NUM];


// CAN Driver Functions

/**
  \fn          ARM_DRIVER_VERSION CAN_GetVersion (void)
  \brief       Get driver version.
  \return      ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION CAN_GetVersion (void) { return can_driver_version; }

/**
  \fn          ARM_CAN_CAPABILITIES CAN0_GetCapabilities (void)
  \fn          ARM_CAN_CAPABILITIES CAN1_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      ARM_CAN_CAPABILITIES
*/
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static ARM_CAN_CAPABILITIES CAN0_GetCapabilities (void) { return can_driver_capabilities[0U]; }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static ARM_CAN_CAPABILITIES CAN1_GetCapabilities (void) { return can_driver_capabilities[1U]; }
#endif
/**
  \fn          int32_t CANx_Initialize (ARM_CAN_SignalUnitEvent_t   cb_unit_event,
                                        ARM_CAN_SignalObjectEvent_t cb_object_event,
                                        uint8_t                     x)
  \brief       Initialize CAN interface and register signal (callback) functions.
  \param[in]   cb_unit_event   Pointer to ARM_CAN_SignalUnitEvent callback function
  \param[in]   cb_object_event Pointer to ARM_CAN_SignalObjectEvent callback function
  \param[in]   x               Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_Initialize (ARM_CAN_SignalUnitEvent_t   cb_unit_event,
                                ARM_CAN_SignalObjectEvent_t cb_object_event,
                                uint8_t                     x) {

  if (x >= CAN_CTRL_NUM)               { return ARM_DRIVER_ERROR; }
  if (can_driver_initialized[x] != 0U) { return ARM_DRIVER_OK;    }

  CAN_SignalUnitEvent  [x] = cb_unit_event;
  CAN_SignalObjectEvent[x] = cb_object_event;

  if (x == 0U) {
#if (RTE_CAN0_RX_PIN_ID != 0U)
    PIO_Configure(CAN_PIN[0], 1U); 
#endif
#if (RTE_CAN0_TX_PIN_ID != 0U)
    PIO_Configure(CAN_PIN[1], 1U); 
#endif
  } else {
#if (RTE_CAN1_RX_PIN_ID != 0U)
    PIO_Configure(CAN_PIN[2], 1U); 
#endif
#if (RTE_CAN1_TX_PIN_ID != 0U)
    PIO_Configure(CAN_PIN[3], 1U); 
#endif
  }

  can_driver_initialized[x] = 1U;

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_Initialize (ARM_CAN_SignalUnitEvent_t cb_unit_event, ARM_CAN_SignalObjectEvent_t cb_object_event) { return CANx_Initialize (cb_unit_event, cb_object_event, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_Initialize (ARM_CAN_SignalUnitEvent_t cb_unit_event, ARM_CAN_SignalObjectEvent_t cb_object_event) { return CANx_Initialize (cb_unit_event, cb_object_event, 1U); }
#endif

/**
  \fn          int32_t CANx_Uninitialize (uint8_t x)
  \brief       De-initialize CAN interface.
  \param[in]   x      Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_Uninitialize (uint8_t x) {

  if (x >= CAN_CTRL_NUM) { return ARM_DRIVER_ERROR; }

  // Pin unconfigure should be done here but Atmel library does not
  // offer that functionality

  can_driver_initialized[x] = 0U;

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_Uninitialize (void) { return CANx_Uninitialize (0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_Uninitialize (void) { return CANx_Uninitialize (1U); }
#endif

/**
  \fn          int32_t CANx_PowerControl (ARM_POWER_STATE state, uint8_t x)
  \brief       Control CAN interface power.
  \param[in]   state  Power state
                 - ARM_POWER_OFF :  power off: no operation possible
                 - ARM_POWER_LOW :  low power mode: retain state, detect and signal wake-up events
                 - ARM_POWER_FULL : power on: full operation at maximum performance
  \param[in]   x      Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_PowerControl (ARM_POWER_STATE state, uint8_t x) {
  Mcan    *ptr_CAN;
  uint32_t tout;

  if (x >= CAN_CTRL_NUM) { return ARM_DRIVER_ERROR; }

  ptr_CAN = ptr_CANx[x];

  switch (state) {
    case ARM_POWER_OFF:
      can_driver_powered[x] = 0U;
      PMC_EnablePeripheral  (ID_MCANx[x]);
      ptr_CAN->MCAN_IE   =   0U;
      ptr_CAN->MCAN_IR   =   0xFFFFFFFFU;
      ptr_CAN->MCAN_ILE &=  ~MCAN_ILE_EINT0;
      NVIC_ClearPendingIRQ  (MCANx_IRQn[x]);
      NVIC_DisableIRQ       (MCANx_IRQn[x]);
      PMC_DisablePeripheral (ID_MCANx[x]);
      break;

    case ARM_POWER_FULL:
      if (can_driver_initialized[x] == 0U) { return ARM_DRIVER_ERROR; }
      if (can_driver_powered[x]     != 0U) { return ARM_DRIVER_OK;    }

      // Not used: Select MCANx clock to = MCK/1 clock = 150 MHz
      // PMC->PMC_PCK[5] = PMC_PCK_PRES(0U) | PMC_PCK_CSS_MCK;

      // Select MCANx clock to = UPLL/6 clock = 80 MHz
      if ((PMC->PMC_SR & PMC_SR_LOCKU) == 0U) {                         // If UPLL is not already running
        PMC->CKGR_UCKR = CKGR_UCKR_UPLLEN | CKGR_UCKR_UPLLCOUNT(0x0FU); // Enable UPLL 480 MHz
        tout = SystemCoreClock / 2U;
        while(((PMC->PMC_SR & PMC_SR_LOCKU) == 0U) && (tout > 0U)) {    // Wait that UPLL is considered locked by the PMC
          tout --;
        }
        if (tout == 0U) { return ARM_DRIVER_ERROR; }
      }
      PMC->PMC_PCK[5] = PMC_PCK_PRES(5U) | PMC_PCK_CSS_UPLL_CLK;

      PMC->PMC_SCER   = PMC_SCER_PCK5;

      PMC_EnablePeripheral (ID_MCANx[x]);

      // Clear runtime variables
      can_driver_fd_mode[x] = 0U;
      can_obj_cfg_access[x] = 0U;
      can_obj_rx_cfg[x][0]  = 0U;
      can_obj_rx_cfg[x][1]  = 0U;
      can_obj_tx_cfg[x]     = 0U;
      can_tx_obj_active[x]  = 0U;
      can_filter_std_num[x] = 0U;
      can_filter_ext_num[x] = 0U;

      // Set peripheral to initialization and configuration change enable to set RAM pointers
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT; while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;

      // Set upper 16 bits of address of RAM for CAN
      if (x == 0U) {
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
        MATRIX->CCFG_CAN0  = (MATRIX->CCFG_CAN0  & 0x0000FFFFU) | (((uint32_t)CAN0_RAM) & 0xFFFF0000U);

        // Clear RAM for CAN0
        memset((void *)CAN0_RAM, 0U, sizeof(CAN0_RAM));
#endif
      } else {
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
        MATRIX->CCFG_SYSIO = (MATRIX->CCFG_SYSIO & 0x0000FFFFU) | (((uint32_t)CAN1_RAM) & 0xFFFF0000U);

        // Clear RAM for CAN1
        memset((void *)CAN1_RAM, 0U, sizeof(CAN1_RAM));
#endif
      }

      // Configure memory areas for filters and Rx/Tx buffers, FIFOs for Rx and Tx are not used
      // Rx Buffers are configured for 64 byte data field to accommodate maximum CAN FD message
      // Tx Buffers are configured for 64 byte data field to accommodate maximum CAN FD message
      ptr_CAN->MCAN_SIDFC = MCAN_SIDFC_LSS  (CAN_DRV_CONFIG[x].FILTER_STD_NUM)                    |
                            MCAN_SIDFC_FLSSA(((uint32_t)(CAN_RAM_PTR[x].PTR_STD_ID))        >> 2) ;
      ptr_CAN->MCAN_XIDFC = MCAN_XIDFC_LSE  (CAN_DRV_CONFIG[x].FILTER_EXT_NUM)                    |
                            MCAN_XIDFC_FLESA(((uint32_t)(CAN_RAM_PTR[x].PTR_EXT_ID))        >> 2) ;
      ptr_CAN->MCAN_RXF0C = MCAN_RXF0C_F0S  (CAN_DRV_CONFIG[x].RX_FIFO0_ELEM_NUM)                 |
                            MCAN_RXF0C_F0SA (((uint32_t)(CAN_RAM_PTR[x].PTR_RX_FIFO0))      >> 2) ;
      ptr_CAN->MCAN_RXF1C = MCAN_RXF1C_F1S  (CAN_DRV_CONFIG[x].RX_FIFO1_ELEM_NUM)                 |
                            MCAN_RXF1C_F1SA (((uint32_t)(CAN_RAM_PTR[x].PTR_RX_FIFO1))      >> 2) ;
      ptr_CAN->MCAN_RXBC  = MCAN_RXBC_RBSA  (((uint32_t)(CAN_RAM_PTR[x].PTR_RX_BUFFER))     >> 2) ;
      ptr_CAN->MCAN_RXESC = MCAN_RXESC_F0DS (7U)                                                  |
                            MCAN_RXESC_F1DS (7U)                                                  |
                            MCAN_RXESC_RBDS (7U)                                                  ;
      ptr_CAN->MCAN_TXEFC = MCAN_TXEFC_EFS  (CAN_DRV_CONFIG[x].TX_FIFO_ELEM_NUM)                  |
                            MCAN_TXEFC_EFSA (((uint32_t)(CAN_RAM_PTR[x].PTR_TX_EVENT_FIFO)) >> 2) ;
      ptr_CAN->MCAN_TXBC  = MCAN_TXBC_NDTB  (CAN_DRV_CONFIG[x].TX_BUF_OBJ_NUM)                    |
                            MCAN_TXBC_TFQS  (CAN_DRV_CONFIG[x].TX_FIFO_ELEM_NUM)                  |
                            MCAN_TXBC_TBSA  (((uint32_t)(CAN_RAM_PTR[x].PTR_TX_BUFFER))     >> 2) ;
      ptr_CAN->MCAN_TXESC = MCAN_TXESC_TBDS (7U)                                                  ;

      // Reject non-matching frames
      ptr_CAN->MCAN_GFC   = MCAN_GFC_ANFE(3U) | MCAN_GFC_ANFS(3U);

      // Route all interrupt sources to m_can_int0 and enable it
      ptr_CAN->MCAN_ILS = 0U;
      ptr_CAN->MCAN_ILE = MCAN_ILE_EINT0;

      // Enable relevant interrupts
      ptr_CAN->MCAN_IR  = 0xFFFFFFFFU;
      ptr_CAN->MCAN_IE  = MCAN_IE_TCE | MCAN_IE_DRXE | MCAN_IE_EPE | MCAN_IE_EWE | MCAN_IE_BOE;

      can_driver_powered[x] = 1U;

      NVIC_ClearPendingIRQ (MCANx_IRQn[x]);
      if ((CAN_SignalUnitEvent[x] != NULL) || (CAN_SignalObjectEvent[x] != NULL)) {
        NVIC_EnableIRQ     (MCANx_IRQn[x]);
      }
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_PowerControl (ARM_POWER_STATE state) { return CANx_PowerControl (state, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_PowerControl (ARM_POWER_STATE state) { return CANx_PowerControl (state, 1U); }
#endif

/**
  \fn          uint32_t CAN_GetClock (void)
  \brief       Retrieve CAN base clock frequency.
  \return      base clock frequency
*/
uint32_t CAN_GetClock (void) {
  uint32_t clk;

  switch ((PMC->PMC_PCK[5] & PMC_PCK_CSS_Msk) >> PMC_PCK_CSS_Pos) {
    case PMC_PCK_CSS_SLOW_CLK:
      if ((SUPC->SUPC_SR & SUPC_SR_OSCSEL) != 0U) {
        clk = CHIP_FREQ_XTAL_32K;
      } else {
        clk = CHIP_FREQ_SLCK_RC;
      }
      break;
    case PMC_PCK_CSS_MAIN_CLK:
      if ((PMC->CKGR_MOR & CKGR_MOR_MOSCSEL) != 0U) {
        clk = CHIP_FREQ_XTAL_12M;
      } else {
        clk = CHIP_FREQ_MAINCK_RC_4MHZ;

        switch (PMC->CKGR_MOR & CKGR_MOR_MOSCRCF_Msk) {
          case CKGR_MOR_MOSCRCF_4_MHz:
            break;
          case CKGR_MOR_MOSCRCF_8_MHz:
            clk *= 2U;
            break;
          case CKGR_MOR_MOSCRCF_12_MHz:
            clk *= 3U;
            break;
          default:
            break;
        }
      }

      if ((PMC->PMC_MCKR & PMC_MCKR_PRES_Msk) == PMC_MCKR_PRES_CLK_3) {
        clk /= 3U;
      } else {
        clk >>= ((PMC->PMC_MCKR & PMC_MCKR_PRES_Msk) >> PMC_MCKR_PRES_Pos);
      }
      break;
    case PMC_PCK_CSS_PLLA_CLK:
      if ((PMC->CKGR_MOR & CKGR_MOR_MOSCSEL) != 0U) {
        clk = CHIP_FREQ_XTAL_12M;
      } else {
        clk = CHIP_FREQ_MAINCK_RC_4MHZ;

        switch (PMC->CKGR_MOR & CKGR_MOR_MOSCRCF_Msk) {
          case CKGR_MOR_MOSCRCF_4_MHz:
            break;
          case CKGR_MOR_MOSCRCF_8_MHz:
            clk *= 2U;
            break;
          case CKGR_MOR_MOSCRCF_12_MHz:
            clk *= 3U;
            break;
          default:
            break;
        }
      }

      clk *= ((((PMC->CKGR_PLLAR) & CKGR_PLLAR_MULA_Msk) >> CKGR_PLLAR_MULA_Pos) + 1U);
      clk /= ((((PMC->CKGR_PLLAR) & CKGR_PLLAR_DIVA_Msk) >> CKGR_PLLAR_DIVA_Pos));
      break;
    case PMC_PCK_CSS_MCK:
      clk = SystemCoreClock / (((PMC->PMC_MCKR & PMC_MCKR_MDIV_Msk) >> PMC_MCKR_MDIV_Pos) + 1U);
      break;
    case PMC_PCK_CSS_UPLL_CLK:
      clk = 480000000U / (((PMC->PMC_MCKR & PMC_MCKR_UPLLDIV2) >> 13) + 1U);
      break;
    default:
      clk = 0U;
      break;
  }
  clk /= ((PMC->PMC_PCK[5] & PMC_PCK_PRES_Msk) >> PMC_PCK_PRES_Pos) + 1U;

  return clk;
}


/**
  \fn          int32_t CANx_SetBitrate (ARM_CAN_BITRATE_SELECT select, uint32_t bitrate, uint32_t bit_segments, uint8_t x)
  \brief       Set bitrate for CAN interface.
  \param[in]   select       Bitrate selection
                 - ARM_CAN_BITRATE_NOMINAL : nominal (flexible data-rate arbitration) bitrate
                 - ARM_CAN_BITRATE_FD_DATA : flexible data-rate data bitrate
  \param[in]   bitrate      Bitrate
  \param[in]   bit_segments Bit segments settings
  \param[in]   x            Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_SetBitrate (ARM_CAN_BITRATE_SELECT select, uint32_t bitrate, uint32_t bit_segments, uint8_t x) {
  Mcan     *ptr_CAN;
  uint32_t  cccr, sjw, prop_seg, phase_seg1, phase_seg2, pclk, brp, brp_max, tq_num;

  if (x >= CAN_CTRL_NUM)                                                            { return ARM_DRIVER_ERROR;               }
  if ((select != ARM_CAN_BITRATE_NOMINAL) && (select != ARM_CAN_BITRATE_FD_DATA)  ) { return ARM_CAN_INVALID_BITRATE_SELECT; }
  if (can_driver_powered[x] == 0U)                                                  { return ARM_DRIVER_ERROR;               }

  prop_seg   = (bit_segments & ARM_CAN_BIT_PROP_SEG_Msk  ) >> ARM_CAN_BIT_PROP_SEG_Pos;
  phase_seg1 = (bit_segments & ARM_CAN_BIT_PHASE_SEG1_Msk) >> ARM_CAN_BIT_PHASE_SEG1_Pos;
  phase_seg2 = (bit_segments & ARM_CAN_BIT_PHASE_SEG2_Msk) >> ARM_CAN_BIT_PHASE_SEG2_Pos;
  sjw        = (bit_segments & ARM_CAN_BIT_SJW_Msk       ) >> ARM_CAN_BIT_SJW_Pos;

#if (CHIP_REV_A == 1)                           // Chip revision MRL A
  if (select == ARM_CAN_BITRATE_NOMINAL) {
    if (((prop_seg + phase_seg1) < 2U) || ((prop_seg + phase_seg1) >  64U)) { return ARM_CAN_INVALID_BIT_PROP_SEG;   }
    if (( phase_seg2             < 1U) || ( phase_seg2             >  16U)) { return ARM_CAN_INVALID_BIT_PHASE_SEG2; }
    if (( sjw                    < 1U) || ( sjw                    >  16U)) { return ARM_CAN_INVALID_BIT_SJW;        }
    brp_max = 1024U;
  } else {
    if (((prop_seg + phase_seg1) < 2U) || ((prop_seg + phase_seg1) >  16U)) { return ARM_CAN_INVALID_BIT_PROP_SEG;   }
    if (( phase_seg2             < 1U) || ( phase_seg2             >   8U)) { return ARM_CAN_INVALID_BIT_PHASE_SEG2; }
    if (( sjw                    < 1U) || ( sjw                    >   4U)) { return ARM_CAN_INVALID_BIT_SJW;        }
    brp_max = 32U;
  }
#else                                           // Chip revision MRL B
  if (select == ARM_CAN_BITRATE_NOMINAL) {
    if (((prop_seg + phase_seg1) < 2U) || ((prop_seg + phase_seg1) > 256U)) { return ARM_CAN_INVALID_BIT_PROP_SEG;   }
    if (( phase_seg2             < 1U) || ( phase_seg2             > 128U)) { return ARM_CAN_INVALID_BIT_PHASE_SEG2; }
    if (( sjw                    < 1U) || ( sjw                    > 128U)) { return ARM_CAN_INVALID_BIT_SJW;        }
    brp_max = 512U;
  } else {
    if (((prop_seg + phase_seg1) < 2U) || ((prop_seg + phase_seg1) >  32U)) { return ARM_CAN_INVALID_BIT_PROP_SEG;   }
    if (( phase_seg2             < 1U) || ( phase_seg2             >  16U)) { return ARM_CAN_INVALID_BIT_PHASE_SEG2; }
    if (( sjw                    < 1U) || ( sjw                    >   8U)) { return ARM_CAN_INVALID_BIT_SJW;        }
    brp_max = 32U;
  }
#endif

  tq_num = 1U + prop_seg + phase_seg1 + phase_seg2;
  pclk   = CAN_GetClock ();           if (pclk == 0U)    { return ARM_DRIVER_ERROR;        }
  brp    = pclk / (tq_num * bitrate); if (brp > brp_max) { return ARM_CAN_INVALID_BITRATE; }
  if (pclk > (brp * tq_num * bitrate)) {
    if ((((pclk - (brp * tq_num * bitrate)) * 1024U) / pclk) > CAN_CLOCK_TOLERANCE) { return ARM_CAN_INVALID_BITRATE; }
  } else if (pclk < (brp * tq_num * bitrate)) {
    if (((((brp * tq_num * bitrate) - pclk) * 1024U) / pclk) > CAN_CLOCK_TOLERANCE) { return ARM_CAN_INVALID_BITRATE; }
  }

  ptr_CAN = ptr_CANx[x];

  cccr = ptr_CAN->MCAN_CCCR;
  if ((cccr & (MCAN_CCCR_CCE | MCAN_CCCR_INIT)) != (MCAN_CCCR_CCE | MCAN_CCCR_INIT)) {
    ptr_CAN->MCAN_CCCR = MCAN_CCCR_CCE  |       // Configuration change enable
                         MCAN_CCCR_INIT ;       // Initialization
    while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
  }
#if (CHIP_REV_A == 1)                           // Chip revision MRL A
  if (select == ARM_CAN_BITRATE_NOMINAL) {
    ptr_CAN->MCAN_BTP  = MCAN_BTP_BRP  (brp - 1U) | MCAN_BTP_TSEG1  (prop_seg + phase_seg1 - 1U) | MCAN_BTP_TSEG2  (phase_seg2 - 1U) | MCAN_BTP_SJW  (sjw - 1U);
  } else {
    ptr_CAN->MCAN_FBTP = MCAN_FBTP_FBRP(brp - 1U) | MCAN_FBTP_FTSEG1(prop_seg + phase_seg1 - 1U) | MCAN_FBTP_FTSEG2(phase_seg2 - 1U) | MCAN_FBTP_FSJW(sjw - 1U);
  }
#else                                           // Chip revision MRL B
  if (select == ARM_CAN_BITRATE_NOMINAL) {
    // Register was renamed to NBTP but it is on same location as BTP
    ptr_CAN->MCAN_NBTP = ((brp - 1U) << 16) | ((prop_seg + phase_seg1 - 1U) << 8) |  (phase_seg2 - 1U)       | ((sjw - 1U) << 25);
  } else {
    // Register was renamed to DBTP but it is on same location as FBTP
    ptr_CAN->MCAN_DBTP = ((brp - 1U) << 16) | ((prop_seg + phase_seg1 - 1U) << 8) | ((phase_seg2 - 1U) << 4) |  (sjw - 1U);
  }
#endif
  ptr_CAN->MCAN_CCCR = cccr;

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_SetBitrate (ARM_CAN_BITRATE_SELECT select, uint32_t bitrate, uint32_t bit_segments) { return CANx_SetBitrate (select, bitrate, bit_segments, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_SetBitrate (ARM_CAN_BITRATE_SELECT select, uint32_t bitrate, uint32_t bit_segments) { return CANx_SetBitrate (select, bitrate, bit_segments, 1U); }
#endif

/**
  \fn          int32_t CANx_SetMode (ARM_CAN_MODE mode, uint8_t x)
  \brief       Set operating mode for CAN interface.
  \param[in]   mode   Operating mode
                 - ARM_CAN_MODE_INITIALIZATION :    initialization mode
                 - ARM_CAN_MODE_NORMAL :            normal operation mode
                 - ARM_CAN_MODE_RESTRICTED :        restricted operation mode
                 - ARM_CAN_MODE_MONITOR :           bus monitoring mode
                 - ARM_CAN_MODE_LOOPBACK_INTERNAL : loopback internal mode
                 - ARM_CAN_MODE_LOOPBACK_EXTERNAL : loopback external mode
  \param[in]   x      Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_SetMode (ARM_CAN_MODE mode, uint8_t x) {
  Mcan     *ptr_CAN;
  uint32_t  event;

  if (x >= CAN_CTRL_NUM)           { return ARM_DRIVER_ERROR; }
  if (can_driver_powered[x] == 0U) { return ARM_DRIVER_ERROR; }

  ptr_CAN = ptr_CANx[x];

  event = 0U;
  switch (mode) {
    case ARM_CAN_MODE_INITIALIZATION:
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE | MCAN_CCCR_TEST_ENABLED;
      ptr_CAN->MCAN_TEST = 0U;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      event              = ARM_CAN_EVENT_UNIT_BUS_OFF;
      break;
    case ARM_CAN_MODE_NORMAL:
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE | MCAN_CCCR_TEST_ENABLED;
      ptr_CAN->MCAN_TEST = 0U;
      ptr_CAN->MCAN_CCCR = 0U;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 1U);
      event              = ARM_CAN_EVENT_UNIT_ACTIVE;
      break;
    case ARM_CAN_MODE_RESTRICTED:
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE | MCAN_CCCR_TEST_ENABLED | MCAN_CCCR_ASM_RESTRICTED;
      ptr_CAN->MCAN_TEST = 0U;
      ptr_CAN->MCAN_CCCR =                                                           MCAN_CCCR_ASM_RESTRICTED;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 1U);
      event              = ARM_CAN_EVENT_UNIT_ACTIVE;
      break;
    case ARM_CAN_MODE_MONITOR:
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE | MCAN_CCCR_TEST_ENABLED | MCAN_CCCR_MON_ENABLED;
      ptr_CAN->MCAN_TEST = 0U;
      ptr_CAN->MCAN_CCCR =                                                           MCAN_CCCR_MON_ENABLED;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 1U);
      event              = ARM_CAN_EVENT_UNIT_PASSIVE;
      break;
    case ARM_CAN_MODE_LOOPBACK_INTERNAL:
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE | MCAN_CCCR_TEST_ENABLED | MCAN_CCCR_MON_ENABLED;
      ptr_CAN->MCAN_TEST = MCAN_TEST_LBCK_ENABLED;
      ptr_CAN->MCAN_CCCR =                                  MCAN_CCCR_TEST_ENABLED | MCAN_CCCR_MON_ENABLED;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 1U);
      event              = ARM_CAN_EVENT_UNIT_PASSIVE;
      break;
    case ARM_CAN_MODE_LOOPBACK_EXTERNAL:
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = MCAN_CCCR_INIT | MCAN_CCCR_CCE | MCAN_CCCR_TEST_ENABLED;
      ptr_CAN->MCAN_TEST = MCAN_TEST_LBCK_ENABLED;
      ptr_CAN->MCAN_CCCR =                                  MCAN_CCCR_TEST_ENABLED;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 1U);
      event              = ARM_CAN_EVENT_UNIT_ACTIVE;
      break;
    default:
      return ARM_DRIVER_ERROR_PARAMETER;
  }
  if ((CAN_SignalUnitEvent[x] != NULL) && (event != 0U)) { CAN_SignalUnitEvent[x](event); }

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_SetMode (ARM_CAN_MODE mode) { return CANx_SetMode (mode, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_SetMode (ARM_CAN_MODE mode) { return CANx_SetMode (mode, 1U); }
#endif

/**
  \fn          ARM_CAN_OBJ_CAPABILITIES CANx_ObjectGetCapabilities (uint32_t obj_idx, uint8_t x)
  \brief       Retrieve capabilities of an object.
  \param[in]   obj_idx  Object index
  \param[in]   x        Controller number (0..1)
  \return      ARM_CAN_OBJ_CAPABILITIES
*/
ARM_CAN_OBJ_CAPABILITIES CANx_ObjectGetCapabilities (uint32_t obj_idx, uint8_t x) {
  ARM_CAN_OBJ_CAPABILITIES obj_cap_null;

  if ((x >= CAN_CTRL_NUM) || (obj_idx >= CAN_DRV_CONFIG[x].TOT_OBJ_NUM)) {
    memset (&obj_cap_null, 0U, sizeof(ARM_CAN_OBJ_CAPABILITIES));
    return obj_cap_null;
  }

  if         (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
    if (obj_idx == 0U) {
      return can_object_capabilities_rx_fifo0[x];
    } else {
      return can_object_capabilities_rx_fifo1[x];
    }
  } else if  (obj_idx < CAN_DRV_CONFIG[x].RX_BUF_OBJ_IDX_MAX) {
    return can_object_capabilities_rx_buffer[x];
  } else if  (obj_idx < CAN_DRV_CONFIG[x].TX_FIFO_OBJ_IDX_MAX) {
    return can_object_capabilities_tx_fifo[x];
  } else {
    return can_object_capabilities_tx_buffer[x];
  }
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
ARM_CAN_OBJ_CAPABILITIES CAN0_ObjectGetCapabilities (uint32_t obj_idx) { return CANx_ObjectGetCapabilities (obj_idx, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
ARM_CAN_OBJ_CAPABILITIES CAN1_ObjectGetCapabilities (uint32_t obj_idx) { return CANx_ObjectGetCapabilities (obj_idx, 1U); }
#endif

/**
  \fn          int32_t CANx_ObjectSetFilter (uint32_t obj_idx, ARM_CAN_FILTER_OPERATION operation, uint32_t id, uint32_t arg, uint8_t x)
  \brief       Add or remove filter for message reception.
  \param[in]   obj_idx      Object index of object that filter should be or is assigned to
  \param[in]   operation    Operation on filter
                 - ARM_CAN_FILTER_ID_EXACT_ADD :       add    exact id filter
                 - ARM_CAN_FILTER_ID_EXACT_REMOVE :    remove exact id filter
                 - ARM_CAN_FILTER_ID_RANGE_ADD :       add    range id filter
                 - ARM_CAN_FILTER_ID_RANGE_REMOVE :    remove range id filter
                 - ARM_CAN_FILTER_ID_MASKABLE_ADD :    add    maskable id filter
                 - ARM_CAN_FILTER_ID_MASKABLE_REMOVE : remove maskable id filter
  \param[in]   id           ID or start of ID range (depending on filter type)
  \param[in]   arg          Mask or end of ID range (depending on filter type)
  \param[in]   x            Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_ObjectSetFilter (uint32_t obj_idx, ARM_CAN_FILTER_OPERATION operation, uint32_t id, uint32_t arg, uint8_t x) {
  uint32_t *ptr_filter_RAM;
  uint32_t *cache_addr;
  uint32_t  i, cache_size, val0, val1;

  if (x >= CAN_CTRL_NUM)                           { return ARM_DRIVER_ERROR; }
  if (obj_idx >= CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) { return ARM_DRIVER_ERROR; }
  if (can_driver_powered[x] == 0U)                 { return ARM_DRIVER_ERROR; }

  switch (operation) {
    case ARM_CAN_FILTER_ID_MASKABLE_ADD:                // Add
    case ARM_CAN_FILTER_ID_RANGE_ADD:
      // Maskable and range filters are only supported on Rx FIFO
      if (obj_idx >= CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) { return ARM_DRIVER_ERROR; }
    case ARM_CAN_FILTER_ID_EXACT_ADD:
      if ((id & ARM_CAN_ID_IDE_Msk) != 0U) {            // Extended Identifier
        if (can_filter_ext_num[x] >= CAN_DRV_CONFIG[x].FILTER_EXT_NUM) { return ARM_DRIVER_ERROR; }
        switch (operation) {
          case ARM_CAN_FILTER_ID_EXACT_ADD:
            if (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
              // If Rx FIFO is used for reception on this object
              ptr_filter_RAM = CAN_RAM_PTR[x].PTR_EXT_ID;
              for (i = 0U; i < can_filter_ext_num[x]; i++) {
                if ((((ptr_filter_RAM[i*2U+1U] >> 30) & 3U) == 1U)             &&       // If Dual ID entry
                    (((ptr_filter_RAM[i*2U]    >> 29) & 7U) == (obj_idx + 1U))) {       // If same FIFO is assigned
                  if        ((ptr_filter_RAM[i*2U]    & 0x1FFFFFFFU) == 0U) {
                    // First entry in Dual ID entry is unused
                    ptr_filter_RAM[i*2U]    |= (id  & 0x1FFFFFFFU);
                    break;
                  } else if ((ptr_filter_RAM[i*2U+1U] & 0x1FFFFFFFU) == 0U) {
                    // Second entry in Dual ID entry is unused
                    ptr_filter_RAM[i*2U+1U] |= (id  & 0x1FFFFFFFU);
                    break;
                  }
                }
              }
              if (i == can_filter_ext_num[x]) {
                // No free entry in Dual ID entries, add new Dual ID entry
                ptr_filter_RAM    = (uint32_t *)(CAN_RAM_PTR[x].PTR_EXT_ID + can_filter_ext_num[x] * 2U);
                ptr_filter_RAM[0] = ((obj_idx + 1U) << 29) | (id  & 0x1FFFFFFFU);
                ptr_filter_RAM[1] = (1U             << 30);
                can_filter_ext_num[x]++;
              }
            } else {
              // If Rx Buffer is used for reception on this object
              ptr_filter_RAM    = (uint32_t *)(CAN_RAM_PTR[x].PTR_EXT_ID + can_filter_ext_num[x] * 2U);
              ptr_filter_RAM[0] = (7U             << 29) | (id & 0x1FFFFFFFU);
              ptr_filter_RAM[1] =  obj_idx - CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM;
              can_filter_ext_num[x]++;
            }
            break;
          case ARM_CAN_FILTER_ID_MASKABLE_ADD:
            ptr_filter_RAM    =   (uint32_t *)(CAN_RAM_PTR[x].PTR_EXT_ID + can_filter_ext_num[x] * 2U);
            ptr_filter_RAM[0] =   ((obj_idx + 1U) << 29) | (id  & 0x1FFFFFFFU);
            ptr_filter_RAM[1] =   (2U             << 30) | (arg & 0x1FFFFFFFU);
            can_filter_ext_num[x]++;
            break;
          case ARM_CAN_FILTER_ID_RANGE_ADD:
            ptr_filter_RAM    =   (uint32_t *)(CAN_RAM_PTR[x].PTR_EXT_ID + can_filter_ext_num[x] * 2U);
            ptr_filter_RAM[0] =   ((obj_idx + 1U) << 29) | (id  & 0x1FFFFFFFU);
            ptr_filter_RAM[1] =                            (arg & 0x1FFFFFFFU);
            can_filter_ext_num[x]++;
            break;
          default:
            return ARM_DRIVER_ERROR;
        }
      } else {                                          // Standard Identifier
        if (can_filter_std_num[x] >= CAN_DRV_CONFIG[x].FILTER_STD_NUM) { return ARM_DRIVER_ERROR; }
        switch (operation) {
          case ARM_CAN_FILTER_ID_EXACT_ADD:
            if (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
              // If Rx FIFO is used for reception on this object
              ptr_filter_RAM = CAN_RAM_PTR[x].PTR_STD_ID;
              for (i = 0U; i < can_filter_std_num[x]; i++) {
                if ((((ptr_filter_RAM[i] >> 30) & 3U) == 1U)             &&     // If Dual ID entry
                    (((ptr_filter_RAM[i] >> 27) & 7U) == (obj_idx + 1U))) {     // If same FIFO is assigned
                  if        ((ptr_filter_RAM[i] & (0x7FFU << 16)) == 0U) {
                    // First entry in Dual ID entry is unused
                    ptr_filter_RAM[i] |= ((id & 0x7FFU) << 16);
                    break;
                  } else if ((ptr_filter_RAM[i] & (0x7FFU)      ) == 0U) {
                    // Second entry in Dual ID entry is unused
                    ptr_filter_RAM[i] |=  (id & 0x7FFU);
                    break;
                  }
                }
              }
              if (i == can_filter_std_num[x]) {
                // No free entry in Dual ID entries, add new Dual ID entry
                ptr_filter_RAM    = (uint32_t *)(CAN_RAM_PTR[x].PTR_STD_ID + can_filter_std_num[x]);
                ptr_filter_RAM[0] = (1U << 30) | ((obj_idx + 1U) << 27) | ((id & 0x7FFU) << 16);
                can_filter_std_num[x]++;
              }
            } else {
              // If Rx Buffer is used for reception on this object
              ptr_filter_RAM    =   (uint32_t *)(CAN_RAM_PTR[x].PTR_STD_ID + can_filter_std_num[x]);
              ptr_filter_RAM[0] =               (7U            << 27) | ((id & 0x7FFU) << 16) | (obj_idx - CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM);
              can_filter_std_num[x]++;
            }
            break;
          case ARM_CAN_FILTER_ID_MASKABLE_ADD:
            ptr_filter_RAM    =     (uint32_t *)(CAN_RAM_PTR[x].PTR_STD_ID + can_filter_std_num[x]);
            ptr_filter_RAM[0] =     (2U << 30) | ((obj_idx + 1U) << 27) | ((id & 0x7FFU) << 16) | (arg & 0x7FFU);
            can_filter_std_num[x]++;
            break;
          case ARM_CAN_FILTER_ID_RANGE_ADD:
            ptr_filter_RAM    =     (uint32_t *)(CAN_RAM_PTR[x].PTR_STD_ID + can_filter_std_num[x]);
            ptr_filter_RAM[0] =                  ((obj_idx + 1U) << 27) | ((id & 0x7FFU) << 16) | (arg & 0x7FFU);
            can_filter_std_num[x]++;
            break;
          default:
            return ARM_DRIVER_ERROR;
        }
      }
      break;
    case ARM_CAN_FILTER_ID_MASKABLE_REMOVE:             // Remove
    case ARM_CAN_FILTER_ID_RANGE_REMOVE:
      // Maskable and range filters are only supported on Rx FIFO
      if (obj_idx >= CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) { return ARM_DRIVER_ERROR; }
    case ARM_CAN_FILTER_ID_EXACT_REMOVE:
      if ((id & ARM_CAN_ID_IDE_Msk) != 0U) {            // Extended Identifier
        if (can_filter_ext_num[x] == 0U) { return ARM_DRIVER_ERROR; }
        switch (operation) {
          case ARM_CAN_FILTER_ID_EXACT_REMOVE:
            if (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
              // If Rx FIFO is used for reception on this object
              ptr_filter_RAM = CAN_RAM_PTR[x].PTR_EXT_ID;
              for (i = 0U; i < can_filter_ext_num[x]; i++) {
                if ((((ptr_filter_RAM[i*2U+1U] >> 30) & 3U) == 1U)             &&       // If Dual ID entry
                    (((ptr_filter_RAM[i*2U]    >> 29) & 7U) == (obj_idx + 1U))) {       // If same FIFO is assigned
                  if        ((ptr_filter_RAM[i*2U]    & 0x1FFFFFFFU) == (id & 0x1FFFFFFFU)) {
                    // First entry in Dual ID entry is the one to be removed
                    ptr_filter_RAM[i*2U] &= ~0x1FFFFFFFU;
                    if ((ptr_filter_RAM[i*2U+1U] & 0x1FFFFFFFU) == 0U) {
                      // Dual ID entry is free it can be removed
                      // Shift all remaining Extended ID filter entries 1 entry (64 bits) towards start of table
                      for (; i < can_filter_ext_num[x]; i++) {
                        ptr_filter_RAM[i*2U   ] = ptr_filter_RAM[i*2U+2U];
                        ptr_filter_RAM[i*2U+1U] = ptr_filter_RAM[i*2U+3U];
                      }
                      can_filter_ext_num[x]--;
                    }
                    return ARM_DRIVER_OK;
                  } else if ((ptr_filter_RAM[i*2U+1U] & 0x1FFFFFFFU) == (id & 0x1FFFFFFFU)) {
                    // Second entry in Dual ID entry is the one to be removed
                    ptr_filter_RAM[i*2U+1U] &= ~0x1FFFFFFFU;
                    if ((ptr_filter_RAM[i*2U]    & 0x1FFFFFFFU) == 0U) {
                      // Dual ID entry is free it can be removed
                      // Shift all remaining Extended ID filter entries 1 entry (64 bits) towards start of table
                      for (; i < can_filter_ext_num[x]; i++) {
                        ptr_filter_RAM[i*2U   ] = ptr_filter_RAM[i*2U+2U];
                        ptr_filter_RAM[i*2U+1U] = ptr_filter_RAM[i*2U+3U];
                      }
                      can_filter_ext_num[x]--;
                    }
                    return ARM_DRIVER_OK;
                  }
                }
              }
            } else {
              // If Rx Buffer is used for reception on this object
              val0 = (7U             << 29) | (id & 0x1FFFFFFFU);
              val1 =  obj_idx - CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM;
            }
            break;
          case ARM_CAN_FILTER_ID_MASKABLE_REMOVE:
            val0 =   ((obj_idx + 1U) << 29) | (id  & 0x1FFFFFFFU);
            val1 =   (2U             << 30) | (arg & 0x1FFFFFFFU);
            break;
          case ARM_CAN_FILTER_ID_RANGE_REMOVE:
            val0 =   ((obj_idx + 1U) << 29) | (id  & 0x1FFFFFFFU);
            val1 =                            (arg & 0x1FFFFFFFU);
            break;
          default:
            return ARM_DRIVER_ERROR;
        }
        ptr_filter_RAM = CAN_RAM_PTR[x].PTR_EXT_ID;
        for (i = 0U; i < can_filter_ext_num[x]; i++) {
          if ((ptr_filter_RAM[i*2U]    == val0) &&
              (ptr_filter_RAM[i*2U+1U] == val1)) {
            // Entry to be removed was found
            // Shift all remaining Extended ID filter entries 1 entry (64 bits) towards start of table
            for (; i < can_filter_ext_num[x]; i++) {
              ptr_filter_RAM[i*2U   ] = ptr_filter_RAM[i*2U+2U];
              ptr_filter_RAM[i*2U+1U] = ptr_filter_RAM[i*2U+3U];
            }
            can_filter_ext_num[x]--;
            break;
          }
        }
      } else {                                          // Standard Identifier
        if (can_filter_std_num[x] == 0U) { return ARM_DRIVER_ERROR; }
        switch (operation) {
          case ARM_CAN_FILTER_ID_EXACT_REMOVE:
            if (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
              // If Rx FIFO is used for reception on this object
              ptr_filter_RAM = CAN_RAM_PTR[x].PTR_STD_ID;
              for (i = 0U; i < can_filter_std_num[x]; i++) {
                if ((((ptr_filter_RAM[i] >> 30) & 3U) == 1U)             &&     // If Dual ID entry
                    (((ptr_filter_RAM[i] >> 27) & 7U) == (obj_idx + 1U))) {     // If same FIFO is assigned
                  if        (((ptr_filter_RAM[i] >> 16) & 0x7FFU) == (id & 0x7FFU)) {
                    // First entry in Dual ID entry is the one to be removed
                    ptr_filter_RAM[i] &= ~(0x7FFU << 16);
                    if ((ptr_filter_RAM[i] & 0x7FFU) == 0U) {
                      // Dual ID entry is free it can be removed
                      // Shift all remaining Standard ID filter entries 1 entry (32 bits) towards start of table
                      for (; i < can_filter_std_num[x]; i++) {
                        ptr_filter_RAM[i] = ptr_filter_RAM[i+1U];
                      }
                      can_filter_std_num[x]--;
                    }
                    return ARM_DRIVER_OK;
                  } else if ( (ptr_filter_RAM[i]        & 0x7FFU) == (id & 0x7FFU)) {
                    // Second entry in Dual ID entry is the one to be removed
                    ptr_filter_RAM[i] &= ~0x7FFU;
                    if ((ptr_filter_RAM[i] & (0x7FFU << 16)) == 0U) {
                      // Dual ID entry is free it can be removed
                      // Shift all remaining Standard ID filter entries 1 entry (32 bits) towards start of table
                      for (; i < can_filter_std_num[x]; i++) {
                        ptr_filter_RAM[i] = ptr_filter_RAM[i+1U];
                      }
                      can_filter_std_num[x]--;
                    }
                    return ARM_DRIVER_OK;
                  }
                }
              }
            } else {
              // If Rx Buffer is used for reception on this object
              val0 =               (7U            << 27) | ((id & 0x7FFU) << 16) | (obj_idx - CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM);
            }
            break;
          case ARM_CAN_FILTER_ID_MASKABLE_REMOVE:
            val0 =   (2U << 30) | ((obj_idx + 1U) << 27) | ((id & 0x7FFU) << 16) | (arg & 0x7FFU);
            break;
          case ARM_CAN_FILTER_ID_RANGE_REMOVE:
            val0 =                ((obj_idx + 1U) << 27) | ((id & 0x7FFU) << 16) | (arg & 0x7FFU);
            break;
          default:
            return ARM_DRIVER_ERROR;
        }
        ptr_filter_RAM = CAN_RAM_PTR[x].PTR_STD_ID;
        for (i = 0U; i < can_filter_std_num[x]; i++) {
          if (ptr_filter_RAM[i] == val0) {
            // Entry to be removed was found
            // Shift all remaining Standard ID filter entries 1 entry (32 bits) towards start of table
            for (; i < can_filter_std_num[x]; i++) {
              ptr_filter_RAM[i] = ptr_filter_RAM[i+1U];
            }
            can_filter_std_num[x]--;
            break;
          }
        }
      }
      break;
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

#if (defined(__DCACHE_PRESENT) && (__DCACHE_PRESENT == 1U))
  if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {              // If Data Cache is enabled
    cache_addr = (uint32_t *)((uint32_t)CAN_RAM_PTR[x].PTR_STD_ID & (~0x1FU));
    cache_size = (uint32_t)CAN_RAM_PTR[x].PTR_RX_BUFFER - (uint32_t)cache_addr;
    SCB_CleanDCache_by_Addr(cache_addr, cache_size);
  }
#endif

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_ObjectSetFilter (uint32_t obj_idx, ARM_CAN_FILTER_OPERATION operation, uint32_t id, uint32_t arg) { return CANx_ObjectSetFilter (obj_idx, operation, id, arg, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_ObjectSetFilter (uint32_t obj_idx, ARM_CAN_FILTER_OPERATION operation, uint32_t id, uint32_t arg) { return CANx_ObjectSetFilter (obj_idx, operation, id, arg, 1U); }
#endif

/**
  \fn          int32_t CANx_ObjectConfigure (uint32_t obj_idx, ARM_CAN_OBJ_CONFIG obj_cfg, uint8_t x)
  \brief       Configure object.
  \param[in]   obj_idx  Object index
  \param[in]   obj_cfg  Object configuration state
                 - ARM_CAN_OBJ_INACTIVE :       deactivate object
                 - ARM_CAN_OBJ_RX :             configure object for reception
                 - ARM_CAN_OBJ_TX :             configure object for transmission
                 - ARM_CAN_OBJ_RX_RTR_TX_DATA : configure object that on RTR reception automatically transmits Data Frame
                 - ARM_CAN_OBJ_TX_RTR_RX_DATA : configure object that transmits RTR and automatically receives Data Frame
  \param[in]   x        Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_ObjectConfigure (uint32_t obj_idx, ARM_CAN_OBJ_CONFIG obj_cfg, uint8_t x) {
  Mcan     *ptr_CAN;
  uint32_t  msk;

  if (x >= CAN_CTRL_NUM)                        { return ARM_DRIVER_ERROR;           }
  if (obj_idx >= CAN_DRV_CONFIG[x].TOT_OBJ_NUM) { return ARM_DRIVER_ERROR_PARAMETER; }
  if (can_driver_powered[x] == 0U)              { return ARM_DRIVER_ERROR;           }
  if (can_obj_cfg_access[x] != 0U)              { return ARM_DRIVER_ERROR_BUSY;      }
  can_obj_cfg_access[x] = 1U; 

  ptr_CAN = ptr_CANx[x];

  switch (obj_cfg) {
    case ARM_CAN_OBJ_INACTIVE:
      if (obj_idx >= CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) {        // If this is Tx object
        msk = 1U << ((obj_idx - CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) & 0x1FU);
        can_obj_tx_cfg[x]    &= ~msk;
        if        (obj_idx < CAN_DRV_CONFIG[x].TX_FIFO_OBJ_IDX_MAX) {
          // If disabling Tx FIFO
          ptr_CAN->MCAN_IE &= ~MCAN_IE_TEFNE;
        } else if (obj_idx < CAN_DRV_CONFIG[x].TX_BUF_OBJ_IDX_MAX) {
          // If disabling Tx Buffer
          ptr_CAN->MCAN_TXBTIE &= ~msk;
        }
      } else {                                          // If this is Rx object
        can_obj_rx_cfg[x][obj_idx >> 5] &= ~(1U << (obj_idx & 0x1FU));
        if        (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
          // If disabling Rx FIFO
          if (obj_idx == 0U) {
            ptr_CAN->MCAN_IE &= ~MCAN_IE_RF0NE;
          } else {
            ptr_CAN->MCAN_IE &= ~MCAN_IE_RF1NE;
          }
          ptr_CAN->MCAN_TXBTIE &= ~msk;
        } else if (obj_idx < CAN_DRV_CONFIG[x].RX_BUF_OBJ_IDX_MAX) {
          // If disabling Rx Buffer
        }
      }
      break;
    case ARM_CAN_OBJ_TX:
      if (obj_idx < CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) { can_obj_cfg_access[x] = 0U; return ARM_DRIVER_ERROR_PARAMETER; }
      msk = 1U << ((obj_idx - CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) & 0x1FU);
      can_obj_tx_cfg[x]    |= msk;
      if        (obj_idx < CAN_DRV_CONFIG[x].TX_FIFO_OBJ_IDX_MAX) {
        // If enabling Tx FIFO
        ptr_CAN->MCAN_IE |=  MCAN_IE_TEFNE;
      } else if (obj_idx < CAN_DRV_CONFIG[x].TX_BUF_OBJ_IDX_MAX) {
        // If enabling Tx Buffer
        ptr_CAN->MCAN_TXBTIE   |=  msk;
      }
      break;
    case ARM_CAN_OBJ_RX:
      if (obj_idx >= CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) { can_obj_cfg_access[x] = 0U; return ARM_DRIVER_ERROR_PARAMETER; }
      if        (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
        // If enabling Rx FIFO
        if (obj_idx == 0U) {
          ptr_CAN->MCAN_IE |=  MCAN_IE_RF0NE;
        } else {
          ptr_CAN->MCAN_IE |=  MCAN_IE_RF1NE;
        }
      }
      can_obj_rx_cfg[x][obj_idx >> 5] |= 1U << (obj_idx & 0x1FU);
      break;
    case ARM_CAN_OBJ_RX_RTR_TX_DATA:
    case ARM_CAN_OBJ_TX_RTR_RX_DATA:
    default:
      return ARM_DRIVER_ERROR;
  }
  can_obj_cfg_access[x] = 0U; 

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_ObjectConfigure (uint32_t obj_idx, ARM_CAN_OBJ_CONFIG obj_cfg) { return CANx_ObjectConfigure (obj_idx, obj_cfg, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_ObjectConfigure (uint32_t obj_idx, ARM_CAN_OBJ_CONFIG obj_cfg) { return CANx_ObjectConfigure (obj_idx, obj_cfg, 1U); }
#endif

/**
  \fn          int32_t CANx_MessageSend (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, const uint8_t *data, uint8_t size, uint8_t x)
  \brief       Send message on CAN bus.
  \param[in]   obj_idx  Object index
  \param[in]   msg_info Pointer to CAN message information
  \param[in]   data     Pointer to data buffer
  \param[in]   size     Number of data bytes to send
  \param[in]   x        Controller number (0..1)
  \return      value >= 0  number of data bytes accepted to send
  \return      value < 0   execution status
*/
static int32_t CANx_MessageSend (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, const uint8_t *data, uint8_t size, uint8_t x) {
  Mcan     *ptr_CAN;
  uint32_t *ptr_buffer_RAM;
  uint32_t *cache_addr;
  uint32_t  t0, i, cache_size, tx_obj_idx, buf_ofs_idx;

  if (x >= CAN_CTRL_NUM)                                                                          { return ARM_DRIVER_ERROR;           }
  if ((obj_idx < CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) || (obj_idx >= CAN_DRV_CONFIG[x].TOT_OBJ_NUM)) { return ARM_DRIVER_ERROR_PARAMETER; }
  if (can_driver_powered[x] == 0U)                                                                { return ARM_DRIVER_ERROR;           }
  tx_obj_idx = (obj_idx - CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) & 0x1FU;
  if ((can_obj_tx_cfg[x] & (1U << tx_obj_idx)) == 0U)                                             { return ARM_DRIVER_ERROR;           }

  ptr_CAN = ptr_CANx[x];

  if (obj_idx < CAN_DRV_CONFIG[x].TX_FIFO_OBJ_IDX_MAX) {
    // Send using Tx FIFO
    buf_ofs_idx = ((ptr_CAN->MCAN_TXFQS & MCAN_TXFQS_TFQPI_Msk) >> MCAN_TXFQS_TFQPI_Pos);
  } else {
    // Send using Tx Buffer
    buf_ofs_idx = tx_obj_idx;
  }
  ptr_buffer_RAM = CAN_RAM_PTR[x].PTR_TX_BUFFER + (buf_ofs_idx * 18U);

  if ((msg_info->id & ARM_CAN_ID_IDE_Msk) == 0U) {      // Standard ID
    t0 = (msg_info->id & 0x7FFU) << 18;
  } else {                                              // Extended ID
    t0 = (msg_info->id & 0x1FFFFFFFU) | (1U << 30);
  }

  if (msg_info->rtr != 0U) {                            // RTR message
    t0 |= 1U << 29;
    ptr_buffer_RAM[0] = t0;
  } else {                                              // Data message
    ptr_buffer_RAM[0] = t0;
    if (can_driver_fd_mode[x] != 0U) {                  // If FD mode
      ptr_buffer_RAM[1] = (1U << 21) | (1U << 20);      // Set FDF (FD Format) and BRS (Bit Rate Switching)
      if (size > 64U) { size = 64U; }
      for (i = 0U; i < 16U; i++) {
        if (CAN_DLC_ENCODE[i] >= size) {
          ptr_buffer_RAM[1] |= i << 16;
          break;
        }
      }
    } else {                                            // If not FD mode
      if (size > 8U) { size = 8U; }
      ptr_buffer_RAM[1] = size << 16;
    }
    if (obj_idx < CAN_DRV_CONFIG[x].TX_FIFO_OBJ_IDX_MAX) {
      // Send using Tx FIFO
      ptr_buffer_RAM[1] |= (1U << 23);
    }

    memcpy ((uint8_t *)(ptr_buffer_RAM + 2U), data, size);
  }

  can_tx_obj_active[x] |= 1U << tx_obj_idx;

#if (defined(__DCACHE_PRESENT) && (__DCACHE_PRESENT == 1U))
  if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {              // If Data Cache is enabled
    cache_addr = (uint32_t *)((uint32_t)ptr_buffer_RAM & (~0x1FU));
    cache_size = (uint32_t)ptr_buffer_RAM + 8U + size - (uint32_t)cache_addr;
    SCB_CleanDCache_by_Addr(cache_addr, cache_size);
  }
#endif

  ptr_CAN->MCAN_TXBAR = 1U << buf_ofs_idx;

  return ((int32_t)size);
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_MessageSend (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, const uint8_t *data, uint8_t size) { return CANx_MessageSend (obj_idx, msg_info, data, size, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_MessageSend (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, const uint8_t *data, uint8_t size) { return CANx_MessageSend (obj_idx, msg_info, data, size, 1U); }
#endif

/**
  \fn          int32_t CANx_MessageRead (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, uint8_t *data, uint8_t size, uint8_t x)
  \brief       Read message received on CAN bus.
  \param[in]   obj_idx  Object index
  \param[out]  msg_info Pointer to read CAN message information
  \param[out]  data     Pointer to data buffer for read data
  \param[in]   size     Maximum number of data bytes to read
  \param[in]   x        Controller number (0..1)
  \return      value >= 0  number of data bytes read
  \return      value < 0   execution status
*/
static int32_t CANx_MessageRead (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, uint8_t *data, uint8_t size, uint8_t x) {
  Mcan     *ptr_CAN;
  uint32_t *ptr_buffer_RAM;
  uint32_t *cache_addr;
  uint32_t  r, cache_size;

  if (x >= CAN_CTRL_NUM)                                                   { return ARM_DRIVER_ERROR;           }
  if (obj_idx >= CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM)                         { return ARM_DRIVER_ERROR_PARAMETER; }
  if (can_driver_powered[x] == 0U)                                         { return ARM_DRIVER_ERROR;           }
  if ((can_obj_rx_cfg[x][obj_idx >> 5] & (1U << (obj_idx & 0x1FU))) == 0U) { return ARM_DRIVER_ERROR;           }

  ptr_CAN = ptr_CANx[x];

  if (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
    // Read message received with Rx FIFO
    if (obj_idx == 0U) {
      ptr_buffer_RAM = CAN_RAM_PTR[x].PTR_RX_FIFO0 + (((ptr_CAN->MCAN_RXF0S & MCAN_RXF0S_F0GI_Msk) >> MCAN_RXF0S_F0GI_Pos) * 18U);
    } else {
      ptr_buffer_RAM = CAN_RAM_PTR[x].PTR_RX_FIFO1 + (((ptr_CAN->MCAN_RXF1S & MCAN_RXF1S_F1GI_Msk) >> MCAN_RXF1S_F1GI_Pos) * 18U);
    }
  } else {
    // Read message received with Rx Buffer
    ptr_buffer_RAM = CAN_RAM_PTR[x].PTR_RX_BUFFER + ((obj_idx - CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM) * 18U);
  }

#if (defined(__DCACHE_PRESENT) && (__DCACHE_PRESENT == 1U))
  if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {      // If Data Cache is enabled
    cache_addr = (uint32_t *)((uint32_t)ptr_buffer_RAM & (~0x1FU));
    cache_size = (uint32_t)ptr_buffer_RAM + size - (uint32_t)cache_addr;
    SCB_InvalidateDCache_by_Addr(cache_addr, cache_size);
  }
#endif

  r = ptr_buffer_RAM[0];                // Read R0
  msg_info->esi = (r >> 31);            // Get Error Status Indicator flag
  if ((r & (1U << 30)) != 0U) {         // If Extended ID
    msg_info->id = ((r      ) & 0x1FFFFFFFU) | ARM_CAN_ID_IDE_Msk;
  } else {                              // If Standard ID
    msg_info->id = ((r >> 18) & 0x7FFU);
  }
  if ((r & (1U << 29)) != 0U) {         // If RTR message
    msg_info->rtr = 1U;
  } else {                              // If Data message
    msg_info->rtr = 0U;
  }
  r = ptr_buffer_RAM[1];                // Read R1
  msg_info->edl = (r >> 21) &  1U;      // Get Extended Data Length flag
  msg_info->brs = (r >> 20) &  1U;      // Get Baud Rate Switching flag
  msg_info->dlc = (r >> 16) & 15U;      // Get Data Length Code
  if (msg_info->rtr == 0U) {            // If Data message read data bytes
    if ((msg_info->edl == 1U) && (can_driver_fd_mode[x] != 0U)) {       // If FD mode and Extended Data Length bit active
      if (size > CAN_DLC_ENCODE[msg_info->dlc]) {
        size = CAN_DLC_ENCODE[msg_info->dlc];
      }
    } else { 
      if (size > msg_info->dlc) {       // If not FD mode or not Extended Data Length bit active
        size = msg_info->dlc;
      }
      if (size > 8U) {
        size = 8U;
      }
    }
  } else {
    size = 0U;
  }

  if (size > 0U) {
    memcpy (data, (uint8_t *)(ptr_buffer_RAM + 2U), size);
  }

  if (obj_idx < CAN_DRV_CONFIG[x].RX_FIFO_OBJ_IDX_MAX) {
    // Acknowledge message received with Rx FIFO
    if (obj_idx == 0U) {
      ptr_CAN->MCAN_RXF0A = (ptr_CAN->MCAN_RXF0S & MCAN_RXF0S_F0GI_Msk) >> MCAN_RXF0S_F0GI_Pos;
    } else {
      ptr_CAN->MCAN_RXF1A = (ptr_CAN->MCAN_RXF1S & MCAN_RXF1S_F1GI_Msk) >> MCAN_RXF1S_F1GI_Pos;
    }
  } else {
    // Clear new data flag and release Rx buffer
    obj_idx -= CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM;
    if (obj_idx < 32U) {
      ptr_CAN->MCAN_NDAT1 = 1 <<  obj_idx;
    } else {
      ptr_CAN->MCAN_NDAT2 = 1 << (obj_idx - 32U);
    }
  }

  return ((int32_t)size);
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_MessageRead (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, uint8_t *data, uint8_t size) { return CANx_MessageRead (obj_idx, msg_info, data, size, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_MessageRead (uint32_t obj_idx, ARM_CAN_MSG_INFO *msg_info, uint8_t *data, uint8_t size) { return CANx_MessageRead (obj_idx, msg_info, data, size, 1U); }
#endif

/**
  \fn          int32_t CANx_Control (uint32_t control, uint32_t arg, uint8_t x)
  \brief       Control CAN interface.
  \param[in]   control  Operation
                 - ARM_CAN_SET_FD_MODE :            set FD operation mode
                 - ARM_CAN_ABORT_MESSAGE_SEND :     abort sending of CAN message
                 - ARM_CAN_CONTROL_RETRANSMISSION : enable/disable automatic retransmission
                 - ARM_CAN_SET_TRANSCEIVER_DELAY :  set transceiver delay
  \param[in]   arg      Argument of operation
  \param[in]   x        Controller number (0..1)
  \return      execution status
*/
static int32_t CANx_Control (uint32_t control, uint32_t arg, uint8_t x) {
  Mcan     *ptr_CAN;
  uint32_t  cccr, tdc_en, tdco;
#if (CHIP_REV_A == 0)                           // Chip revision higher than MRL A
  uint32_t  tdcf;
#endif

  if (x >= CAN_CTRL_NUM)           { return ARM_DRIVER_ERROR; }
  if (can_driver_powered[x] == 0U) { return ARM_DRIVER_ERROR; }

  ptr_CAN = ptr_CANx[x];

  switch (control & ARM_CAN_CONTROL_Msk) {
    case ARM_CAN_ABORT_MESSAGE_SEND:
      if ((arg < CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM) || (arg >= CAN_DRV_CONFIG[x].TOT_OBJ_NUM)) { return ARM_DRIVER_ERROR_PARAMETER; }
      if ( arg < CAN_DRV_CONFIG[x].TX_FIFO_OBJ_IDX_MAX)                                       { return ARM_DRIVER_ERROR;           }
      if ((can_obj_tx_cfg[x] & (1U << (arg - CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM))) == 0U)       { return ARM_DRIVER_OK;              }
      if (can_driver_powered[x] == 0U)                                                        { return ARM_DRIVER_ERROR;           }
      ptr_CAN->MCAN_TXBCR   =  (1U << (arg - CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM - CAN_DRV_CONFIG[x].TX_FIFO_OBJ_NUM));
      can_tx_obj_active[x] &= ~(1U << (arg - CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM));
      break;
    case ARM_CAN_CONTROL_RETRANSMISSION:
      switch (arg) {
        case 0:                                 // Disable automatic retransmission
          ptr_CAN->MCAN_CCCR |=  MCAN_CCCR_DAR;
          break;
        case 1:                                 // Enable automatic retransmission
          ptr_CAN->MCAN_CCCR &= ~MCAN_CCCR_DAR;
          break;
        default:
          return ARM_DRIVER_ERROR_PARAMETER;
      }
      break;
    case ARM_CAN_SET_FD_MODE:
      if (x == 0U) {
        tdc_en = CAN0_TRANSMITTER_DELAY_COMPENSATION_EN;
        tdco   = CAN0_TRANSMITTER_DELAY_OFFSET;
      } else {
        tdc_en = CAN1_TRANSMITTER_DELAY_COMPENSATION_EN;
        tdco   = CAN1_TRANSMITTER_DELAY_OFFSET;
      }
      cccr  = ptr_CAN->MCAN_CCCR;
      cccr |= MCAN_CCCR_INIT;
      ptr_CAN->MCAN_CCCR = cccr;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 0U);
      cccr |= MCAN_CCCR_CCE;
      ptr_CAN->MCAN_CCCR = cccr;
#if (CHIP_REV_A == 1)                           // Chip revision MRL A
      cccr &= ~(MCAN_CCCR_CME_Msk | MCAN_CCCR_CMR_Msk);
      switch (arg) {
        case 0:                                 // Disable FD mode
          cccr |= MCAN_CCCR_CMR(3U);
          if (tdc_en != 0U) {                   // If TDC is used
            ptr_CAN->MCAN_FBTP &= ~MCAN_FBTP_TDC;
          }
          break;
        case 1:                                 // Enable FD mode with bit rate switching
          cccr |= MCAN_CCCR_CME(2U) | MCAN_CCCR_CMR(2U);
          if (tdc_en != 0U) {                   // If TDC is used
            ptr_CAN->MCAN_FBTP |=  MCAN_FBTP_TDCO(tdco) | MCAN_FBTP_TDC;
          }
          break;
        default:
          return ARM_DRIVER_ERROR_PARAMETER;
      }
#else                                           // Chip revision MRL B
      if (x == 0U) {
        tdcf   = CAN0_TRANSMITTER_DELAY_FILTER;
      } else {
        tdcf   = CAN1_TRANSMITTER_DELAY_FILTER;
      }
      switch (arg) {
        case 0:                                 // Disable FD mode
          cccr &= ~(MCAN_CCCR_FDOE_Msk | MCAN_CCCR_BRSE_Msk | MCAN_CCCR_NISO_Msk);
          if (tdc_en != 0U) {                   // If TDC is used
            ptr_CAN->MCAN_DBTP &= ~MCAN_DBTP_TDC;
          }
          break;
        case 1:                                 // Enable FD mode with bit rate switching
          cccr |=   MCAN_CCCR_FDOE_Msk | MCAN_CCCR_BRSE_Msk;
#if (CAN_NON_ISO_EN == 1U)
          cccr |=   MCAN_CCCR_NISO_Msk;         // Enable non-ISO mode
#endif
          if (tdc_en != 0U) {                   // If TDC is used
            ptr_CAN->MCAN_TDCR  = ((tdco & 0x7FU) << 8) | (tdcf & 0x7FU);
            ptr_CAN->MCAN_DBTP |=  MCAN_DBTP_TDC;
          }
          break;
        default:
          return ARM_DRIVER_ERROR_PARAMETER;
      }
#endif
      ptr_CAN->MCAN_CCCR = cccr;
      cccr &= ~(MCAN_CCCR_INIT | MCAN_CCCR_CCE);
      ptr_CAN->MCAN_CCCR = cccr;
      while ((ptr_CAN->MCAN_CCCR & MCAN_CCCR_INIT) == 1U);
      if (arg == 1U) {
        can_driver_fd_mode[x] = 1U;
      } else {
        can_driver_fd_mode[x] = 0U;
      }
      break;
    case ARM_CAN_SET_TRANSCEIVER_DELAY:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static int32_t CAN0_Control (uint32_t control, uint32_t arg) { return CANx_Control (control, arg, 0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static int32_t CAN1_Control (uint32_t control, uint32_t arg) { return CANx_Control (control, arg, 1U); }
#endif

/**
  \fn          ARM_CAN_STATUS CANx_GetStatus (uint8_t x)
  \brief       Get CAN status.
  \param[in]   x      Controller number (0..1)
  \return      CAN status ARM_CAN_STATUS
*/
static ARM_CAN_STATUS CANx_GetStatus (uint8_t x) {
  Mcan            *ptr_CAN;
  ARM_CAN_STATUS   can_status;
  uint32_t         psr, ecr;

  if ((x >= CAN_CTRL_NUM) || (can_driver_powered[x] == 0U)) {
    memset((void *)&can_status, 0U, sizeof(ARM_CAN_STATUS));
    return can_status;
  }

  ptr_CAN = ptr_CANx[x];
  psr     = ptr_CAN->MCAN_PSR;
  ecr     = ptr_CAN->MCAN_ECR;

  if       ((psr & MCAN_PSR_BO) != 0U) { can_status.unit_state = ARM_CAN_UNIT_STATE_INACTIVE; }
  else if  ((psr & MCAN_PSR_EP) != 0U) { can_status.unit_state = ARM_CAN_UNIT_STATE_PASSIVE;  }
  else                                 { can_status.unit_state = ARM_CAN_UNIT_STATE_ACTIVE;   }

  can_status.last_error_code = CAN_LEC_ENCODE[psr & MCAN_PSR_LEC_Msk];
  can_status.tx_error_count  = (uint8_t)(ecr & MCAN_ECR_TEC_Msk);
  if ((ecr & MCAN_ECR_RP) != 0U) {
    can_status.rx_error_count = 128U;
  } else {
    can_status.rx_error_count = (uint8_t)((ecr & MCAN_ECR_REC_Msk) >> MCAN_ECR_REC_Pos);
  }

  return can_status;
}
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
static ARM_CAN_STATUS CAN0_GetStatus (void) { return CANx_GetStatus (0U); }
#endif
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
static ARM_CAN_STATUS CAN1_GetStatus (void) { return CANx_GetStatus (1U); }
#endif


/**
  \fn          void MCANx_Handler (void)
  \brief       CANx Interrupt Routine (IRQ).
*/
void MCANx_Handler (uint8_t x) {
  Mcan             *ptr_CAN;
  uint32_t          ir, psr, ndat1, ndat2, txbto, msk, obj_idx;
  volatile uint32_t get_idx, put_idx;

  ptr_CAN = ptr_CANx[x];

  ir      = ptr_CAN->MCAN_IR & ptr_CAN->MCAN_IE;
  ndat1   = ptr_CAN->MCAN_NDAT1;
  ndat2   = ptr_CAN->MCAN_NDAT2;
  txbto   = ptr_CAN->MCAN_TXBTO;
  ptr_CAN->MCAN_IR = ir;

  if (CAN_SignalUnitEvent[x] != NULL) {
    psr = ptr_CAN->MCAN_PSR;
    if (ir & (MCAN_IR_BO | MCAN_IR_EP)) {               // Handle error interrupt
      if      (((ir & MCAN_IR_BO) != 0U) && ((psr & MCAN_PSR_BO) != 0U)) { CAN_SignalUnitEvent[x](ARM_CAN_EVENT_UNIT_BUS_OFF); }
      else if (((ir & MCAN_IR_EP) != 0U) && ((psr & MCAN_PSR_EP) != 0U)) { CAN_SignalUnitEvent[x](ARM_CAN_EVENT_UNIT_PASSIVE); }
      else                                                               { CAN_SignalUnitEvent[x](ARM_CAN_EVENT_UNIT_ACTIVE);  }
    }
    if (((ir & MCAN_IR_EW) != 0U) && ((psr & MCAN_PSR_EW) != 0U))        { CAN_SignalUnitEvent[x](ARM_CAN_EVENT_UNIT_WARNING); }
  }

  if (CAN_SignalObjectEvent[x] != NULL) {
                                                        // Handle transmit events
    if (ir & MCAN_IR_TEFN) {                            // If message was transmitted using Tx FIFO
      get_idx = (ptr_CAN->MCAN_TXEFS & MCAN_TXEFS_EFGI_Msk) >> MCAN_TXEFS_EFGI_Pos;
      put_idx = (ptr_CAN->MCAN_TXEFS & MCAN_TXEFS_EFPI_Msk) >> MCAN_TXEFS_EFPI_Pos;
      do {
        ptr_CAN->MCAN_TXEFA = get_idx;
        CAN_SignalObjectEvent[x](CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM, ARM_CAN_EVENT_SEND_COMPLETE);
        get_idx = (ptr_CAN->MCAN_TXEFS & MCAN_TXEFS_EFGI_Msk) >> MCAN_TXEFS_EFGI_Pos;
        put_idx = (ptr_CAN->MCAN_TXEFS & MCAN_TXEFS_EFPI_Msk) >> MCAN_TXEFS_EFPI_Pos;
      } while (get_idx != put_idx);
    }
    if (ir & MCAN_IR_TC) {                              // If message was transmitted using dedicated Tx buffer
      for (obj_idx = 0U, msk = 1U; (txbto != 0U) || (msk != 0U); obj_idx++, msk <<= 1U) {
        if ((txbto & msk) != 0U) {
          txbto &= ~msk;
          CAN_SignalObjectEvent[x](obj_idx + CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM, ARM_CAN_EVENT_SEND_COMPLETE);
        }
      }
    }

                                                        // Handle receive events
    if (ir & MCAN_IR_RF0N) {                            // If message was received to Rx FIFO0
      CAN_SignalObjectEvent[x](0U, ARM_CAN_EVENT_RECEIVE);
    }
    if (ir & MCAN_IR_RF1N) {                            // If message was received to Rx FIFO1
      CAN_SignalObjectEvent[x](1U, ARM_CAN_EVENT_RECEIVE);
    }
    if (ir & MCAN_IR_DRX) {                             // If message was received to dedicated Rx buffer
      if ((ndat1 | ndat2) != 0U) {
        for (obj_idx = 0U, msk = 1U; (ndat1 != 0U) && (msk != 0U); obj_idx++, msk <<= 1U) {
          if ((ndat1 & msk) != 0U) {
            ndat1 &= ~msk;
            CAN_SignalObjectEvent[x](obj_idx + CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM, ARM_CAN_EVENT_RECEIVE);
          }
        }
        if (CAN_DRV_CONFIG[x].RX_TOT_OBJ_NUM > 32U) {
          for (obj_idx = 32U, msk = 1U; (ndat2 != 0U) && (msk != 0U); obj_idx++, msk <<= 1U) {
            if ((ndat2 & msk) != 0U) {
              ndat2 &= ~msk;
              CAN_SignalObjectEvent[x](obj_idx + CAN_DRV_CONFIG[x].RX_FIFO_OBJ_NUM, ARM_CAN_EVENT_RECEIVE);
            }
          }
        }
      }
    }
  }
}

/**
  \fn          void MCAN0_Handler (void)
  \brief       CAN0 Interrupt Routine (IRQ).
*/
#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
void MCAN0_Handler (void) { MCANx_Handler (0); }
#endif

/**
  \fn          void MCAN1_Handler (void)
  \brief       CAN1 Interrupt Routine (IRQ).
*/
#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
void MCAN1_Handler (void) { MCANx_Handler (1); }
#endif


#if (defined(RTE_CAN0) && (RTE_CAN0 == 1U))
ARM_DRIVER_CAN Driver_CAN0 = {
  CAN_GetVersion,
  CAN0_GetCapabilities,
  CAN0_Initialize,
  CAN0_Uninitialize,
  CAN0_PowerControl,
  CAN_GetClock,
  CAN0_SetBitrate,
  CAN0_SetMode,
  CAN0_ObjectGetCapabilities,
  CAN0_ObjectSetFilter,
  CAN0_ObjectConfigure,
  CAN0_MessageSend,
  CAN0_MessageRead,
  CAN0_Control,
  CAN0_GetStatus
};
#endif

#if (defined(RTE_CAN1) && (RTE_CAN1 == 1U))
ARM_DRIVER_CAN Driver_CAN1 = {
  CAN_GetVersion,
  CAN1_GetCapabilities,
  CAN1_Initialize,
  CAN1_Uninitialize,
  CAN1_PowerControl,
  CAN_GetClock,
  CAN1_SetBitrate,
  CAN1_SetMode,
  CAN1_ObjectGetCapabilities,
  CAN1_ObjectSetFilter,
  CAN1_ObjectConfigure,
  CAN1_MessageSend,
  CAN1_MessageRead,
  CAN1_Control,
  CAN1_GetStatus
};
#endif
