/* -----------------------------------------------------------------------------
 * Copyright (c) 2015 ARM Ltd.
 *  
 * This software is provided 'as-is', without any express or implied warranty. 
 * In no event will the authors be held liable for any damages arising from 
 * the use of this software. Permission is granted to anyone to use this 
 * software for any purpose, including commercial applications, and to alter 
 * it and redistribute it freely, subject to the following restrictions:
 *  
 * 1. The origin of this software must not be misrepresented; you must not 
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be 
 *    appreciated but is not required. 
 * 
 * 2. Altered source versions must be plainly marked as such, and must not be 
 *    misrepresented as being the original software. 
 * 
 * 3. This notice may not be removed or altered from any source distribution.
 *
 * $Date:        28. April 2015
 * $Revision:    V1.02
 *
 * Project:      CMSIS Ethernet MAC Driver for Atmel SAM-ESV7
 * -------------------------------------------------------------------------- */

#include "GMAC_SAM-ESV7.h"

#define ARM_ETH_MAC_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,02) /* driver version */

/* Receive/transmit Checksum offload enable */
#ifndef EMAC_CHECKSUM_OFFLOAD
#define EMAC_CHECKSUM_OFFLOAD   1
#endif

/* ETH Memory Buffer configuration */
#define RX_BUFFERS           16        /* Must be a power of 2 */
#define TX_BUFFERS           16        /* Must be a power of 2 */
#define ETH_BUF_SIZE         1536U     /* ETH Receive/Transmit buffer size */
#define DUMMY_SIZE           2
#define DUMMY_BUFF_SIZE      128

/* PIN list of PIO for GMAC RMII */
#define BOARD_GMAC_PINS_RMII {(PIO_PD0  | \
                               PIO_PD1  | \
                               PIO_PD2  | \
                               PIO_PD3  | \
                               PIO_PD4  | \
                               PIO_PD5  | \
                               PIO_PD6  | \
                               PIO_PD7  | \
                               PIO_PD8  | \
                               PIO_PD9  ),\
                               PIOD, ID_PIOD, PIO_PERIPH_A, PIO_DEFAULT}
                          
/* PIN list of PIO for GMAC MII */
#define BOARD_GMAC_PINS_MII  {(PIO_PD0  | \
                               PIO_PD1  | \
                               PIO_PD2  | \
                               PIO_PD3  | \
                               PIO_PD15 | \
                               PIO_PD16 | \
                               PIO_PD17 | \
                               PIO_PD14 | \
                               PIO_PD4  | \
                               PIO_PD5  | \
                               PIO_PD6  | \
                               PIO_PD11 | \
                               PIO_PD12 | \
                               PIO_PD7  | \
                               PIO_PD10 | \
                               PIO_PD13 | \
                               PIO_PD8  | \
                               PIO_PD9  ),\
                               PIOD, ID_PIOD, PIO_PERIPH_A, PIO_DEFAULT}
                             
/* PIN list of PIO for GMAC */
#define BOARD_GMAC_RESET_PIN {PIO_PC10, PIOC, ID_PIOC, PIO_OUTPUT_1, PIO_PULLUP} 

/* PINs for GMAC */
#if (RTE_ETH_RMII==1)
static const Pin sGmacPins[] = {BOARD_GMAC_PINS_RMII};
#else
static const Pin sGmacPins[] = {BOARD_GMAC_PINS_MII};
#endif
static const Pin sGmacResetPin[] = BOARD_GMAC_RESET_PIN;

/* Driver Version */
static const ARM_DRIVER_VERSION DriverVersion = {
  ARM_ETH_MAC_API_VERSION,
  ARM_ETH_MAC_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_ETH_MAC_CAPABILITIES DriverCapabilities = {
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_rx_ip4  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_rx_ip6  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_rx_udp  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_rx_tcp  */
  0U,                                       /* checksum_offload_rx_icmp */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_tx_ip4  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_tx_ip6  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_tx_udp  */
  (EMAC_CHECKSUM_OFFLOAD != 0) ? 1U : 0U,   /* checksum_offload_tx_tcp  */
  0U,                                       /* checksum_offload_tx_icmp */
  (RTE_ETH_RMII == 0) ?
  ARM_ETH_INTERFACE_MII :
  ARM_ETH_INTERFACE_RMII,                   /* media_interface          */
  0U,                                       /* mac_address              */
  1U,                                       /* event_rx_frame           */
  0U,                                       /* event_tx_frame           */
  0U,                                       /* event_wakeup             */
  0U                                        /* precision_timer          */
};

/* The GMAC descriptors must be located in a section without data cache.  
   Define RAM_NOCACHE section in the linker scatter file
   and setup the Memory Protection Unit accordingly */
extern unsigned int Image$$RAM_NOCACHE$$Length;

/* Local variables */
static ARM_ETH_MAC_SignalEvent_t cb_Event;

/* Ethernet DMA Descriptor and Buffer memory */
/* Ensure it is in a non-cacheable RAM region   */
__attribute__ ((at(RTE_ETH_DMA_MEM_ADDR))) COMPILER_ALIGNED(8) static sGmacTxDescriptor tx_desc[TX_BUFFERS];
__attribute__ ((at(RTE_ETH_DMA_MEM_ADDR))) COMPILER_ALIGNED(8) static sGmacRxDescriptor rx_desc[RX_BUFFERS];
__attribute__ ((at(RTE_ETH_DMA_MEM_ADDR))) COMPILER_ALIGNED(8) static sGmacTxDescriptor gDummyTxDs[DUMMY_SIZE];
__attribute__ ((at(RTE_ETH_DMA_MEM_ADDR))) COMPILER_ALIGNED(8) static sGmacRxDescriptor gDummyRxDs[DUMMY_SIZE];

COMPILER_ALIGNED(8) static fGmacdTransferCallback gTxCbs[TX_BUFFERS];
COMPILER_ALIGNED(8) static fGmacdTransferCallback gDummyTxCbs[DUMMY_SIZE];
COMPILER_ALIGNED(8) static uint8_t rx_buf [RX_BUFFERS*ETH_BUF_SIZE];
COMPILER_ALIGNED(8) static uint8_t tx_buf [TX_BUFFERS*ETH_BUF_SIZE];
COMPILER_ALIGNED(8) static uint8_t pTxDummyBuffer[DUMMY_SIZE*DUMMY_BUFF_SIZE];
COMPILER_ALIGNED(8) static uint8_t pRxDummyBuffer[DUMMY_SIZE*DUMMY_BUFF_SIZE];

/* GMAC driver instance */
static sGmacd gGmacd;

/* GMAC driver structure Queue*/
static sGmacInit Que0, Que;

// ETH_MAC State
static uint8_t eth_state = 0;


/* GMAC interrupt handler */
void GMAC_Handler(void)
{
  GMACD_Handler(&gGmacd, GMAC_QUE_0);
}

/* GMAC RX callback */
void GMAC_RxCb(uint32_t status) {
  if ((status & GMAC_RSR_REC) != 0) {
    cb_Event (ARM_ETH_MAC_EVENT_RX_FRAME);
  }
}

/**
  \fn          void init_queue (void)
  \brief       Initialize GMAC Queues.
  \return      none.
*/
static void init_queue (void) {
  
  /* Initialize GMAC driver structure Queue 0*/
  memset(&Que0, 0, sizeof(Que0));
  Que0.bIsGem = 1;
  Que0.bDmaBurstLength = 4;
  Que0.pRxBuffer = rx_buf;
  Que0.pRxD = rx_desc;
  Que0.wRxBufferSize = ETH_BUF_SIZE;
  Que0.wRxSize = RX_BUFFERS;
  Que0.pTxBuffer = tx_buf;
  Que0.pTxD = tx_desc;
  Que0.wTxBufferSize = ETH_BUF_SIZE;
  Que0.wTxSize = TX_BUFFERS;
  Que0.pTxCb = gTxCbs;     

  /* Initialize GMAC driver structure of other Queue */
  memset(&Que, 0, sizeof(Que));
  Que.bIsGem = 1;
  Que.bDmaBurstLength = 4;
  Que.pRxBuffer = pRxDummyBuffer;
  Que.pRxD = gDummyRxDs;
  Que.wRxBufferSize = DUMMY_BUFF_SIZE;
  Que.wRxSize = DUMMY_SIZE;
  Que.pTxBuffer = pTxDummyBuffer;
  Que.pTxD = gDummyTxDs;
  Que.wTxBufferSize = DUMMY_BUFF_SIZE;
  Que.wTxSize = DUMMY_SIZE;
  Que.pTxCb = gDummyTxCbs;  

#if (NUM_GMAC_QUEUES > 3)
  GMACD_InitTransfer(&gGmacd, &Que, GMAC_QUE_5);
  GMACD_InitTransfer(&gGmacd, &Que, GMAC_QUE_4);
  GMACD_InitTransfer(&gGmacd, &Que, GMAC_QUE_3);
#endif  
  GMACD_InitTransfer(&gGmacd, &Que, GMAC_QUE_2);
  GMACD_InitTransfer(&gGmacd, &Que, GMAC_QUE_1);
  GMACD_InitTransfer(&gGmacd, &Que0, GMAC_QUE_0);
  GMACD_SetRxCallback(&gGmacd, GMAC_RxCb, GMAC_QUE_0);

}

/* Ethernet Driver functions */

/**
  \fn          ARM_DRIVER_VERSION ETH_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION ETH_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_ETH_MAC_CAPABILITIES ETH_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_ETH_MAC_CAPABILITIES
*/
static ARM_ETH_MAC_CAPABILITIES ETH_GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn          int32_t ETH_Initialize (ARM_ETH_MAC_SignalEvent_t cb_event)
  \brief       Initialize Ethernet MAC Device.
  \param[in]   cb_event  Pointer to \ref ARM_ETH_MAC_SignalEvent
  \return      \ref execution_status
*/
static int32_t ETH_Initialize (ARM_ETH_MAC_SignalEvent_t cb_event) {
  
  if (eth_state & ETH_MAC_INITIALIZED) { return ARM_DRIVER_OK; }
  if (eth_state & ETH_MAC_POWERED)     { return ARM_DRIVER_ERROR; }
  
  /* Configure IO */
  PIO_Configure(sGmacResetPin, 1);
  PIO_Clear(sGmacResetPin);
  PIO_Configure(sGmacPins, PIO_LISTSIZE(sGmacPins));
  
  /* Register driver callback function */
  cb_Event = cb_event;
  
  eth_state = ETH_MAC_INITIALIZED;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_Uninitialize (void)
  \brief       De-initialize Ethernet MAC Device.
  \return      \ref execution_status
*/
static int32_t ETH_Uninitialize (void) {
  eth_state = 0;
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_PowerControl (ARM_POWER_STATE state)
  \brief       Control Ethernet MAC Device Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t ETH_PowerControl (ARM_POWER_STATE state) {
uint32_t val;
  
   switch (state) {
    case ARM_POWER_OFF:
      /* Disable interrupts */  
      NVIC_ClearPendingIRQ(GMAC_IRQn);
      NVIC_DisableIRQ(GMAC_IRQn); 
      /* Power Off */
      PMC_DisablePeripheral(ID_GMAC);
      PIO_Clear(sGmacResetPin);

      eth_state &= ~ETH_MAC_POWERED;
      break;

    case ARM_POWER_LOW:
      return ARM_DRIVER_ERROR_UNSUPPORTED;

    case ARM_POWER_FULL:
      if (eth_state & ETH_MAC_POWERED) { return ARM_DRIVER_OK; }
      if ((eth_state & ETH_MAC_INITIALIZED) == 0U) { return ARM_DRIVER_ERROR; }
                 
      /* Power On */
      PIO_Set(sGmacResetPin);      
      /* Initialize peripheral */
      GMACD_Init(&gGmacd, GMAC, ID_GMAC, 0, 0);  
      
      /* Set peripheral clock */
      val = SystemCoreClock;
      switch (PMC->PMC_MCKR & PMC_MCKR_MDIV_Msk) {
        case PMC_MCKR_MDIV_PCK_DIV2: val /= 2; break;
        case PMC_MCKR_MDIV_PCK_DIV4: val /= 4; break;
        case PMC_MCKR_MDIV_PCK_DIV3: val /= 3; break;
      }
      GMAC_SetMdcClock(GMAC,val);
      init_queue(); 

      GMAC_EnableMdio(GMAC); 
      /* Enable interrupts */  
      NVIC_ClearPendingIRQ(GMAC_IRQn);
      NVIC_EnableIRQ(GMAC_IRQn);   

      eth_state |= ETH_MAC_POWERED;
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_GetMacAddress (ARM_ETH_MAC_ADDR *ptr_addr)
  \brief       Get Ethernet MAC Address.
  \param[in]   ptr_addr  Pointer to address
  \return      \ref execution_status
*/
static int32_t ETH_GetMacAddress (ARM_ETH_MAC_ADDR *ptr_addr) {
  uint32_t val;

  if (ptr_addr == NULL) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }

  val = GMAC->GMAC_SA[0].GMAC_SAT;
  ptr_addr->b[5] = (uint8_t)(val >>  8);
  ptr_addr->b[4] = (uint8_t)(val);
  val = GMAC->GMAC_SA[0].GMAC_SAB;
  ptr_addr->b[3] = (uint8_t)(val >> 24);
  ptr_addr->b[2] = (uint8_t)(val >> 16);
  ptr_addr->b[1] = (uint8_t)(val >>  8);
  ptr_addr->b[0] = (uint8_t)(val);
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_SetMacAddress (const ARM_ETH_MAC_ADDR *ptr_addr)
  \brief       Set Ethernet MAC Address.
  \param[in]   ptr_addr  Pointer to address
  \return      \ref execution_status
*/
static int32_t ETH_SetMacAddress (const ARM_ETH_MAC_ADDR *ptr_addr) {
  if (ptr_addr == NULL) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }
  
  GMAC->GMAC_SA[0].GMAC_SAB = (ptr_addr->b[3] << 24) | (ptr_addr->b[2] << 16) | 
                              (ptr_addr->b[1] <<  8) | (ptr_addr->b[0]);
  GMAC->GMAC_SA[0].GMAC_SAT = (ptr_addr->b[5] <<  8) | (ptr_addr->b[4]);
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_SetAddressFilter (const ARM_ETH_MAC_ADDR *ptr_addr,
                                               uint32_t          num_addr)
  \brief       Configure Address Filter.
  \param[in]   ptr_addr  Pointer to addresses
  \param[in]   num_addr  Number of addresses to configure
  \return      \ref execution_status
*/
static int32_t ETH_SetAddressFilter (const ARM_ETH_MAC_ADDR *ptr_addr, uint32_t num_addr) {
uint8_t index;
uint8_t i,j,m,n;
uint32_t hashtable[2];
  
  if ((ptr_addr == NULL) && (num_addr != 0)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }
  
  // set specific address filter
  for (index=1; index<GMACSA_NUMBER; index++) {
    if (num_addr>0) {
      GMAC->GMAC_SA[index].GMAC_SAB = (ptr_addr->b[3] << 24) | (ptr_addr->b[2] << 16) | 
                                      (ptr_addr->b[1] <<  8) | (ptr_addr->b[0]);
      GMAC->GMAC_SA[index].GMAC_SAT = (ptr_addr->b[5] <<  8) | (ptr_addr->b[4]);
      num_addr--;
      ptr_addr++;
    } else {
      GMAC->GMAC_SA[index].GMAC_SAT = 0U; 
      GMAC->GMAC_SA[index].GMAC_SAB = 0U;
    }
  }
  
  // clear the hash table
  hashtable[0] = 0;
  hashtable[1] = 0;
  
  // calculate hash table for remaining MAC addresses
  for(i = 0; i < num_addr; i++) {
    // reset hash value
    index = 0; 
    // apply the hash function
    for(j = 0; j < 48; j += 6) {
      // calculate the shift count
      n = j / 8;
      m = j % 8; 
      // update hash value
      if(!m) index ^= ptr_addr->b[n];
      else index ^= (ptr_addr->b[n] >> m) | (ptr_addr->b[n + 1] << (8 - m));
    } 
    // the hash value is reduced to a 6-bit index
    index &= 0x3F;
    // update hash table contents
    hashtable[index / 32] |= (1 << (index % 32));    
  }

  // write the hash table
  GMAC->GMAC_HRB = hashtable[0];
  GMAC->GMAC_HRT = hashtable[1];
  
  // enable multicast and unicast hash
  GMAC->GMAC_NCFGR |= GMAC_NCFGR_MTIHEN | GMAC_NCFGR_UNIHEN;
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_SendFrame (const uint8_t *frame, uint32_t len, uint32_t flags)
  \brief       Send Ethernet frame.
  \param[in]   frame  Pointer to frame buffer with data to send
  \param[in]   len    Frame buffer length in bytes
  \param[in]   flags  Frame transmit flags (see ARM_ETH_MAC_TX_FRAME_...)
  \return      \ref execution_status
*/
static int32_t ETH_SendFrame (const uint8_t *frame, uint32_t len, uint32_t flags) {

  if (GMACD_Send(&gGmacd, (void*)frame, len, NULL, GMAC_QUE_0) != GMACD_OK)
    return ARM_DRIVER_ERROR;
  else  
    return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_ReadFrame (uint8_t *frame, uint32_t len)
  \brief       Read data of received Ethernet frame.
  \param[in]   frame  Pointer to frame buffer for data to read into
  \param[in]   len    Frame buffer length in bytes
  \return      number of data bytes read or execution status
                 - value >= 0: number of data bytes read
                 - value < 0: error occurred, value is execution status as defined with \ref execution_status 
*/
static int32_t ETH_ReadFrame (uint8_t *frame, uint32_t len) {
uint32_t frmSize;
  
  if(GMACD_Poll(&gGmacd, frame, len, &frmSize, GMAC_QUE_0) != GMACD_OK)
    return ARM_DRIVER_ERROR_BUSY;
  else 
    return ((int32_t)frmSize);
}

/**
  \fn          uint32_t GetRxFrameSize (void)
  \brief       Get size of received Ethernet frame.
  \return      number of bytes in received frame
*/
static uint32_t ETH_GetRxFrameSize (void) {
  volatile sGmacRxDescriptor *pRxTd = &gGmacd.queueList[GMAC_QUE_0].pRxD[gGmacd.queueList[GMAC_QUE_0].wRxI]; 
  if ((pRxTd->addr.val & GMAC_RX_OWNERSHIP_BIT) != GMAC_RX_OWNERSHIP_BIT) return 0;
  if ((pRxTd->status.val & GMAC_RX_EOF_BIT) != GMAC_RX_EOF_BIT) return 0;
  return (pRxTd->status.val & GMAC_LENGTH_FRAME);
}


/**
  \fn          int32_t ETH_GetRxFrameTime (ARM_ETH_MAC_TIME *time)
  \brief       Get time of received Ethernet frame.
  \param[in]   time  Pointer to time structure for data to read into
  \return      \ref execution_status
*/
static int32_t ETH_GetRxFrameTime (ARM_ETH_MAC_TIME *time) {
  //TODO: implement support for precision timer protocol
  return ARM_DRIVER_ERROR;
}

/**
  \fn          int32_t ETH_GetTxFrameTime (ARM_ETH_MAC_TIME *time)
  \brief       Get time of transmitted Ethernet frame.
  \param[in]   time  Pointer to time structure for data to read into
  \return      \ref execution_status
*/
static int32_t ETH_GetTxFrameTime (ARM_ETH_MAC_TIME *time) {
  //TODO: implement support for precision timer protocol
  return ARM_DRIVER_ERROR;
}

/**
  \fn          int32_t ETH_Control (uint32_t control, uint32_t arg)
  \brief       Control Ethernet Interface.
  \param[in]   control  operation
  \param[in]   arg      argument of operation (optional)
  \return      \ref execution_status
*/
static int32_t ETH_Control (uint32_t control, uint32_t arg) {
uint32_t duplex, speed;
  
  if ((eth_state & ETH_MAC_POWERED) == 0U) { return ARM_DRIVER_ERROR; }
    
  switch (control) {
    case ARM_ETH_MAC_CONFIGURE:

      /* Configure 1GBit/100MBit/10MBit mode */
      switch (arg & ARM_ETH_MAC_SPEED_Msk) {
        case ARM_ETH_MAC_SPEED_10M:
          speed = GMAC_SPEED_10M;
          break;
        case ARM_ETH_MAC_SPEED_100M:
          speed = GMAC_SPEED_100M;
          break;
        case ARM_ETH_MAC_SPEED_1G:
          speed = GMAC_SPEED_1000M;
          break;        
        default:
          return ARM_DRIVER_ERROR_UNSUPPORTED;
      }

      /* Confige Half/Full duplex mode */
      switch (arg & ARM_ETH_MAC_DUPLEX_Msk) {
        case ARM_ETH_MAC_DUPLEX_FULL:
          duplex = GMAC_DUPLEX_FULL;
          break;
        case ARM_ETH_MAC_DUPLEX_HALF:
          duplex = GMAC_DUPLEX_HALF;
          break;
        default:
          return ARM_DRIVER_ERROR;
      }      
                  
      /* Configure loopback mode */
      if (arg & ARM_ETH_MAC_LOOPBACK) {
        GMAC->GMAC_NCR |= GMAC_NCR_LBL;
      }

#if (EMAC_CHECKSUM_OFFLOAD != 0)
      /* Enable rx checksum verification */
      if (arg & ARM_ETH_MAC_CHECKSUM_OFFLOAD_RX) {
        GMAC->GMAC_DCFGR |= GMAC_NCFGR_RXCOEN;
      }
      /* Enable tx checksum generation */
      if (arg & ARM_ETH_MAC_CHECKSUM_OFFLOAD_TX) {
        GMAC->GMAC_DCFGR |= GMAC_DCFGR_TXCOEN;
      }
#else
      if ((arg & ARM_ETH_MAC_CHECKSUM_OFFLOAD_RX) ||
          (arg & ARM_ETH_MAC_CHECKSUM_OFFLOAD_TX)) {
        /* Checksum offload is disabled in the driver */
        return ARM_DRIVER_ERROR;
      }
#endif

      /* Enable broadcast frame receive */
      if ((arg & ARM_ETH_MAC_ADDRESS_BROADCAST) == 0) {
        GMAC->GMAC_NCFGR |= GMAC_NCFGR_NBC;
      }

      /* Enable all multicast frame receive */
      if (arg & ARM_ETH_MAC_ADDRESS_MULTICAST) {
        GMAC->GMAC_HRB = GMAC_HRB_ADDR(0xFFFFFFFF);
        GMAC->GMAC_HRT = GMAC_HRT_ADDR(0xFFFFFFFF);
        GMAC->GMAC_NCFGR |= GMAC_NCFGR_MTIHEN;
      }

      /* Enable promiscuous mode (no filtering) */
      if (arg & ARM_ETH_MAC_ADDRESS_ALL) {
        GMAC->GMAC_NCFGR |= GMAC_NCFGR_CAF;
      }
      
      /* Setup GMAC mode  */
#if (RTE_ETH_RMII==1)
      GMAC_EnableRGMII(GMAC, duplex, speed);
#else
      GMAC_EnableGMII(GMAC, duplex, speed);
#endif
      eth_state |= ETH_MAC_CONFIGURED;
      break;

    case ARM_ETH_MAC_CONTROL_TX:
      /* Enable/disable MAC transmitter */
      GMAC_TransmitEnable(GMAC,arg);
      break;

    case ARM_ETH_MAC_CONTROL_RX:
      /* Enable/disable MAC receiver */
      GMAC_ReceiveEnable(GMAC,arg);
      break;

    case ARM_ETH_MAC_FLUSH:
      /* Flush tx and rx buffers */
      GMACD_Reset(&gGmacd);
      break;

    case ARM_ETH_MAC_VLAN_FILTER:
      /* Configure VLAN filter */
      GMAC->GMAC_SVLAN = GMAC_SVLAN_ESVLAN | GMAC_SVLAN_VLAN_TYPE(arg);
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_ControlTimer (uint32_t control, ARM_ETH_MAC_TIME *time)
  \brief       Control Precision Timer.
  \param[in]   control  operation
  \param[in]   time     Pointer to time structure
  \return      \ref execution_status
*/
static int32_t ETH_ControlTimer (uint32_t control, ARM_ETH_MAC_TIME *time) {
  //TODO: implement support for precision timer protocol
  return ARM_DRIVER_ERROR;
}

/**
  \fn          int32_t ETH_PHY_Read (uint8_t phy_addr, uint8_t reg_addr, uint16_t *data)
  \brief       Read Ethernet PHY Register through Management Interface.
  \param[in]   phy_addr  5-bit device address
  \param[in]   reg_addr  5-bit register address
  \param[out]  data      Pointer where the result is written to
  \return      \ref execution_status
*/
static int32_t ETH_PHY_Read (uint8_t phy_addr, uint8_t reg_addr, uint16_t *data) {
uint32_t loop;
  
  GMAC->GMAC_MAN =   (~GMAC_MAN_WZO & GMAC_MAN_CLTTO) 
                      | GMAC_MAN_OP(0x02)
                      | GMAC_MAN_WTN(0x02)
                      | GMAC_MAN_PHYA(phy_addr)
                      | GMAC_MAN_REGA(reg_addr) ;
 
  /* Wait until IDLE bit in Network Status register is cleared */
  loop = SystemCoreClock;
  while((GMAC->GMAC_NSR & GMAC_NSR_IDLE) == 0) {
    /* Loop counter timeout */
    if (loop-- == 0) { return ARM_DRIVER_ERROR; }
  }
  
  *data = (uint16_t)(GMAC->GMAC_MAN & GMAC_MAN_DATA_Msk);
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ETH_PHY_Write (uint8_t phy_addr, uint8_t reg_addr, uint16_t data)
  \brief       Write Ethernet PHY Register through Management Interface.
  \param[in]   phy_addr  5-bit device address
  \param[in]   reg_addr  5-bit register address
  \param[in]   data      16-bit data to write
  \return      \ref execution_status
*/
static int32_t ETH_PHY_Write (uint8_t phy_addr, uint8_t reg_addr, uint16_t data) {
uint32_t loop;
  
  GMAC->GMAC_MAN =   (~GMAC_MAN_WZO & GMAC_MAN_CLTTO) 
                      | GMAC_MAN_OP(0x01)
                      | GMAC_MAN_WTN(0x02)
                      | GMAC_MAN_PHYA(phy_addr)
                      | GMAC_MAN_REGA(reg_addr)
                      | GMAC_MAN_DATA(data) ;

  /* Wait until IDLE bit in Network Status register is cleared */
  loop = SystemCoreClock;
  while((GMAC->GMAC_NSR & GMAC_NSR_IDLE) == 0) {
    /* Loop counter timeout */
    if (loop-- == 0) { return ARM_DRIVER_ERROR; }
  }
  
  return ARM_DRIVER_OK;
}


/* MAC Driver Control Block */
ARM_DRIVER_ETH_MAC Driver_ETH_MAC0 = {
  ETH_GetVersion,
  ETH_GetCapabilities,
  ETH_Initialize,
  ETH_Uninitialize,
  ETH_PowerControl,
  ETH_GetMacAddress,
  ETH_SetMacAddress,
  ETH_SetAddressFilter,
  ETH_SendFrame,
  ETH_ReadFrame,
  ETH_GetRxFrameSize,
  ETH_GetRxFrameTime,
  ETH_GetTxFrameTime,
  ETH_ControlTimer,
  ETH_Control,
  ETH_PHY_Read,
  ETH_PHY_Write
};
