/* -----------------------------------------------------------------------------
 * Copyright (c) 2015 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty. 
 * In no event will the authors be held liable for any damages arising from 
 * the use of this software. Permission is granted to anyone to use this 
 * software for any purpose, including commercial applications, and to alter 
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not 
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be 
 *    appreciated but is not required. 
 * 
 * 2. Altered source versions must be plainly marked as such, and must not be 
 *    misrepresented as being the original software. 
 * 
 * 3. This notice may not be removed or altered from any source distribution.
 *   
 * $Date:        23. April 2015
 * $Revision:    V1.01
 *  
 * Project:      CMSIS MCI Driver for Atmel SAM-ESV7
 * ---------------------------------------------------------------------------*/

#include "MCI_SAM-ESV7.h"

#define ARM_MCI_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,01)  /* driver version */

/* MCI0 Clock */
#define BOARD_MCI_PIN_CK  {PIO_PA25, PIOA, ID_PIOA, PIO_PERIPH_D, PIO_DEFAULT}
/* MCI0 Slot A IO pins definition. (PC4-PC13) */
#define BOARD_MCI_PINS_SLOTA {(PIO_PA30 | \
                               PIO_PA31 | \
                               PIO_PA26 | \
                               PIO_PA27 | \
                               PIO_PA28 ),\
                               PIOA, ID_PIOA, PIO_PERIPH_C, PIO_DEFAULT}
/* SD card pins instance. */
static const Pin sMCIPins[] = {BOARD_MCI_PINS_SLOTA, BOARD_MCI_PIN_CK};
/* SD card detection pin instance. */
#if (RTE_MCI_CD_PORT_ID != 0)
static const Pin sMCICardDetectPin = RTE_MCI_CD_PORT;
#endif
/* SD write protection pin instance. */
#if (RTE_MCI_WP_PORT_ID != 0)
static const Pin sMCIWriteProtectPin = RTE_MCI_WP_PORT;
#endif

/* DMA driver instance */
static sXdmad dmaDrv;
/* MCI driver instance. */
static sMcid mciDrv;
/* SDCard driver instance. */
//static sSdCard sdDrv;
/* sSdmmcCommand */
static sSdmmcCommand sdcmd;

static MCI_INFO MCI;

/* Driver Version */
static const ARM_DRIVER_VERSION DriverVersion = {
  ARM_MCI_API_VERSION,
  ARM_MCI_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_MCI_CAPABILITIES DriverCapabilities = {
  (RTE_MCI_CD_PORT_ID != 0) ? 1U : 0U,            /* cd_state          */
  0U,                                             /* cd_event          */
  (RTE_MCI_WP_PORT_ID != 0) ? 1U : 0U,            /* wp_state          */
  0U,                                             /* vdd               */
  0U,                                             /* vdd_1v8           */
  0U,                                             /* vccq              */
  0U,                                             /* vccq_1v8          */
  0U,                                             /* vccq_1v2          */
  1U,                                             /* data_width_4      */
  1U,                                             /* data_width_8      */
  0U,                                             /* data_width_4_ddr  */
  0U,                                             /* data_width_8_ddr  */
  1U,                                             /* high_speed        */
  0U,                                             /* uhs_signaling     */
  0U,                                             /* uhs_tuning        */
  0U,                                             /* uhs_sdr50         */
  0U,                                             /* uhs_sdr104        */
  0U,                                             /* uhs_ddr50         */
  0U,                                             /* uhs_driver_type_a */
  0U,                                             /* uhs_driver_type_c */
  0U,                                             /* uhs_driver_type_d */
  0U,                                             /* sdio_interrupt    */
  0U,                                             /* read_wait         */
  0U,                                             /* suspend_resume    */
  0U,                                             /* mmc_interrupt     */
  0U,                                             /* mmc_boot          */
  0U,                                             /* rst_n             */
  0U,                                             /* ccs               */
  0U                                              /* ccs_timeout       */
};


/* MCI Callback */
void MCI_Cb (uint32_t status, void *pArg) {
  uint32_t event;
  uint32_t r[4];
  if (status == SDMMC_OK) {
    event = ARM_MCI_EVENT_COMMAND_COMPLETE;
    if (MCI.flags & MCI_DATA_XFER) {
      MCI.flags &= ~MCI_DATA_XFER;
      MCI.status.transfer_active = 0U;
      event |= ARM_MCI_EVENT_TRANSFER_COMPLETE;
    }
    if (sdcmd.cmdOp.bmBits.respType==2) {
      memcpy(r,sdcmd.pResp,16);
      sdcmd.pResp[0]=r[3];
      sdcmd.pResp[1]=r[2];
      sdcmd.pResp[2]=r[1];
      sdcmd.pResp[3]=r[0];
    }
  } else {
    event = ARM_MCI_EVENT_COMMAND_ERROR;
  }
  MCI.status.command_active = 0U;
  MCI.cb_event (event);
}


/**
  \fn          ARM_DRV_VERSION MCI_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRIVER_VERSION MCI_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_MCI_CAPABILITIES MCI_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_MCI_CAPABILITIES
*/
static ARM_MCI_CAPABILITIES MCI_GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn            int32_t MCI_Initialize (ARM_MCI_SignalEvent_t cb_event)
  \brief         Initialize the Memory Card Interface
  \param[in]     cb_event  Pointer to \ref ARM_MCI_SignalEvent
  \return        \ref execution_status
*/
static int32_t MCI_Initialize (ARM_MCI_SignalEvent_t cb_event) {

  if (MCI.flags & MCI_POWER) { return ARM_DRIVER_ERROR; }
  if (MCI.flags & MCI_INIT)  { return ARM_DRIVER_OK;    }

  MCI.cb_event = cb_event;

  /* Clear status */
  MCI.status.command_active  = 0U;
  MCI.status.transfer_active = 0U;
  
  /* Configure SD card pins */
  PIO_Configure(sMCIPins, PIO_LISTSIZE(sMCIPins));
#if (RTE_MCI_CD_PORT_ID != 0)
  PIO_Configure(&sMCICardDetectPin, 1);
#endif
#if (RTE_MCI_WP_PORT_ID != 0)
  PIO_Configure(&sMCIWriteProtectPin, 1);
#endif
  
  MCI.flags = MCI_INIT;

  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t MCI_Uninitialize (void)
  \brief         De-initialize Memory Card Interface.
  \return        \ref execution_status
*/
static int32_t MCI_Uninitialize (void) {

  MCI.flags = 0U;

  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t MCI_PowerControl (ARM_POWER_STATE state)
  \brief         Control Memory Card Interface Power.
  \param[in]     state   Power state \ref ARM_POWER_STATE
  \return        \ref execution_status
*/
static int32_t MCI_PowerControl (ARM_POWER_STATE state) {
uint32_t mck;
  
  switch (state) {
    case ARM_POWER_OFF:
      MCI.flags &= ~(MCI_POWER | MCI_SETUP);
      /* Disable interrupts */
      NVIC_ClearPendingIRQ(HSMCI_IRQn);
      NVIC_DisableIRQ( HSMCI_IRQn );
      NVIC_ClearPendingIRQ(XDMAC_IRQn);
      NVIC_DisableIRQ( XDMAC_IRQn );
      /* Disable HSMCI */
      HSMCI_EnableDma( HSMCI, 0 );
      PMC_DisablePeripheral( ID_HSMCI );
      break;

    case ARM_POWER_FULL:
      if ((MCI.flags & MCI_INIT) == 0U) { return ARM_DRIVER_ERROR; }
      
      if ((MCI.flags & MCI_POWER) == 0) {        
        /* Initialize the DMA driver */
        XDMAD_Initialize(&dmaDrv,0);

        /* Enable XDMA interrupt and give it priority over any other peripheral interrupt */
        NVIC_ClearPendingIRQ(XDMAC_IRQn);
        NVIC_SetPriority(XDMAC_IRQn, 1);
        NVIC_EnableIRQ( XDMAC_IRQn );

        /* Initialize the HSMCI driver */
        mck = SystemCoreClock;
        switch (PMC->PMC_MCKR & PMC_MCKR_MDIV_Msk) {
          case PMC_MCKR_MDIV_PCK_DIV2: mck /= 2; break;
          case PMC_MCKR_MDIV_PCK_DIV4: mck /= 4; break;
          case PMC_MCKR_MDIV_PCK_DIV3: mck /= 3; break;
        }
        MCID_Init(&mciDrv, HSMCI, ID_HSMCI, mck, &dmaDrv, 0 ) ;

        /* Enable MCI interrupt and give it priority lower than DMA*/
        NVIC_ClearPendingIRQ(HSMCI_IRQn);
        NVIC_SetPriority(HSMCI_IRQn, 3);
        NVIC_EnableIRQ( HSMCI_IRQn );

        MCI.flags |= MCI_POWER;
      }
      break;

    case ARM_POWER_LOW:
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }
  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t MCI_CardPower (uint32_t voltage)
  \brief         Set Memory Card supply voltage.
  \param[in]     voltage  Memory Card supply voltage
  \return        \ref execution_status
*/
static int32_t MCI_CardPower (uint32_t voltage) {

  if ((MCI.flags & MCI_POWER) == 0U) { return ARM_DRIVER_ERROR; }
  return ARM_DRIVER_ERROR_UNSUPPORTED;
}


/**
  \fn            int32_t MCI_ReadCD (void)
  \brief         Read Card Detect (CD) state.
  \return        1:card detected, 0:card not detected, or error
*/
static int32_t MCI_ReadCD (void) {

  if ((MCI.flags & MCI_POWER) == 0U) { return ARM_DRIVER_ERROR; }  
#if (RTE_MCI_CD_PORT_ID != 0)
  return (PIO_Get(&sMCICardDetectPin) ? 0 : 1);
#else
  return (0);
#endif
}


/**
  \fn            int32_t MCI_ReadWP (void)
  \brief         Read Write Protect (WP) state.
  \return        1:write protected, 0:not write protected, or error
*/
static int32_t MCI_ReadWP (void) {
  
  if ((MCI.flags & MCI_POWER) == 0U) { return ARM_DRIVER_ERROR; }
#if (RTE_MCI_WP_PORT_ID==1)
  return = (PIO_Get(&sMCIWriteProtectPin) ? 0 : 1);
#else
  return (0);
#endif
}


/**
  \fn            int32_t MCI_SendCommand (uint32_t  cmd,
                                      uint32_t  arg,
                                      uint32_t  flags,
                                      uint32_t *response)
  \brief         Send Command to card and get the response.
  \param[in]     cmd       Memory Card command
  \param[in]     arg       Command argument
  \param[in]     flags     Command flags
  \param[out]    response  Pointer to buffer for response
  \return        \ref execution_status
*/
static int32_t MCI_SendCommand (uint32_t cmd, uint32_t arg, uint32_t flags, uint32_t *response) {
    
  if (((flags & MCI_RESPONSE_EXPECTED_Msk) != 0U) && (response == NULL)) {
    return ARM_DRIVER_ERROR_PARAMETER;
  }
  if ((MCI.flags & MCI_SETUP) == 0U) {
    return ARM_DRIVER_ERROR;
  }
  if (MCI.status.command_active) {
    return ARM_DRIVER_ERROR_BUSY;
  }
  MCI.status.command_active   = 1U;
  MCI.flags &= ~(MCI_RESP_CRC | MCI_RESP_LONG);
    
  /* Card initialize */
  if (flags & ARM_MCI_CARD_INITIALIZE) {
    memset(&sdcmd, 0, sizeof(sSdmmcCommand));
    sdcmd.cmdOp.wVal = SDMMC_CMD_POWERONINIT;
    MCID_SendCmd(&mciDrv, &sdcmd);
    while(!MCID_IsCmdCompleted(&mciDrv));
  }
  
  /* Reset sdmmc command structure */
  memset(&sdcmd, 0, sizeof(sSdmmcCommand));
  sdcmd.cmdOp.wVal |= SDMMC_CMD_bmCOMMAND;
  
  switch (flags & ARM_MCI_RESPONSE_Msk) {
    case ARM_MCI_RESPONSE_NONE:
      /* No response expected */
      break;

    case ARM_MCI_RESPONSE_SHORT_BUSY: 
      /* Busy check */
      sdcmd.cmdOp.wVal |= SDMMC_CMD_bmBUSY;      
    case ARM_MCI_RESPONSE_SHORT:  
      /* Short response */      
      sdcmd.cmdOp.wVal |= SDMMC_CMD_bmRESP(3);
      /* Stop command */
      if (cmd==12) sdcmd.cmdOp.wVal |= SDMMC_CMD_bmSTOPXFR;
      break;

    case ARM_MCI_RESPONSE_LONG:
     /* Long response */ 
      MCI.flags |= MCI_RESP_LONG;
      sdcmd.cmdOp.wVal |= SDMMC_CMD_bmRESP(2);
      break;

    default:
      return ARM_DRIVER_ERROR;
  }
  
  /* Open drain command */
  if (MCI.bus_cmd_mode & ARM_MCI_BUS_CMD_OPEN_DRAIN) {
    sdcmd.cmdOp.wVal |= SDMMC_CMD_bmOD;
  }
  
  /* CRC */ 
  if (flags & ARM_MCI_RESPONSE_CRC) {
    MCI.flags |= MCI_RESP_CRC;
    sdcmd.cmdOp.wVal |= SDMMC_CMD_bmCRC;
  }
  
  /* Set data transfer parameters */
  if (flags & ARM_MCI_TRANSFER_DATA) {
    MCI.flags |= MCI_DATA_XFER;
    MCI.status.transfer_active = 1U;
    
    sdcmd.wBlockSize = MCI.xfer.block_size;
    sdcmd.wNbBlocks  = MCI.xfer.block_count;
    sdcmd.pData      = MCI.xfer.buf;
    
    if (MCI.flags & MCI_DATA_READ)
      sdcmd.cmdOp.wVal |= SDMMC_CMD_bmDATARX;
    else
      sdcmd.cmdOp.wVal |= SDMMC_CMD_bmDATATX;
  }  

  /* Set other command parameters */
  sdcmd.fCallback = MCI_Cb;
  sdcmd.pResp = response;
  sdcmd.dwArg = arg;
  sdcmd.bCmd = cmd;
  
  /* Send command */
  MCID_SendCmd(&mciDrv, &sdcmd);

  return ARM_DRIVER_OK;
}


/**
  \fn            int32_t MCI_SetupTransfer (uint8_t *data,
                                        uint32_t block_count,
                                        uint32_t block_size,
                                        uint32_t mode)
  \brief         Setup read or write transfer operation.
  \param[in,out] data         Pointer to data block(s) to be written or read
  \param[in]     block_count  Number of blocks
  \param[in]     block_size   Size of a block in bytes
  \param[in]     mode         Transfer mode
  \return        \ref execution_status
*/
static int32_t MCI_SetupTransfer (uint8_t *data, uint32_t block_count, uint32_t block_size, uint32_t mode) {

  if ((data == NULL) || (block_count == 0U) || (block_size == 0U)) { return ARM_DRIVER_ERROR_PARAMETER; }

  if ((MCI.flags & MCI_SETUP) == 0U) {
    return ARM_DRIVER_ERROR;
  }
  if (MCI.status.transfer_active) {
    return ARM_DRIVER_ERROR_BUSY;
  }

  /* Set transfer buffer and block size */
  MCI.xfer.buf = data;
  MCI.xfer.block_count = block_count;
  MCI.xfer.block_size = block_size;  

  /* Set transfer direction */
  if ((mode & ARM_MCI_TRANSFER_WRITE) == 0)
    MCI.flags |= MCI_DATA_READ;
  else 
    MCI.flags &= ~MCI_DATA_READ;

  return (ARM_DRIVER_OK);
}


/**
  \fn            int32_t MCI_AbortTransfer (void)
  \brief         Abort current read/write data transfer.
  \return        \ref execution_status
*/
static int32_t MCI_AbortTransfer (void) {
  int32_t  status;

  if ((MCI.flags & MCI_SETUP) == 0U) { return ARM_DRIVER_ERROR; }
  
  /* Cancel command */ 
  MCID_CancelCmd(&mciDrv);
  
  MCI.status.command_active  = 0U;
  MCI.status.transfer_active = 0U;
  
  return status;
}


/**
  \fn            int32_t MCI_Control (uint32_t control, uint32_t arg)
  \brief         Control MCI Interface.
  \param[in]     control  Operation
  \param[in]     arg      Argument of operation (optional)
  \return        \ref execution_status
*/
static int32_t MCI_Control (uint32_t control, uint32_t arg) {
  uint32_t bps, mck, clkdiv;

  if ((MCI.flags & MCI_POWER) == 0U) { return ARM_DRIVER_ERROR; }

  switch (control) {
    case ARM_MCI_BUS_SPEED:
      mck = SystemCoreClock;
      switch (PMC->PMC_MCKR & PMC_MCKR_MDIV_Msk) {
        case PMC_MCKR_MDIV_PCK_DIV2: mck /= 2; break;
        case PMC_MCKR_MDIV_PCK_DIV4: mck /= 4; break;
        case PMC_MCKR_MDIV_PCK_DIV3: mck /= 3; break;
      }
      if((mck % arg) == 0) clkdiv = mck/arg;
      else clkdiv = ((mck + arg)/arg);
      bps = mck/clkdiv;
      // Set clock divider
      HSMCI_DivCtrl(HSMCI, clkdiv, 0x7);  
      // Bus speed configured
      MCI.flags |= MCI_SETUP;
      return ((int32_t)bps);

    case ARM_MCI_BUS_SPEED_MODE:
      switch (arg) {
        case ARM_MCI_BUS_DEFAULT_SPEED:
        case ARM_MCI_BUS_HIGH_SPEED:
          /* Enable/Disable high speed */ 
          HSMCI_HsEnable(HSMCI,arg);
          break;
        default: return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_MCI_BUS_CMD_MODE:
      switch (arg) {
        case ARM_MCI_BUS_CMD_PUSH_PULL:
        case ARM_MCI_BUS_CMD_OPEN_DRAIN:
          /* Configure command line open-drain/push-pull mode */
          MCI.bus_cmd_mode = arg;
          break;
        default:
          return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_MCI_BUS_DATA_WIDTH:
      switch (arg) {
        case ARM_MCI_BUS_DATA_WIDTH_1:
          HSMCI_SetBusWidth(HSMCI,1U);
          break;
        case ARM_MCI_BUS_DATA_WIDTH_4:
          HSMCI_SetBusWidth(HSMCI,4U);
          break;
        case ARM_MCI_BUS_DATA_WIDTH_8:
          HSMCI_SetBusWidth(HSMCI,8U);
          break;
        default:
          return ARM_DRIVER_ERROR_UNSUPPORTED;
      }
      break;

    case ARM_MCI_DATA_TIMEOUT:
      HSMCI_ConfigureDataTO(HSMCI, HSMCI_DTOR_DTOCYC(1+arg/1048576)|HSMCI_DTOR_DTOMUL_1048576);
      break;

    case ARM_MCI_CONTROL_CLOCK_IDLE:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
    
    case ARM_MCI_MONITOR_SDIO_INTERRUPT:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
    
    case ARM_MCI_CONTROL_READ_WAIT:
      return ARM_DRIVER_ERROR_UNSUPPORTED;

    default: return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}


/**
  \fn            ARM_MCI_STATUS MCI_GetStatus (void)
  \brief         Get MCI status.
  \return        MCI status \ref ARM_MCI_STATUS
*/
static ARM_MCI_STATUS MCI_GetStatus (void) {
  return MCI.status;
}


/**
 * MCI interrupt handler. Forwards the event to the MCI driver handlers.
 */
void HSMCI_Handler(void)
{
  MCID_Handler(&mciDrv);
}


/* MCI Driver Control Block */
ARM_DRIVER_MCI Driver_MCI0 = {
  MCI_GetVersion,
  MCI_GetCapabilities,
  MCI_Initialize,
  MCI_Uninitialize,
  MCI_PowerControl,
  MCI_CardPower,
  MCI_ReadCD,
  MCI_ReadWP,
  MCI_SendCommand,
  MCI_SetupTransfer,
  MCI_AbortTransfer,
  MCI_Control,
  MCI_GetStatus
};
