/* -----------------------------------------------------------------------------
 * Copyright (c) 2013-2017 Arm Limited
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 *
 * $Date:        29. January 2018
 * $Revision:    V1.5
 *
 * Driver:       Driver_USBD0
 * Project:      USB High/Full/Low-Speed Device Driver for Atmel SAM-ESV7
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                  Value
 *   ---------------------                  -----
 *   Connect to hardware via Driver_USBD# = 0
 * --------------------------------------------------------------------------
 * Defines used for driver configuration (at compile time):
 *
 *   USBD_MAX_ENDPOINT_NUM:  defines maximum number of IN/OUT Endpoint pairs
 *                           that driver will support with Control Endpoint 0
 *                           not included, this value impacts driver memory
 *                           requirements
 *     - default value:      9
 *     - maximum value:      9
 *
 *   USB_PEPx_BANK_NUM:      defines number of banks used by endpoint x (x=0..9)
 *     - default settings:
 *       PEP0: 1 bank (for control endpoint 0)
 *       PEP1: 1 bank (for bulk/interrupt endpoint)
 *       PEP2: 1 bank (for bulk/interrupt endpoint)
 *       PEP3: 1 bank (for bulk/interrupt endpoint)
 *       PEP4: 1 bank (for bulk/interrupt endpoint)
 *       PEP5: 1 bank (for bulk/interrupt endpoint)
 *       PEP6: 1 bank (for bulk/interrupt endpoint)
 *       PEP7: 1 bank (for bulk/interrupt endpoint)
 *       PEP8: 1 bank (for bulk/interrupt endpoint)
 *       PEP9: 1 bank (for bulk/interrupt endpoint)
 *
 *   USB_DMA1_EN:            defines if DMA is used by endpoint x (x=1..7)
 *     - default settings:
 *       DMA1: disabled (for bulk/interrupt endpoint)
 *       DMA2: disabled (for bulk/interrupt endpoint)
 *       DMA3: disabled (for bulk/interrupt endpoint)
 *       DMA4: disabled (for bulk/interrupt endpoint)
 *       DMA5: disabled (for bulk/interrupt endpoint)
 *       DMA6: disabled (for bulk/interrupt endpoint)
 *       DMA7: disabled (for bulk/interrupt endpoint)
 *
 *   USB_WT_RAM_EN:          defines if USB data is in RAM with write-through
 *                           policy enabled (= 1)
 *     - default value:      0
 * -------------------------------------------------------------------------- */

/* History:
 *  Version 1.5
 *    Corrected IRQ routine to process also last endpoint (number 9)
 *  Version 1.4
 *    Added warning that DMA can only be used with USB data in non-cacheable or 
 *    shareable or write-through policy enabled RAM 
 *    (otherwise USB data would need to be of size multiple of 32 and 32 byte 
 *     aligned for cache handling with DMA)
 *  Version 1.3
 *    Corrected configuration for Isochronous Endpoint
 *  Version 1.2
 *    Enabled multiple bank usage for improved performance
 *    Enabled DMA usage for improved performance
 *    Corrected OUT reception in full-speed mode
 *    Added delay for stalling as otherwise it can stall already active transfer
 *    All endless loops changed to end with timeout
 *    Corrected suspend/resume signaling
 *  Version 1.0
 *    Initial release
 */

#include "USBD_SAM-ESV7.h"

// Macros

#define GET_EPSIZE_VAL(x)    (((x) <=    8U) ? 0U : \
                              ((x) <=   16U) ? 1U : \
                              ((x) <=   32U) ? 2U : \
                              ((x) <=   64U) ? 3U : \
                              ((x) <=  128U) ? 4U : \
                              ((x) <=  256U) ? 5U : \
                              ((x) <=  512U) ? 6U : \
                                               7U)

#define EP_NUM(ep_addr)       ((ep_addr) & ARM_USB_ENDPOINT_NUMBER_MASK)
#define EP_DIR(ep_addr)      (((ep_addr) & ARM_USB_ENDPOINT_DIRECTION_MASK) >> 7)

// Check configuration settings

#ifndef USBD_MAX_ENDPOINT_NUM
#define USBD_MAX_ENDPOINT_NUM  (9U)
#define USBD_MAX_ENDPOINT_MSK  ((1U << USBD_MAX_ENDPOINT_NUM) - 1U)
#endif
#if    (USBD_MAX_ENDPOINT_NUM > 9U)
#error  Too many Endpoints, maximum IN/OUT Endpoint pairs that this driver supports is 9 !!!
#endif

#ifndef USBHS_RAM_ADDR         
#define USBHS_RAM_ADDR         (0xA0100000U)
#endif

// Bank number (single, double or triple) and bank size configuration
// (total available memory for all used banks is 4096 bytes)
#ifndef USB_PEP0_BANK_NUM
#define USB_PEP0_BANK_NUM      (1U)
#endif
#if    (USB_PEP0_BANK_NUM == 0U)
#error Endpoint 0 has to contain 1 bank!
#elif  (USB_PEP0_BANK_NUM != 1U)
#error Endpoint 0 does not support multiple banks!
#endif

#ifndef USB_PEP1_BANK_NUM
#define USB_PEP1_BANK_NUM      (1U)
#endif
#if    (USB_PEP1_BANK_NUM > 3U)
#error Driver supports maximum 3 banks for Endpoint 1!
#endif

#ifndef USB_PEP2_BANK_NUM
#define USB_PEP2_BANK_NUM      (1U)
#endif
#if    (USB_PEP2_BANK_NUM > 3U)
#error Driver supports maximum 3 banks for Endpoint 2!
#endif

#ifndef USB_PEP3_BANK_NUM
#define USB_PEP3_BANK_NUM      (1U)
#endif
#if    (USB_PEP3_BANK_NUM > 2U)
#error Endpoint 3 does not support more than 2 banks!
#endif

#ifndef USB_PEP4_BANK_NUM
#define USB_PEP4_BANK_NUM      (1U)
#endif
#if    (USB_PEP4_BANK_NUM > 2U)
#error Endpoint 4 does not support more than 2 banks!
#endif

#ifndef USB_PEP5_BANK_NUM
#define USB_PEP5_BANK_NUM      (1U)
#endif
#if    (USB_PEP5_BANK_NUM > 2U)
#error Endpoint 5 does not support more than 2 banks!
#endif

#ifndef USB_PEP6_BANK_NUM
#define USB_PEP6_BANK_NUM      (1U)
#endif
#if    (USB_PEP6_BANK_NUM > 2U)
#error Endpoint 6 does not support more than 2 banks!
#endif

#ifndef USB_PEP7_BANK_NUM
#define USB_PEP7_BANK_NUM      (1U)
#endif
#if    (USB_PEP7_BANK_NUM > 2U)
#error Endpoint 7 does not support more than 2 banks!
#endif

#ifndef USB_PEP8_BANK_NUM
#define USB_PEP8_BANK_NUM      (1U)
#endif
#if    (USB_PEP8_BANK_NUM > 2U)
#error Endpoint 8 does not support more than 2 banks!
#endif

#ifndef USB_PEP9_BANK_NUM
#define USB_PEP9_BANK_NUM      (1U)
#endif
#if    (USB_PEP9_BANK_NUM > 2U)
#error Endpoint 9 does not support more than 2 banks!
#endif

// Endpoint DMA usage configuration
#ifndef USB_DMA1_EN
#define USB_DMA1_EN            (0U)
#endif

#ifndef USB_DMA2_EN
#define USB_DMA2_EN            (0U)
#endif

#ifndef USB_DMA3_EN
#define USB_DMA3_EN            (0U)
#endif

#ifndef USB_DMA4_EN
#define USB_DMA4_EN            (0U)
#endif

#ifndef USB_DMA5_EN
#define USB_DMA5_EN            (0U)
#endif

#ifndef USB_DMA6_EN
#define USB_DMA6_EN            (0U)
#endif

#ifndef USB_DMA7_EN
#define USB_DMA7_EN            (0U)
#endif

#ifndef USB_WT_RAM_EN
#define USB_WT_RAM_EN          (0U)
#endif

#if   ((USB_DMA1_EN != 0) || \
       (USB_DMA2_EN != 0) || \
       (USB_DMA3_EN != 0) || \
       (USB_DMA4_EN != 0) || \
       (USB_DMA5_EN != 0) || \
       (USB_DMA6_EN != 0) || \
       (USB_DMA7_EN != 0))
#define USB_DMA_USED           (1U)
#warning Ensure that USB data is in non-cacheable or shareable or write-through policy enabled RAM !
#warning If USB data is RAM with write-through policy enabled define USB_WT_RAM_EN = 1 !
#else
#define USB_DMA_USED           (0U)
#endif

static const uint32_t ep_bnk_num = (USB_PEP0_BANK_NUM      ) | 
                                   (USB_PEP1_BANK_NUM <<  2) | 
                                   (USB_PEP2_BANK_NUM <<  4) | 
                                   (USB_PEP3_BANK_NUM <<  6) | 
                                   (USB_PEP4_BANK_NUM <<  8) | 
                                   (USB_PEP5_BANK_NUM << 10) | 
                                   (USB_PEP6_BANK_NUM << 12) | 
                                   (USB_PEP7_BANK_NUM << 14) | 
                                   (USB_PEP8_BANK_NUM << 16) | 
                                   (USB_PEP9_BANK_NUM << 18) ;


static const uint32_t ep_dma_en  = (USB_DMA1_EN << 1) | 
                                   (USB_DMA2_EN << 2) | 
                                   (USB_DMA3_EN << 3) | 
                                   (USB_DMA4_EN << 4) | 
                                   (USB_DMA5_EN << 5) | 
                                   (USB_DMA6_EN << 6) | 
                                   (USB_DMA7_EN << 7) ;

// USBD Driver *****************************************************************

#define ARM_USBD_DRV_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,5)

// Driver Version
static const ARM_DRIVER_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

// Driver Capabilities
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  0U,   // VBUS Detection
  0U,   // Event VBUS On
  0U    // Event VBUS Off
};

typedef struct {                        // Endpoint structure definition
  uint8_t  *data;
  uint32_t  num;
  uint8_t   ep_addr;
  uint8_t   ep_type;
  uint16_t  ep_max_packet_size;
  uint32_t  num_transferring;
  uint32_t  num_transferred_total;
  uint32_t  active[2];                  // active[0] = out activity, active[1] = in activity
} ENDPOINT_t;

static ARM_USBD_SignalDeviceEvent_t   SignalDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t SignalEndpointEvent;

static bool           hw_powered     = false;
static bool           hw_initialized = false;
static volatile bool  setup_received;   // Setup packet received
static uint32_t       setup_packet[2];  // Setup packet data
static ARM_USBD_STATE usbd_state;

// Endpoints runtime information
static ENDPOINT_t     ep[USBD_MAX_ENDPOINT_NUM + 1U];


// Auxiliary functions

/**
  \fn          void USBD_FlushTxFIFO (uint8_t ep_num)
  \brief       Flush Transmit Endpoint FIFO.
  \param[in]   ep_num   Endpoint Number (0..USBD_MAX_ENDPOINT_NUM)
*/
static void USBD_FlushTxFIFO (uint8_t ep_num) {
  uint32_t tout1, tout2;

  // Flush any pending data
  USBHS->USBHS_DEVIDR            = USBHS_DEVIDR_PEP_0 << ep_num;
  USBHS->USBHS_DEVEPTIDR[ep_num] = USBHS_DEVEPTIDR_TXINEC;
  tout1 = 10U;
  while (((USBHS->USBHS_DEVEPTISR[ep_num] & USBHS_DEVEPTISR_NBUSYBK_Msk) != 0U) && (tout1-- != 0U)) {
    // While there is data that needs to be flushed, flush it
    USBHS->USBHS_DEVEPTIER[ep_num] = USBHS_DEVEPTIER_KILLBKS;
    tout2 = SystemCoreClock;
    while (((USBHS->USBHS_DEVEPTIMR[ep_num] & USBHS_DEVEPTIMR_KILLBK) != 0U) && (tout2-- != 0U));
  }
}

/**
  \fn          void USBD_Reset (void)
  \brief       Reset USB Endpoint settings and variables.
*/
static void USBD_Reset (void) {
  ENDPOINT_t *ptr_ep;
  uint8_t     ep_num;
  
  // All Endpoint except Endpoint 0 are disabled by hardware upon USB Reset signal

  for (ep_num = 1U; ep_num < USBD_MAX_ENDPOINT_NUM; ep_num++) {
    // Flush active transmits
    ptr_ep = &ep[ep_num];
    if (ptr_ep->active[1] != 0U) {
      USBD_FlushTxFIFO (ep_num);
    }
  }

  // Reset global variables
  setup_received  = false;
  setup_packet[0] = 0U;
  setup_packet[1] = 0U;
  memset((void *)&usbd_state, 0, sizeof(usbd_state));
  memset((void *)ep,          0, sizeof(ep));
}

/**
  \fn          void USBD_ReadFromFIFO (uint8_t ep_num, uint8_t *data, uint32_t num)
  \brief       Read from Endpoint FIFO.
  \param[in]   ep_num   Endpoint Number (0..USBD_MAX_ENDPOINT_NUM)
  \param[out]  data     Pointer to buffer for data to read
  \param[in]   num      Number of data bytes to read
  \return               Number of data bytes read
*/
static uint32_t USBD_ReadFromFIFO (uint8_t ep_num, uint8_t *data, uint32_t num) {
  uint8_t  *ptr_FIFO;
  uint32_t  len;

  ptr_FIFO = (uint8_t *)((uint32_t *)USBHS_RAM_ADDR + ((EPT_VIRTUAL_SIZE/2) * ep_num));
  len      =  num;

#if (defined(__DCACHE_PRESENT) && (__DCACHE_PRESENT == 1U))
  if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {      // If Data Cache is enabled
    SCB_InvalidateDCache_by_Addr((uint32_t*)USBHS_RAM_ADDR + ((EPT_VIRTUAL_SIZE/2) * ep_num), len);
  }
#endif

  while (len-- > 0U) {
    *(data++) = *(ptr_FIFO++);
  }

  return num;
}

/**
  \fn          void USBD_WriteToFIFO (uint8_t ep_num, uint8_t ep_dir, uint8_t *data, uint32_t num)
  \brief       Write to Endpoint FIFO.
  \param[in]   ep_num   Endpoint Number (0..USBD_MAX_ENDPOINT_NUM)
  \param[out]  data     Pointer to buffer with data to write
  \param[in]   num      Number of data bytes to write
  \return               Number of data bytes written
*/
static uint32_t USBD_WriteToFIFO (uint8_t ep_num, uint8_t *data, uint32_t num) {
  uint8_t  *ptr_FIFO;
  uint32_t  len;

  ptr_FIFO = (uint8_t *)((uint32_t *)USBHS_RAM_ADDR + ((EPT_VIRTUAL_SIZE/2) * ep_num));

  if (num > ep[ep_num].ep_max_packet_size) { num = ep[ep_num].ep_max_packet_size; }
  len = num;

  while (len-- > 0) {
    *(ptr_FIFO++) = *(data++);
  }

#if (defined(__DCACHE_PRESENT) && (__DCACHE_PRESENT == 1U))
  if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {      // If Data Cache is enabled
    SCB_CleanDCache_by_Addr((uint32_t*)USBHS_RAM_ADDR + ((EPT_VIRTUAL_SIZE/2) * ep_num), num);
  }
#endif

  return num;
}


// USBD Driver functions

/**
  \fn          ARM_DRIVER_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION USBD_GetVersion (void) { return usbd_driver_version; }

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES USBD_GetCapabilities (void) { return usbd_driver_capabilities; }

/**
  \fn          int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                        ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref execution_status
*/
static int32_t USBD_Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  if (hw_initialized == true) { return ARM_DRIVER_OK; }

  SignalDeviceEvent   = cb_device_event;
  SignalEndpointEvent = cb_endpoint_event;

  hw_initialized = true;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref execution_status
*/
static int32_t USBD_Uninitialize (void) {

  hw_initialized = false;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t USBD_PowerControl (ARM_POWER_STATE state) {
  uint32_t tout;

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_DisableIRQ      (USBHS_IRQn);                // Disable interrupt
      NVIC_ClearPendingIRQ (USBHS_IRQn);                // Clear pending interrupt
      hw_powered = false;                               // Clear powered flag

      USBHS->USBHS_CTRL &= ~USBHS_CTRL_USBE;            // Disable USBHS

      setup_received =  0U;                             // Reset variables
      memset((void *)(&usbd_state), 0, sizeof(usbd_state));
      memset((void *)(ep),          0, sizeof(ep));

      PMC_DisablePeripheral (ID_USBHS);                 // Stop peripheral clock
      break;

    case ARM_POWER_FULL:
      if (hw_initialized == false) { return ARM_DRIVER_ERROR; }
      if (hw_powered     == true)  { return ARM_DRIVER_OK;    }

      // UTMI parallel mode, High/Full/Low-Speed
      PMC->PMC_SCDR = PMC_SCDR_USBCLK;                  // USBCLK not used in this configuration
      PMC->PMC_USB  = PMC_USB_USBS;                     // USB Clock Input is UTMI PLL
      PMC_EnablePeripheral (ID_USBHS);                  // Enable peripheral clock for USBHS
      PMC->CKGR_UCKR = CKGR_UCKR_UPLLEN | CKGR_UCKR_UPLLCOUNT(0x0FU);   // Enable PLL 480 MHz
      tout = SystemCoreClock;
      while(((PMC->PMC_SR & PMC_SR_LOCKU) == 0U) && (tout-- != 1U));    // Wait that PLL is considered locked by the PMC
      if (tout == 0U) {
        return ARM_DRIVER_ERROR;
      }

      USBHS->USBHS_CTRL  =  USBHS_CTRL_UIMOD_DEVICE;    // Select Device Mode and Unfreeze clock and disable USBHS
      USBHS->USBHS_CTRL |=  USBHS_CTRL_USBE;            // Enable USBHS
      USBHS->USBHS_DEVCTRL&=~USBHS_DEVCTRL_SPDCONF_Msk; // Select normal mode (High-speed capable)
      if (RTE_USBD_HIGH_SPEED == 0U) {                  // Force Full-speed
        USBHS->USBHS_DEVCTRL |= USBHS_DEVCTRL_SPDCONF_FORCED_FS;
      }

      tout = SystemCoreClock;
      while (((USBHS->USBHS_SR & USBHS_SR_CLKUSABLE) == 0U) && (tout-- != 1U)); // Wait for usable UTMI clock
      if (tout == 0U) {
        return ARM_DRIVER_ERROR;
      }

      USBHS->USBHS_DEVCTRL |= USBHS_DEVCTRL_DETACH;     // Detach Device

      USBHS->USBHS_DEVIER = USBHS_DEVIER_EORSTES  |     // Enable interrupts
                            USBHS_DEVIER_WAKEUPES |
                            USBHS_DEVIER_SUSPES   ;

      USBHS->USBHS_DEVICR = USBHS_DEVICR_EORSTC   |     // Clear interrupts
                            USBHS_DEVICR_SOFC     |
                            USBHS_DEVICR_MSOFC    |
                            USBHS_DEVICR_SUSPC    ;

      hw_powered = true;                                // Set powered flag
      NVIC_EnableIRQ (USBHS_IRQn);                      // Enable interrupt
      break;

    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceConnect (void) {

  if (hw_powered == false) { return ARM_DRIVER_ERROR; }

  USBHS->USBHS_DEVCTRL &= ~USBHS_DEVCTRL_DETACH;        // Attach Device

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceDisconnect (void) {

  if (hw_powered == false) { return ARM_DRIVER_ERROR; }

  USBHS->USBHS_DEVCTRL |= USBHS_DEVCTRL_DETACH;         // Detach Device

  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBD_STATE USBD_DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      Device State \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE USBD_DeviceGetState (void) {
  return usbd_state;
}

/**
  \fn          int32_t USBD_DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref execution_status
*/
static int32_t USBD_DeviceRemoteWakeup (void) {

  if (hw_powered == false) { return ARM_DRIVER_ERROR; }

  USBHS->USBHS_DEVCTRL |= USBHS_DEVCTRL_RMWKUP;         // Start upstream resume

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_DeviceSetAddress (uint8_t dev_addr)
  \brief       Set USB Device Address.
  \param[in]   dev_addr  Device Address
  \return      \ref execution_status
*/
static int32_t USBD_DeviceSetAddress (uint8_t dev_addr) {

  if (hw_powered == false) { return ARM_DRIVER_ERROR; }

  USBHS->USBHS_DEVCTRL &= ~USBHS_DEVCTRL_UADD_Msk;
  USBHS->USBHS_DEVCTRL |=  USBHS_DEVCTRL_UADD(dev_addr);
  USBHS->USBHS_DEVCTRL |=  USBHS_DEVCTRL_ADDEN;

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_ReadSetupPacket (uint8_t *setup)
  \brief       Read setup packet received over Control Endpoint.
  \param[out]  setup  Pointer to buffer for setup packet
  \return      \ref execution_status
*/
static int32_t USBD_ReadSetupPacket (uint8_t *setup) {

  if (hw_powered == false)  { return ARM_DRIVER_ERROR; }
  if (setup_received == 0U) { return ARM_DRIVER_ERROR; }

  setup_received  = false;
  memcpy(setup, setup_packet, 8);

  if (setup_received != 0U) {           // If new setup packet was received while this was being read
    return ARM_DRIVER_ERROR;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                               uint8_t  ep_type,
                                               uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type  Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t USBD_EndpointConfigure (uint8_t  ep_addr,
                                       uint8_t  ep_type,
                                       uint16_t ep_max_packet_size) {
  ENDPOINT_t *ptr_ep;
  uint32_t    deveptcfg, tout;
  uint8_t     ep_num, ep_dir;

  ep_num =  EP_NUM(ep_addr);
  ep_dir =  EP_DIR(ep_addr);
  ptr_ep = &ep[ep_num];

  if (hw_powered == false)            { return ARM_DRIVER_ERROR;      }
  if (ep_num > USBD_MAX_ENDPOINT_NUM) { return ARM_DRIVER_ERROR;      }
  if (ptr_ep->active[ep_dir] != 0U)   { return ARM_DRIVER_ERROR_BUSY; }

  if ((ep_num == 0U) && (ep_dir == 1U)) {
    // Endpoint 0 is used in both direction so we configure it only for direction OUT
    return ARM_DRIVER_OK;
  }

  // Disable and reset endpoint
  USBHS->USBHS_DEVIDR             =   USBHS_DEVIDR_PEP_0  << ep_num;
  USBHS->USBHS_DEVEPTIDR[ep_num]  =   0U;
  USBHS->USBHS_DEVEPT            |=  (USBHS_DEVEPT_EPRST0 << ep_num);
  USBHS->USBHS_DEVEPT            &= ~(USBHS_DEVEPT_EPRST0 << ep_num);
  USBHS->USBHS_DEVEPTCFG[ep_num]  =   USBHS_DEVEPTIER_RSTDTS;
  USBHS->USBHS_DEVEPT            |=  (USBHS_DEVEPT_EPEN0  << ep_num);
  USBHS->USBHS_DEVEPTCFG[ep_num] &=  ~USBHS_DEVEPTCFG_ALLOC;

  // Clear Endpoint transfer and configuration information
  memset((void *)(ptr_ep), 0U, sizeof (ENDPOINT_t));

  ptr_ep->ep_addr            = ep_addr;
  ptr_ep->ep_type            = ep_type;
  ptr_ep->ep_max_packet_size = ep_max_packet_size;

  // Configure endpoint
  deveptcfg   = ((GET_EPSIZE_VAL(ep_max_packet_size) & 0x07U) << 4) | ((((ep_bnk_num >> (ep_num << 1)) & 3U) - 1U) << 2) | USBHS_DEVEPTCFG_ALLOC;
  if ((ep_dma_en & (1U << ep_num)) != 0U) {             // If DMA used for endpoint
    deveptcfg |= USBHS_DEVEPTCFG_AUTOSW;
  }
  deveptcfg |= ((uint32_t)ep_dir) << 8;
  switch (ep_type) {
    case ARM_USB_ENDPOINT_CONTROL:
      deveptcfg |= USBHS_DEVEPTCFG_EPTYPE_CTRL;
      break;
    case ARM_USB_ENDPOINT_ISOCHRONOUS:
      deveptcfg |= 1U << 13;                            // Support only 1 transaction per microframe
      deveptcfg |= USBHS_DEVEPTCFG_EPTYPE_ISO;
      break;
    case ARM_USB_ENDPOINT_BULK:
      deveptcfg |= USBHS_DEVEPTCFG_EPTYPE_BLK;
      break;
    case ARM_USB_ENDPOINT_INTERRUPT:
      deveptcfg |= USBHS_DEVEPTCFG_EPTYPE_INTRPT;
      break;
    default:
      return ARM_DRIVER_ERROR;
  }

  USBHS->USBHS_DEVEPTCFG[ep_num] = deveptcfg;
  tout = SystemCoreClock;
  while (((USBHS->USBHS_DEVEPTISR[ep_num] & USBHS_DEVEPTISR_CFGOK) == 0U) && (tout-- != 1U));
  if (tout == 0U) {
    return ARM_DRIVER_ERROR;
  }

  if (ep_num == 0U) {
    // Enable reception of SETUP packet for Control Endpoint 0
    USBHS->USBHS_DEVEPTIER[0] = USBHS_DEVEPTIER_RXSTPES;
  }

  // Enable endpoint interrupts
  USBHS->USBHS_DEVIER = USBHS_DEVIER_PEP_0 << ep_num;
  if ((ep_dma_en & (1U << ep_num)) != 0U) {             // If DMA used for endpoint
    USBHS->USBHS_DEVIER = (1U << 24) << ep_num;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointUnconfigure (uint8_t ep_addr) {
  ENDPOINT_t *ptr_ep;
  uint8_t     ep_num, ep_dir;

  ep_num =  EP_NUM(ep_addr);
  ep_dir =  EP_DIR(ep_addr);
  ptr_ep = &ep[ep_num];

  if (hw_powered == false)            { return ARM_DRIVER_ERROR;      }
  if (ep_num > USBD_MAX_ENDPOINT_NUM) { return ARM_DRIVER_ERROR;      }
  if (ptr_ep->active[ep_dir] != 0U)   { return ARM_DRIVER_ERROR_BUSY; }

  if (ep_num != 0U) {
    USBHS->USBHS_DEVIDR             =   USBHS_DEVIDR_PEP_0 << ep_num;
    USBHS->USBHS_DEVEPTIDR[ep_num]  =   0U;
    USBHS->USBHS_DEVEPTCFG[ep_num]  =   0U;
    USBHS->USBHS_DEVEPT            &= ~(USBHS_DEVEPT_EPEN0 << ep_num);
  }

  // Clear Endpoint transfer and configuration information
  memset((void *)(ptr_ep), 0, sizeof (ENDPOINT_t));

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   stall  Operation
                - \b false Clear
                - \b true Set
  \return      \ref execution_status
*/
static int32_t USBD_EndpointStall (uint8_t ep_addr, bool stall) {
  ENDPOINT_t *ptr_ep;
  uint8_t     ep_num, ep_dir;
  uint32_t    tout;

  ep_num =  EP_NUM(ep_addr);
  ep_dir =  EP_DIR(ep_addr);
  ptr_ep = &ep[ep_num];

  if (hw_powered == false)            { return ARM_DRIVER_ERROR; }
  if (ep_num > USBD_MAX_ENDPOINT_NUM) { return ARM_DRIVER_ERROR; }

  if (stall != 0U) {                    // Activate stall
    if (ep_dir != 0U) {                 // IN Endpoint
      for (tout = SystemCoreClock; tout > 0U; tout--) {
        if (((USBHS->USBHS_DEVEPTISR[ep_num] & USBHS_DEVEPTISR_NBUSYBK_Msk) >> USBHS_DEVEPTISR_NBUSYBK_Pos) == 0U) {
          break;
        }
      }
    }
    USBHS->USBHS_DEVEPTIER[ep_num] = USBHS_DEVEPTIER_STALLEDES | USBHS_DEVEPTIER_STALLRQS;
    ptr_ep->active[ep_dir] = 0U;
  } else {                              // De-activate stall
    USBHS->USBHS_DEVEPTIDR[ep_num] = USBHS_DEVEPTIDR_STALLEDEC | USBHS_DEVEPTIDR_STALLRQC;
    USBHS->USBHS_DEVEPTIER[ep_num] = USBHS_DEVEPTIER_RSTDTS;
  }

  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num)
  \brief       Read data from or Write data to USB Endpoint.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[out]  data Pointer to buffer for data to read or with data to write
  \param[in]   num  Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransfer (uint8_t ep_addr, uint8_t *data, uint32_t num) {
  ENDPOINT_t *ptr_ep;
  uint32_t    tout;
  uint8_t     ep_num, ep_dir;
#if (defined(USB_DMA_USED) && (USB_DMA_USED == 1U))
  uint32_t    dma_buf_len, devdmacontrol;
#if (USB_WT_RAM_EN == 1U)
  uint32_t    cache_addr, cache_len;
#endif
#endif

  ep_num =  EP_NUM(ep_addr);
  ep_dir =  EP_DIR(ep_addr);
  ptr_ep = &ep[ep_num];

  if (hw_powered == false)            { return ARM_DRIVER_ERROR;      }
  if (ep_num > USBD_MAX_ENDPOINT_NUM) { return ARM_DRIVER_ERROR;      }
  if (ptr_ep->active[ep_dir] != 0U)   { return ARM_DRIVER_ERROR_BUSY; }

  if ((ep_num == 0U) && (ep_dir == 0U) && (num == 0U)) {
    return ARM_DRIVER_OK;
  }

  if (ep_dir != 0U) {                   // IN Endpoint
    for (tout = SystemCoreClock; tout > 0U; tout--) {
      if (((USBHS->USBHS_DEVEPTISR[ep_num] & USBHS_DEVEPTISR_NBUSYBK_Msk) >> USBHS_DEVEPTISR_NBUSYBK_Pos) == 0U) {
        break;
      }
    }
    if (tout == 0U) { return ARM_DRIVER_ERROR_BUSY; }
  }

  ptr_ep->data                  = data;
  ptr_ep->num                   = num;
  ptr_ep->num_transferred_total = 0U;
  ptr_ep->num_transferring      = 0U;

  ptr_ep->active[ep_dir] = 1U;          // Transfer active
#if (defined(USB_DMA_USED) && (USB_DMA_USED == 1U))
  // If DMA is used
  if ((ep_dma_en & (1U << ep_num)) != 0U) {
#if (USB_WT_RAM_EN == 1U)
    if (ep_dir != 0U) {                 // IN Endpoint
      if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {          // If Data Cache is enabled
        cache_addr = (uint32_t)ptr_ep->data & (~0x1FU);
        cache_len  = (cache_addr - (uint32_t)ptr_ep->data) + ptr_ep->num;
        SCB_CleanDCache_by_Addr((uint32_t *)cache_addr, cache_len);
      }
    }
#endif

    if (num >= 65536U) {
      dma_buf_len = 0U;
      ptr_ep->num_transferring = 65536U;
    } else {
      dma_buf_len = num;
      ptr_ep->num_transferring = num;
    }
    devdmacontrol  = USBHS_DEVDMACONTROL_BUFF_LENGTH(dma_buf_len) |
                     USBHS_DEVDMACONTROL_END_TR_IT                |
                     USBHS_DEVDMACONTROL_END_TR_EN                |
                     USBHS_DEVDMACONTROL_CHANN_ENB                ;
    if (num > ptr_ep->ep_max_packet_size) {
    devdmacontrol |= USBHS_DEVDMACONTROL_END_B_EN                 |
                     USBHS_DEVDMACONTROL_END_BUFFIT               ;
    }

    USBHS->USBHS_DEVDMA[ep_num-1U].USBHS_DEVDMANXTDSC  = 0U;
    USBHS->USBHS_DEVDMA[ep_num-1U].USBHS_DEVDMAADDRESS = (uint32_t)data;
    USBHS->USBHS_DEVDMA[ep_num-1U].USBHS_DEVDMACONTROL = USBHS_DEVDMACONTROL_BUFF_LENGTH(dma_buf_len) |
                                                         USBHS_DEVDMACONTROL_END_BUFFIT               |
                                                         USBHS_DEVDMACONTROL_END_TR_IT                |
                                                         USBHS_DEVDMACONTROL_END_B_EN                 |
                                                         USBHS_DEVDMACONTROL_END_TR_EN                |
                                                         USBHS_DEVDMACONTROL_CHANN_ENB                ;
  } else {
#endif
    // If DMA not used
    if (ep_dir != 0U) {                 // IN Endpoint
      if (ep_num != 0U) { USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_TXINIC;   }
      if (ptr_ep->num != 0U) {
        ptr_ep->num_transferring = USBD_WriteToFIFO(ep_num, ptr_ep->data, ptr_ep->num);
      } else { 
        ptr_ep->num_transferring = 0U;
      }
      if (ep_num == 0U) { USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_TXINIC;   }
      if (ep_num != 0U) { USBHS->USBHS_DEVEPTIDR[ep_num] = USBHS_DEVEPTIDR_FIFOCONC; }

                                        // Enable TXINI interrupt to handle sending in IRQ
      USBHS->USBHS_DEVEPTIER[ep_num] = USBHS_DEVEPTIER_TXINES;
    } else {                            // OUT Endpoint

                                        // Enable RXOUTI interrupt to handle reception in IRQ
      USBHS->USBHS_DEVEPTIER[ep_num] = USBHS_DEVEPTIER_RXOUTES;
    }
#if (defined(USB_DMA_USED) && (USB_DMA_USED == 1U))
  }
#endif

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr)
  \brief       Get result of USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      number of successfully transferred data bytes
*/
static uint32_t USBD_EndpointTransferGetResult (uint8_t ep_addr) {

  return (ep[EP_NUM(ep_addr)].num_transferred_total);
}

/**
  \fn          int32_t USBD_EndpointTransferAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr  Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \return      \ref execution_status
*/
static int32_t USBD_EndpointTransferAbort (uint8_t ep_addr) {
  ENDPOINT_t *ptr_ep;
  uint8_t     ep_num, ep_dir;

  ep_num =  EP_NUM(ep_addr);
  ep_dir =  EP_DIR(ep_addr);
  ptr_ep = &ep[ep_num];

  if (hw_powered == false)            { return ARM_DRIVER_ERROR; }
  if (ep_num > USBD_MAX_ENDPOINT_NUM) { return ARM_DRIVER_ERROR; }
  if (ptr_ep->active[ep_dir] == 0U)   { return ARM_DRIVER_OK;    }

  if (ep_dir != 0U) {                   // IN Endpoint
    USBD_FlushTxFIFO (ep_num);          // Flush pending transfer
  }

  ptr_ep->active[ep_dir] = 0U;          // Transfer not active

  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t USBD_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t USBD_GetFrameNumber (void) {

  if (hw_powered == false) { return 0U; }

  return ((USBHS->USBHS_DEVFNUM & USBHS_DEVFNUM_FNUM_Msk) >> USBHS_DEVFNUM_FNUM_Pos);
}

/**
  \fn          void USBHS_Handler (void)
  \brief       USB Device Interrupt Routine (IRQ).
*/
void USBHS_Handler (void) {
  ENDPOINT_t *ptr_ep;
  uint32_t    devisr, deveptisr, ep_num, num_received, tout;
#if (defined(USB_DMA_USED) && (USB_DMA_USED == 1U))
  uint32_t    devdmastatus, dma_buf_len, num_remaining;
#if (USB_WT_RAM_EN == 1U)
  uint32_t    cache_addr, cache_len;
#endif
#endif

  devisr = USBHS->USBHS_DEVISR & USBHS->USBHS_DEVIMR;
  if (devisr == 0U) { return; }

  if ((devisr & USBHS_DEVISR_SOF) != 0U) {      // If SOF interrupt
    USBHS->USBHS_DEVICR = USBHS_DEVICR_SOFC;
  }
  if ((devisr & USBHS_DEVISR_MSOF) != 0U) {     // If micro SOF interrupt
    USBHS->USBHS_DEVICR = USBHS_DEVICR_MSOFC;
  }
  if ((devisr & USBHS_DEVISR_UPRSM) != 0U) {    // If upstream resume interrupt
    USBHS->USBHS_DEVICR = USBHS_DEVICR_UPRSMC;
  }
  if ((devisr & USBHS_DEVISR_SUSP) != 0U) {     // If suspend interrupt
    USBHS->USBHS_DEVIDR = USBHS_DEVIDR_SUSPEC;
    USBHS->USBHS_DEVIER = USBHS_DEVIER_WAKEUPES | USBHS_DEVIER_UPRSMES;
    usbd_state.active = 0U;
    SignalDeviceEvent(ARM_USBD_EVENT_SUSPEND);
    USBHS->USBHS_DEVICR = USBHS_DEVICR_SUSPC;
  }
  if ((devisr & USBHS_DEVISR_EORSM) != 0U) {    // If end of resume interrupt
    USBHS->USBHS_DEVIDR = USBHS_DEVIDR_WAKEUPEC | USBHS_DEVIDR_UPRSMEC;
    USBHS->USBHS_DEVIER = USBHS_DEVIER_SUSPES;
    usbd_state.active = 1U;
    SignalDeviceEvent(ARM_USBD_EVENT_RESUME);
    USBHS->USBHS_DEVICR = USBHS_DEVICR_EORSMC;
  }
  if ((devisr & USBHS_DEVISR_WAKEUP) != 0U) {   // If wake-up interrupt
    USBHS->USBHS_DEVIDR = USBHS_DEVIDR_WAKEUPEC | USBHS_DEVIDR_UPRSMEC;
    USBHS->USBHS_DEVIER = USBHS_DEVIER_SUSPES;
    SignalDeviceEvent(ARM_USBD_EVENT_RESUME);
    USBHS->USBHS_DEVICR = USBHS_DEVICR_WAKEUPC;
  }
  if ((devisr & USBHS_DEVISR_EORST) != 0U) {    // If USB end of reset interrupt
    USBD_Reset();
    SignalDeviceEvent(ARM_USBD_EVENT_RESET);
    usbd_state.active = 1U;
    if ((USBHS->USBHS_SR & USBHS_SR_SPEED_Msk) == USBHS_SR_SPEED_HIGH_SPEED) {
      usbd_state.speed  = ARM_USB_SPEED_HIGH;
      SignalDeviceEvent(ARM_USBD_EVENT_HIGH_SPEED);
    }
    USBHS->USBHS_DEVICR = USBHS_DEVICR_EORSTC;
  }

  if ((devisr & 0xFEFFF000U) != 0U) {
    for (ep_num = 0U; ep_num < (USBD_MAX_ENDPOINT_NUM + 1U); ep_num++) {
#if (defined(USB_DMA_USED) && (USB_DMA_USED == 1U))
      if ((devisr & ((1U << 24) << ep_num)) != 0U) {                            // If DMA interrupt
        ptr_ep    = (ENDPOINT_t *)(&ep[ep_num]);
        USBHS->USBHS_DEVDMA[ep_num-1U].USBHS_DEVDMACONTROL &= ~(USBHS_DEVDMACONTROL_END_B_EN  |
                                                                USBHS_DEVDMACONTROL_END_TR_EN);
        if (ptr_ep->active[1] != 0U) {                                          // If IN transfer is active
          devdmastatus = USBHS->USBHS_DEVDMA[ep_num - 1U].USBHS_DEVDMASTATUS;
          ptr_ep->num_transferred_total += ptr_ep->num_transferring - (devdmastatus >> 16);
          if (ptr_ep->num == ptr_ep->num_transferred_total) {                   // If sending done
            // If all required IN data was sent, set ARM_USBD_EP_EVENT_IN
            ptr_ep->active[1] = 0U;                                             // IN not active any more
            SignalEndpointEvent(ep_num | 0x80U, ARM_USBD_EVENT_IN);             // Send IN event
          } else {                                                              // If continue sending
            // If transfer is not yet finished
            num_remaining = ptr_ep->num - ptr_ep->num_transferred_total;
            if (num_remaining >= 65536U) {
              dma_buf_len = 0U;
              ptr_ep->num_transferring = 65536U;
            } else {
              dma_buf_len = num_remaining;
              ptr_ep->num_transferring = num_remaining;
            }
            USBHS->USBHS_DEVDMA[ep_num-1U].USBHS_DEVDMACONTROL = USBHS_DEVDMACONTROL_BUFF_LENGTH(dma_buf_len) |
                                                                 USBHS_DEVDMACONTROL_END_BUFFIT               |
                                                                 USBHS_DEVDMACONTROL_END_TR_IT                |
                                                                 USBHS_DEVDMACONTROL_END_B_EN                 |
                                                                 USBHS_DEVDMACONTROL_END_TR_EN                |
                                                                 USBHS_DEVDMACONTROL_CHANN_ENB                ;
          }
        }
        if (ptr_ep->active[0] != 0U) {                                          // If OUT transfer is active
          devdmastatus = USBHS->USBHS_DEVDMA[ep_num - 1U].USBHS_DEVDMASTATUS;
          num_received = ptr_ep->num_transferring - (devdmastatus >> 16);
          ptr_ep->num_transferred_total += num_received;
          if (((num_received & (ptr_ep->ep_max_packet_size - 1U)) != 0U) || (num_received == 0U) || (ptr_ep->num == ptr_ep->num_transferred_total)){
            // If all OUT data was received (required size or data terminated with ZLP or short packet), set ARM_USBD_EP_EVENT_OUT
            ptr_ep->active[0] = 0U;                                             // OUT not active any more
#if (USB_WT_RAM_EN == 1U)
            if ((SCB->CCR & SCB_CCR_DC_Msk) != 0U) {          // If Data Cache is enabled
              cache_addr = (uint32_t)ptr_ep->data & (~0x1FU);
              cache_len  = (cache_addr - (uint32_t)ptr_ep->data) + ptr_ep->num_transferred_total;
              SCB_InvalidateDCache_by_Addr((uint32_t*)(cache_addr), cache_len);
            }
#endif
            SignalEndpointEvent(ep_num, ARM_USBD_EVENT_OUT);                    // Send OUT event
          } else {
            // If more data should be received
            num_remaining = ptr_ep->num - ptr_ep->num_transferred_total;
            if (num_remaining >= 65536U) {
              dma_buf_len = 0U;
              ptr_ep->num_transferring = 65536U;
            } else {
              dma_buf_len = num_remaining;
              ptr_ep->num_transferring = num_remaining;
            }
            USBHS->USBHS_DEVDMA[ep_num-1U].USBHS_DEVDMACONTROL = USBHS_DEVDMACONTROL_BUFF_LENGTH(dma_buf_len) |
                                                                 USBHS_DEVDMACONTROL_END_BUFFIT               |
                                                                 USBHS_DEVDMACONTROL_END_TR_IT                |
                                                                 USBHS_DEVDMACONTROL_END_B_EN                 |
                                                                 USBHS_DEVDMACONTROL_END_TR_EN                |
                                                                 USBHS_DEVDMACONTROL_CHANN_ENB                ;
          }
        }
      }
#endif
      if ((devisr & (USBHS_DEVISR_PEP_0 << ep_num)) != 0U) {
        deveptisr = USBHS->USBHS_DEVEPTISR[ep_num] & USBHS->USBHS_DEVEPTIMR[ep_num];
        ptr_ep    = (ENDPOINT_t *)(&ep[ep_num]);
        if ((deveptisr & USBHS_DEVEPTISR_STALLEDI) != 0U) {
          USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_STALLEDIC;
          USBHS->USBHS_DEVEPTIER[ep_num] = USBHS_DEVEPTIER_STALLRQS;
        }
        if ((deveptisr & USBHS_DEVEPTISR_TXINI) != 0U) {                        // If transmit done interrupt
          if (ptr_ep->active[1] != 0U) {                                        // If IN transfer is active
            ptr_ep->num_transferred_total += ptr_ep->num_transferring;
            if (ptr_ep->num == ptr_ep->num_transferred_total) {                 // If sending done
              // If all required IN data was sent, set ARM_USBD_EP_EVENT_IN
              ptr_ep->active[1] = 0U;                                           // IN not active any more
              USBHS->USBHS_DEVEPTIDR[ep_num]  =  USBHS_DEVEPTIDR_TXINEC;        // Disable TXINI interrupt
              SignalEndpointEvent(ep_num | 0x80U, ARM_USBD_EVENT_IN);           // Send IN event
            } else {                                                            // If continue sending
              // If transfer is not yet finished
              if (ep_num != 0U) { USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_TXINIC;   }
              ptr_ep->num_transferring = USBD_WriteToFIFO(ep_num, ptr_ep->data + ptr_ep->num_transferred_total, ptr_ep->num - ptr_ep->num_transferred_total);
              if (ep_num == 0U) { USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_TXINIC;   }
              if (ep_num != 0U) { USBHS->USBHS_DEVEPTIDR[ep_num] = USBHS_DEVEPTIDR_FIFOCONC; }
            }
          }
        }
        if ((deveptisr & USBHS_DEVEPTISR_RXOUTI) != 0U) {                       // If receive done interrupt
          if (ptr_ep->active[0] != 0U) {                                        // If OUT transfer is active
            if (ep_num != 0U) { USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_RXOUTIC; }
            tout = SystemCoreClock;
            while (((USBHS->USBHS_DEVEPTISR[ep_num] & USBHS_DEVEPTISR_NBUSYBK_Msk) == 0U) && (tout-- != 0U));
            num_received = (USBHS->USBHS_DEVEPTISR[ep_num] & USBHS_DEVEPTISR_BYCT_Msk) >> USBHS_DEVEPTISR_BYCT_Pos;
            if (num_received != 0U) {
              if (num_received > ptr_ep->num) {
                num_received = ptr_ep->num;
              }
              ptr_ep->num_transferred_total += USBD_ReadFromFIFO (ep_num, ptr_ep->data + ptr_ep->num_transferred_total, num_received);
            }
            if (ep_num != 0U) { USBHS->USBHS_DEVEPTIDR[ep_num] = USBHS_DEVEPTIDR_FIFOCONC; }
            if (((num_received & (ptr_ep->ep_max_packet_size - 1U)) != 0U) || (num_received == 0U) || (ptr_ep->num == ptr_ep->num_transferred_total)){
              // If all OUT data was received (required size or data terminated with ZLP or short packet), set ARM_USBD_EP_EVENT_OUT
              ptr_ep->active[0] = 0U;                                           // OUT not active any more
              USBHS->USBHS_DEVEPTIDR[ep_num] = USBHS_DEVEPTIDR_RXOUTEC;         // Disable RXOUTI interrupt
              SignalEndpointEvent(ep_num, ARM_USBD_EVENT_OUT);                  // Send OUT event
            }
            if (ep_num == 0U) { USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_RXOUTIC;  }
          }
        }
        if ((deveptisr & USBHS_DEVEPTISR_RXSTPI) != 0U) {                       // If SETUP packed received interrupt
          ptr_ep->active[0] = 0U;
          ptr_ep->active[1] = 0U;
          setup_received    = true;
          USBD_ReadFromFIFO  (ep_num, (uint8_t *)setup_packet, 8);
          SignalEndpointEvent(ep_num | 0x80U, ARM_USBD_EVENT_SETUP);
          USBHS->USBHS_DEVEPTICR[ep_num] = USBHS_DEVEPTICR_RXSTPIC;
        }
      }
    }
  }
}

ARM_DRIVER_USBD Driver_USBD0 = {
  USBD_GetVersion,
  USBD_GetCapabilities,
  USBD_Initialize,
  USBD_Uninitialize,
  USBD_PowerControl,
  USBD_DeviceConnect,
  USBD_DeviceDisconnect,
  USBD_DeviceGetState,
  USBD_DeviceRemoteWakeup,
  USBD_DeviceSetAddress,
  USBD_ReadSetupPacket,
  USBD_EndpointConfigure,
  USBD_EndpointUnconfigure,
  USBD_EndpointStall,
  USBD_EndpointTransfer,
  USBD_EndpointTransferGetResult,
  USBD_EndpointTransferAbort,
  USBD_GetFrameNumber
};
