/* -----------------------------------------------------------------------------
 * Copyright (c) 2016 ARM Ltd.
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software. Permission is granted to anyone to use this
 * software for any purpose, including commercial applications, and to alter
 * it and redistribute it freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software in
 *    a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 *
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 3. This notice may not be removed or altered from any source distribution.
 *
 * $Date:        04. January 2016
 * $Revision:    V1.1
 *
 * Project:      CMSIS USB High-Speed Host Driver for Atmel SAM-ESV7
 * --------------------------------------------------------------------------
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 *
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------------- */
 
#include "USBH_SAM-ESV7.h"

// USBD Driver *****************************************************************

#define ARM_USBH_DRIVER_VERSION ARM_DRIVER_VERSION_MAJOR_MINOR(1,1) // USBD driver version

// Driver Version
static const ARM_DRIVER_VERSION usbh_driver_version = { ARM_USBH_API_VERSION, ARM_USBH_DRIVER_VERSION };

// Driver Capabilities
static const ARM_USBH_CAPABILITIES usbh_driver_capabilities = {
  0x0001, /* Root HUB available Ports Mask   */
  false,  /* Automatic SPLIT packet handling */
  true,   /* Signal Connect event            */
  true,   /* Signal Disconnect event         */
  false   /* Signal Overcurrent event        */
};

// Number of Endpoints
#ifndef USBH_PIPE_NUM
#define USBH_PIPE_NUM             CHIP_USB_NUMPIPE
#endif

// RAM Address
#ifndef USBHS_RAM_ADDR         
#define USBHS_RAM_ADDR         (0xA0100000U)
#endif

// Pipe handle mask
#define  USBH_PIPE_HANDLE        0x40

// Local variables and structures
static ARM_USBH_SignalPortEvent_t signal_port_event;
static ARM_USBH_SignalPipeEvent_t signal_pipe_event;
static uint32_t controlpipe_enabled;
static uint8_t drv_state = {0};
ARM_USBH_PORT_STATE port_state = {0}; 
const Pin USBH_VBUS_PORT[] = RTE_USBH_VBUS_PORT;

// Pipe resources structure
typedef struct {
  uint32_t bank;
  uint32_t ep_type;
  uint32_t ep_dir;
  uint32_t ep_addr;
  uint32_t dev_addr;
  uint32_t ep_size;
  uint32_t ep_interval;
  uint32_t num;
  uint8_t* buf;
  uint32_t buf_size;
  uint8_t  periodic_start;
} PIPE;

// Pipe resources
static PIPE pipe[USBH_PIPE_NUM];


// USB HAL FUNCTIONS

/**
  \fn          sysclk_enable_usb(void)
  \brief       Enable USB clock
*/
void sysclk_enable_usb(void)
{
  /* Disable FS USB clock */
  PMC->PMC_SCDR = PMC_SCDR_USBCLK;    
  /* Enable PLL 480 MHz */
  PMC->CKGR_UCKR = CKGR_UCKR_UPLLEN | CKGR_UCKR_UPLLCOUNT(0xF);    
  /* Wait that PLL is considered locked by the PMC */
  while( !(PMC->PMC_SR & PMC_SR_LOCKU) );    
  /* USB clock register: USB Clock Input is UTMI PLL */
  PMC->PMC_USB = (PMC_USB_USBS | PMC_USB_USBDIV(10 - 1) );    
  PMC->PMC_SCER = PMC_SCER_USBCLK;
}

/**
  \fn          sysclk_disable_usb(void)
  \brief       Disable USB clock
*/
void sysclk_disable_usb(void)
{
  /* Disable FS USB clock*/
  PMC->PMC_SCDR = PMC_SCDR_USBCLK;    
  /* Enable PLL 480 MHz */
  PMC->CKGR_UCKR = CKGR_UCKR_UPLLEN | CKGR_UCKR_UPLLCOUNT(0xF);    
  /* Wait that PLL is considered locked by the PMC */
  while( !(PMC->PMC_SR & PMC_SR_LOCKU) );    
  /* USB clock register: USB Clock Input is UTMI PLL */
  PMC->PMC_USB = (PMC_USB_USBS | PMC_USB_USBDIV(10 - 1) );   
}

/**
  \fn          USBH_DisableUsb(void)
  \brief       Disable USB peripheral
*/
static void USBH_DisableUsb(void)
{
  irqflags_t flags;

  // Check USB clock ready after a potential sleep mode < IDLE
  while (!USBHS_ISUsableClock(USBHS));
  USBHS_UnFreezeClock(USBHS);
    
  // Disable interrupts - connection, disconnection, SOF and reset
  USBHS_HostIntDisable(USBHS, (USBHS_HSTICR_DCONNIC | USBHS_HSTICR_DDISCIC
      | USBHS_HSTICR_HSOFIC  | USBHS_HSTICR_HWUPIC
      | USBHS_HSTICR_RSMEDIC | USBHS_HSTICR_RSTIC
      | USBHS_HSTICR_RXRSMIC));
  USBHS_DisableSOF();
  USBHS_Ack(USBHS,(0x1U << 9U)); // USBHS_SR_VBUSRQ
  
  // Send disconnection event
  port_state.connected = false;
  signal_port_event(0, ARM_USBH_EVENT_DISCONNECT);

  // Disable USB peripheral
  flags = cpu_irq_save();
  USBHS_FreezeClock(USBHS);
  USBHS_UsbEnable(USBHS, false);
  sysclk_disable_usb();
  PMC_DisablePeripheral(ID_USBHS);
  cpu_irq_restore(flags);
}

/**
  \fn          USBH_EnableUsbHost(void)
  \brief       Enable USB peripheral
*/
static  void USBH_EnableUsbHost(void)
{
  irqflags_t flags;
    
  // To avoid USB interrupt before end of initialization
  flags = cpu_irq_save();

  //* Enable USB hardware clock
  sysclk_enable_usb();
  PMC_EnablePeripheral(ID_USBHS);

  // Enable USB interrupt
  NVIC_EnableIRQ((IRQn_Type) ID_USBHS);
  // Configure VBUS port
  PIO_Configure(USBH_VBUS_PORT, 1); 
  // Set VBUS off  
  PIO_Set(USBH_VBUS_PORT);  
  // Set host mode
  USBHS_UsbMode(USBHS, HOST_MODE);     
  USBHS_UsbEnable(USBHS, true);

#if (RTE_USBH_HIGH_SPEED==0)
  USBHS->USBHS_HSTCTRL |= USBHS_HSTCTRL_SPDCONF_LOW_POWER;
#endif

  // Check USB clock
  USBHS_UnFreezeClock(USBHS);
  while (!USBHS_ISUsableClock(USBHS)){}
    
  // Clear all interrupts that may have been set by a previous host mode
  USBHS_ClearHostStatus(USBHS, (USBHS_HSTICR_DCONNIC | USBHS_HSTICR_DDISCIC
      | USBHS_HSTICR_HSOFIC  | USBHS_HSTICR_HWUPIC
      | USBHS_HSTICR_RSMEDIC | USBHS_HSTICR_RSTIC
      | USBHS_HSTICR_RXRSMIC));
  memory_sync();

  // Host wakeup interrupt enable
  USBHS_HostIntEnable(USBHS, USBHS_HSTIER_HWUPIES);
  // Set VBUS on
  PIO_Clear(USBH_VBUS_PORT);   
    
  // Enable main control interrupt
  // Connection, SOF and reset
  USBHS_HostIntEnable(USBHS, (USBHS_HSTIER_RSTIES | USBHS_HSTIER_HSOFIES | USBHS_HSTIER_DCONNIES));
  USBHS_FreezeClock(USBHS);     
  cpu_irq_restore(flags);
}

/**
  \fn          USBH_PipeDmaTransfer(void)
  \param[in]   pipe_hndl    Pipe Handle
  \brief       Handle the DMA pipe transfer 
*/
static void USBH_PipeDmaTransfer(ARM_USBH_PIPE_HANDLE pipe_hndl)
{
  irqflags_t flags;
  uint32_t dma_ctrl = 0;
  uint32_t max_trans;
  uint32_t next_trans;

  if (pipe[pipe_hndl].num != pipe[pipe_hndl].buf_size) {
    // Need to send or receive other data
    next_trans = pipe[pipe_hndl].buf_size - pipe[pipe_hndl].num;
    max_trans = UHD_PIPE_MAX_TRANS;
    if (USBHS_HostGetToken(USBHS, pipe_hndl) == USBHS_HSTPIPCFG_PTOKEN_IN) {
      // 256 is the maximum of IN requests via UPINRQ
      if ((256L*USBHS_HostGetSize(USBHS, pipe_hndl)) < UHD_PIPE_MAX_TRANS) {
        max_trans = 256L * USBHS_HostGetSize(USBHS, pipe_hndl);
      }
    }
    if (max_trans < next_trans) {
      // The USB hardware supports a maximum
      // transfer size of UHD_PIPE_MAX_TRANS Bytes
      next_trans = max_trans;
    }
    if (next_trans == UHD_PIPE_MAX_TRANS) {
      // Set 0 to transfer the maximum
      dma_ctrl = USBHS_HSTDMACONTROL_BUFF_LENGTH(0);
    } else {
      dma_ctrl = USBHS_HSTDMACONTROL_BUFF_LENGTH(next_trans);
    }
    if (USBHS_HSTPIPCFG_PTOKEN_OUT == USBHS_HostGetToken(USBHS, pipe_hndl)) {
      if (0 != (next_trans % USBHS_HostGetSize(USBHS, pipe_hndl))) {
        // Enable short packet option
        // else the DMA transfer is accepted
        // and interrupt DMA valid but nothing is sent.
        dma_ctrl |= USBHS_HSTDMACONTROL_END_B_EN;
      }
    } else {
      if ((USBHS_HostGetPipeType(USBHS, pipe_hndl) != ARM_USB_ENDPOINT_ISOCHRONOUS)
        || ((int)next_trans <= USBHS_HostGetSize(USBHS, pipe_hndl))) {
        // Enable short packet reception
        dma_ctrl |= USBHS_HSTDMACONTROL_END_TR_IT | USBHS_HSTDMACONTROL_END_TR_EN;
      }
    }     
    // Start USB DMA to fill or read fifo of the selected endpoint
    USBHS_SetHostDmaBuffAdd(&USBHS->USBHS_HSTDMA[pipe_hndl-1], (uint32_t) &pipe[pipe_hndl].buf[pipe[pipe_hndl].num]);
    dma_ctrl |= USBHS_HSTDMACONTROL_END_BUFFIT |
        USBHS_HSTDMACONTROL_CHANN_ENB;
    // Disable IRQs to have a short sequence
    // between read of EOT_STA and DMA enable
    flags = cpu_irq_save();
    if (!(USBHS_GetHostPipeDmaStatus(&USBHS->USBHS_HSTDMA[pipe_hndl-1])
        & USBHS_HSTDMASTATUS_END_TR_ST)) {
      if (USBHS_HSTPIPCFG_PTOKEN_IN == USBHS_HostGetToken(USBHS, pipe_hndl)) {
        uint32_t pipe_size = USBHS_HostGetSize(USBHS, pipe_hndl);
        USBHS_HostInReq(USBHS, pipe_hndl, (next_trans + pipe_size-1) / pipe_size);
      }
      if (!pipe[pipe_hndl].periodic_start) {
        USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_NBUSYBKEC);
        USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_PFREEZEC);
      } else {
        // Last bank not sent, just start
        if (USBHS_IsHostPipeIntTypeEnable(USBHS, pipe_hndl, USBHS_HSTPIPIMR_NBUSYBKE)) {
          USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPIDR_NBUSYBKEC | USBHS_HSTPIPIDR_PFREEZEC));
        } 
      }
            SCB_CleanInvalidateDCache();
      USBHS_HostConfigureDma(&USBHS->USBHS_HSTDMA[pipe_hndl-1], dma_ctrl);
      pipe[pipe_hndl].num += next_trans;
      cpu_irq_restore(flags);
      return;
    }
    cpu_irq_restore(flags);
    // Here a ZLP has been received
    // and the DMA transfer must be not started.
    // It is the end of transfer
    pipe[pipe_hndl].buf_size = pipe[pipe_hndl].num;
  }  
  // Transfer complete
  signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_TRANSFER_COMPLETE);
}

/**
  \fn          USBH_ConfigurePipe(void)
  \brief       Configure pipe
  \param[in]   pipe_hndl    Pipe Handle
  \return      \ref execution_status
*/
static int32_t USBH_ConfigurePipe (ARM_USBH_PIPE_HANDLE pipe_hndl) {    
  // Configure HSTPIPCFG
  USBHS_HostConfigure(USBHS, pipe_hndl, pipe[pipe_hndl].bank, 
                                        pipe[pipe_hndl].ep_size, 
                                        pipe[pipe_hndl].ep_type, 
                                        pipe[pipe_hndl].ep_dir, 
                                        pipe[pipe_hndl].ep_addr, 
                                        pipe[pipe_hndl].ep_interval);    
  if (pipe[pipe_hndl].ep_type==USBHS_HSTPIPCFG_PTYPE_CTRL) {
    // For control pipe disable automatic bank switching
    USBHS_HostDisableAutoSw(USBHS, pipe_hndl);
  } else {
    // Enable automatic bank switching and DMA interrupt
    USBHS_HostEnableAutoSw(USBHS, pipe_hndl);    
    USBHS_HostDmaIntEnable(USBHS, pipe_hndl-1);
  }
  // Allocate pipe memory
  USBHS_HostAllocMem(USBHS, pipe_hndl);
  if (!USBHS_IsHostConfigOk(USBHS, pipe_hndl)) {
    USBHS_HostPipeDisable(USBHS, pipe_hndl);
    return ARM_DRIVER_ERROR;
  }  
  // Set address
  USBHS_HostSetAddr(USBHS, pipe_hndl, pipe[pipe_hndl].dev_addr);  
  // Enable pipe
  USBHS_HostPipeEnable(USBHS, pipe_hndl);   
  // Enable endpoint interrupts
  USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPIER_RXSTALLDES | USBHS_HSTPIPIER_PERRES) );    
  USBHS_HostPipeIntEnable(USBHS, pipe_hndl);     
  return ARM_DRIVER_OK;
}

/**
  \fn          void USBH_ConfigurePacketSize (ARM_USBH_PIPE_HANDLE pipe_hndl, uint16_t ep_max_packet_size) 
  \brief       Configure pipe's maximum packet size
  \param[in]   pipe_hndl          Pipe Handle
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
*/
static void USBH_ConfigurePacketSize (ARM_USBH_PIPE_HANDLE pipe_hndl, uint16_t ep_max_packet_size) 
{
  if(ep_max_packet_size <= 8 )
      pipe[pipe_hndl].ep_size = 0;
  else if (ep_max_packet_size <= 16)
      pipe[pipe_hndl].ep_size = 1;
  else if (ep_max_packet_size <= 32)
      pipe[pipe_hndl].ep_size = 2;
  else if (ep_max_packet_size <= 64)
      pipe[pipe_hndl].ep_size = 3;
  else if (ep_max_packet_size <= 128)
      pipe[pipe_hndl].ep_size = 4;
  else if (ep_max_packet_size <= 256)
      pipe[pipe_hndl].ep_size = 5;
  else if (ep_max_packet_size <= 512)
      pipe[pipe_hndl].ep_size = 6;
  else if (ep_max_packet_size <= 1024)
      pipe[pipe_hndl].ep_size = 7;
}

/**
  \fn          USBH_ControlSetupPacket(void)
  \param[in]   pipe_hndl    Pipe Handle
  \param[in]   data         Pointer to data buffer
  \param[in]   num          Number of bytes to transfer
  \brief       Start transfering a setup packet through a control pipe  
*/
static uint32_t USBH_FillFifo(uint8_t *data, uint32_t num)
{
  volatile uint8_t *pEpData; 

  // Fill fifo
  pEpData = (volatile uint8_t*)USBHS_RAM_ADDR;
  memory_sync();
  while (num-->0)
    *pEpData++ = *data++;
  memory_barrier();
  
  return ARM_DRIVER_OK;  
}

/**
  \fn          USBH_InPacketReceived(void)
  \param[in]   pipe_hndl    Pipe Handle
  \brief       Get IN packet data  
*/
static void USBH_InPacketReceived(ARM_USBH_PIPE_HANDLE pipe_hndl)
{
  uint8_t *pEpData;
  uint8_t nb = USBHS_HostGetPipeByteCount(USBHS, pipe_hndl);
  bool b_short_packet;
  
  // Get information to read data
  pEpData =  (uint8_t*)((uint32_t*)USBHS_RAM_ADDR + (EPT_VIRTUAL_SIZE * pipe_hndl));
  memory_sync();  
  b_short_packet = (USBHS_HostGetIntTypeStatus(USBHS, pipe_hndl, USBHS_HSTPIPISR_SHORTPACKETI) == USBHS_HSTPIPISR_SHORTPACKETI)? true:false;

  // Read fifo
  while (nb-->0) {
    *pipe[pipe_hndl].buf++ = *pEpData++;
    pipe[pipe_hndl].num++;
  }  
         
  if ((pipe[pipe_hndl].num < pipe[pipe_hndl].buf_size)&&(!b_short_packet)) {
    // Send a new IN packet request
    USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_RXINES);
    // Ack fifo and unfreeze pipe
    USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPIDR_FIFOCONC | USBHS_HSTPIPIDR_PFREEZEC )); 
  } else {
    // Transfer complete
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_TRANSFER_COMPLETE);
  }
}

/**
  \fn          USBH_ControlInterrupt(void)
  \param[in]   pipe_hndl    Pipe Handle
  \brief       Interrupt service routine for control pipes  
*/
static void USBH_ControlInterrupt(ARM_USBH_PIPE_HANDLE pipe_hndl)
{
  // Disable setup, IN and OUT interrupts of control endpoint
  USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, (  USBHS_HSTPIPIDR_TXSTPEC
                                            | USBHS_HSTPIPIDR_RXINEC
                                            | USBHS_HSTPIPIDR_TXOUTEC) );

  // Search event on control endpoint
  if (USBHS_HostGetIntTypeStatus(USBHS, pipe_hndl, USBHS_HSTPIPISR_TXSTPI) ) 
    {
    // SETUP packet sent
    USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_PFREEZES);
    USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_TXSTPIC);    
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_TRANSFER_COMPLETE);  
    return;
  }
  if (USBHS_HostGetIntTypeStatus(USBHS, pipe_hndl,USBHS_HSTPIPISR_RXINI)) {
    // In case of low USB speed and with a high CPU frequency,
    // an ACK from host can be always running on USB line
    // then wait end of ACK on IN pipe.
    while (!USBHS_IsHostPipeIntTypeEnable(USBHS, pipe_hndl, USBHS_HSTPIPIMR_PFREEZE)){}
    // IN packet received, Acknowledge
    USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_RXINIC);
    USBH_InPacketReceived(pipe_hndl);
    return;
  }
  if (USBHS_HostGetIntTypeStatus(USBHS, pipe_hndl,USBHS_HSTPIPISR_TXOUTI)) {
    // OUT packet sent
    USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_PFREEZES);
    USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_TXOUTIC);
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_TRANSFER_COMPLETE);
    return;
  }  
  if (USBHS_HostGetIntTypeStatus(USBHS, 0,USBHS_HSTPIPISR_RXSTALLDI)) {
    // Stall Handshake received
    USBHS_HostAckPipeIntType(USBHS, 0, USBHS_HSTPIPICR_RXSTALLDIC);   
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_HANDSHAKE_STALL);  
    return;
  }  
  if (USBHS_HostGetIntTypeStatus(USBHS, 0,USBHS_HSTPIPISR_PERRI)) {
    // Error event
    USBHS_HostClearErr(USBHS, pipe_hndl, 0xFF);
    USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_TXOUTEC);
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_BUS_ERROR);
    return;
  }
}

/**
  \fn          USBH_PipeInterrupt(void)
  \param[in]   pipe_hndl    Pipe Handle
  \brief       Interrupt service routine for other pipes  
*/
static void USBH_PipeInterrupt(ARM_USBH_PIPE_HANDLE pipe_hndl)
{
  uint32_t status = USBHS_HostGetIntTypeStatus(USBHS, pipe_hndl, 0xFFFFFFFF);
  uint32_t statusInt = status & USBHS_IsHostPipeIntTypeEnable(USBHS, pipe_hndl, 0xFFFFFFFF);  
 
  // for DMA endpoints
  if (USBHS_IsHostPipeIntTypeEnable(USBHS, pipe_hndl, USBHS_HSTPIPIMR_NBUSYBKE) && (USBHS_HostGetNumOfBusyBank(USBHS, pipe_hndl)==0)) {
    USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_PFREEZES);
    USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_NBUSYBKEC);
    // For ISO, no ACK, finished when DMA done    
    if (!(USBHS_HostGetPipeType(USBHS, pipe_hndl) == USBHS_HSTPIPCFG_PTYPE_ISO)) {
      signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_TRANSFER_COMPLETE);
    }    
    return;    
  }
  if (status & USBHS_HSTPIPISR_RXSTALLDI) {
    USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_RXSTALLDIC);
    USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_RSTDTS);
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_HANDSHAKE_STALL);      
    return;
  }
  if (status & USBHS_HSTPIPISR_PERRI) {
    // Error event
    USBHS_HostClearErr(USBHS, pipe_hndl, 0xFF);
    USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_TXOUTEC);
    signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_BUS_ERROR);
    return;
  }
}

/**
  \fn          USBH_PipeDmaInterrupt(void)
  \param[in]   pipe_hndl    Pipe Handle
  \brief       Interrupt service routine for DMA transfers  
*/
static void USBH_PipeDmaInterrupt(ARM_USBH_PIPE_HANDLE pipe_hndl)
{
  uint32_t nb_remaining;

  SCB_CleanInvalidateDCache();
  // Ignore EOT_STA interrupt
  if (USBHS_GetHostPipeDmaStatus(&USBHS->USBHS_HSTDMA[pipe_hndl-1]) & USBHS_HSTDMASTATUS_CHANN_ENB) return;
  // Save number of data no transfered
  nb_remaining = ((USBHS_GetHostPipeDmaStatus(&USBHS->USBHS_HSTDMA[pipe_hndl-1]) & USBHS_HSTDMASTATUS_BUFF_COUNT_Msk) >> USBHS_HSTDMASTATUS_BUFF_COUNT_Pos);
  memory_sync();
  
  if (nb_remaining) {
    // Transfer no complete (short packet or ZLP) then
    // Update number of transferred data
    pipe[pipe_hndl].num -= nb_remaining;
    // Set transfer complete to stop the transfer
    pipe[pipe_hndl].buf_size = pipe[pipe_hndl].num;
  }

  if (USBHS_HSTPIPCFG_PTOKEN_OUT == USBHS_HostGetToken(USBHS, pipe_hndl)) {
    // Wait that all banks are free to freeze clock of OUT endpoint
    // and call callback
    USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_NBUSYBKES);
    // For ISO out, start another DMA transfer since no ACK needed
    if (USBHS_HostGetPipeType(USBHS, pipe_hndl) == USBHS_HSTPIPCFG_PTYPE_ISO) 
    {
      signal_pipe_event(pipe_hndl|USBH_PIPE_HANDLE, ARM_USBH_EVENT_TRANSFER_COMPLETE);
    }
  } else {
    if (!USBHS_IsHostPipeIntTypeEnable(USBHS, pipe_hndl, USBHS_HSTPIPIMR_PFREEZE)) {
      // Pipe is not freeze in case of :
      // - incomplete transfer when the request number INRQ is not complete.
      // - low USB speed and with a high CPU frequency,
      // a ACK from host can be always running on USB line.
      if (nb_remaining) {
        // Freeze pipe in case of incomplete transfer
        USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_PFREEZES);
      } else {
        // Wait freeze in case of ASK on going
        while (!USBHS_IsHostPipeIntTypeEnable(USBHS, pipe_hndl, USBHS_HSTPIPIMR_PFREEZE)) {}
      }
    }
    USBH_PipeDmaTransfer(pipe_hndl);
  }
}

/**
  \fn          USBHS_Handler(void)
  \param[in]   pipe_hndl    Pipe Handle
  \brief       Interrupt service routine entry point 
*/
void USBHS_Handler(void)
{
  uint8_t pipe_int, pipe_dma_int;
  uint8_t pipe_hndl;
  uint32_t status = USBHS_GetHostStatus(USBHS, 0xFF);
  uint32_t statusInt = (status & USBHS_IsHostIntEnable(USBHS, 0xFF));
    
  if (status & USBHS_HSTISR_HSOFI) 
  {
    //Acknowledge SOF
    USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_HSOFIC);
    for (pipe_hndl = 1; pipe_hndl < USBHS_EPT_NUM; pipe_hndl++) {
      if (pipe[pipe_hndl].periodic_start) {
        pipe[pipe_hndl].periodic_start = false;
        USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_PFREEZEC);
      }
    }
  } else {
    // Pipe/DMA interrupt
    pipe_int = USBHS_GetInterruptPipeNum();
    pipe_dma_int = USBHS_GetInterruptPipeDmaNum();
    
    SCB_CleanInvalidateDCache();

    if (pipe_int != USBHS_EPT_NUM) {
        if (pipe[pipe_int].ep_type == USBHS_HSTPIPCFG_PTYPE_CTRL) {
          // Interrupt acked by control endpoint managed
          USBH_ControlInterrupt(pipe_int);
        } else {        
          // Interrupt acked by bulk/interrupt/isochronous endpoint
          USBH_PipeInterrupt(pipe_int);
        }
    }    
    else if (pipe_dma_int != USBHS_EPT_NUM) {
        // Interrupt DMA acked by bulk/interrupt/isochronous endpoint
        USBH_PipeDmaInterrupt(pipe_dma_int);
    }
    // USB bus reset detection
    else if (status & USBHS_HSTISR_RSTI) 
    {
      if (USBHS_GetUsbSpeed(USBHS)!=USBHS_SR_SPEED_HIGH_SPEED) {
        USBHS->USBHS_HSTCTRL |= USBHS_HSTCTRL_SPDCONF_LOW_POWER;
      }      
      USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_RSTIC);
      // Reconfigure control pipe
      if (controlpipe_enabled) USBH_ConfigurePipe(0);
      // Send event
      signal_port_event(0, ARM_USBH_EVENT_RESET);
    }    
    // Disconnection event
    else if (statusInt & USBHS_HSTIMR_DDISCIE) {         
      // Reset device address            
      pipe[0].dev_addr = 0x00;            
      // Re-enable USB and send disconnection event
      if (USBHS_IsHostPipeEnable(USBHS, 0)) {
        USBH_DisableUsb();           
        USBH_EnableUsbHost();   
      } else {         
        USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_DDISCIC);
        USBHS_HostIntDisable(USBHS, USBHS_HSTIDR_DDISCIEC);
        USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_DCONNIC);
        USBHS_HostIntEnable(USBHS, USBHS_HSTIER_DCONNIES);
        USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_HWUPIC);
        USBHS_HostIntEnable(USBHS, USBHS_HSTIER_HWUPIES);        
      }
    }
    // Connection event
    else if (statusInt & USBHS_HSTIMR_DCONNIE) {        
        USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_DCONNIC);
        USBHS_HostIntDisable(USBHS,USBHS_HSTIDR_DCONNIEC);
        USBHS_ClearHostStatus(USBHS, USBHS_HSTICR_DDISCIC);
        USBHS_HostIntEnable(USBHS, USBHS_HSTIER_DDISCIES);
        USBHS_EnableSOF(USBHS);
        // Send event
        port_state.connected = true;
        signal_port_event(0, ARM_USBH_EVENT_CONNECT);     
    }
    // If wakeup interrupt is enabled and triggered and connection interrupt is enabled 
    else if((statusInt & USBHS_HSTISR_HWUPI) && USBHS_IsHostIntEnable(USBHS, USBHS_HSTIMR_DCONNIE))
    {        
        while (!USBHS_ISUsableClock(USBHS));
        USBHS_UnFreezeClock(USBHS);
        // Here the wakeup interrupt has been used to detect connection
        // with an asynchrone interrupt
        USBHS_HostIntDisable(USBHS, USBHS_HSTIDR_HWUPIEC);    
        USBHS_Set(USBHS, (0x1U << 9U)); // USBHS_SR_VBUSRQ
    }
    else if (USBHS_IsHostIntEnable(USBHS, USBHS_HSTIMR_HWUPIE) && 
            (USBHS_GetHostStatus(USBHS, (USBHS_HSTISR_HWUPI   | 
                                         USBHS_HSTISR_RSMEDI  | 
                                         USBHS_HSTISR_RXRSMI) ))) 
    {
      while (!USBHS_ISUsableClock(USBHS)){}
      USBHS_UnFreezeClock(USBHS);
      // Disable wakeup/resumes interrupts
      USBHS_HostIntDisable(USBHS, (USBHS_HSTIDR_HWUPIEC   |
                                   USBHS_HSTIDR_RSMEDIEC  |
                                   USBHS_HSTIDR_RXRSMIEC) );
      USBHS_EnableSOF(USBHS);
      if ((!(USBHS_HSTISR_RSMEDI & status))&&(!(USBHS_HSTISR_DDISCI & status))) {
        // In High speed mode a downstream resume must be sent
        // after a upstream to avoid a disconnection.
        if (USBHS_IsUsbHighSpeed(USBHS)) {
          USBHS_Resume();
        }
      }
    } else {
       status = status;
    }
  }   
}


// USB HOST DRIVER FUNCTIONS

/**
  \fn          ARM_DRIVER_VERSION ARM_USBH_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRIVER_VERSION
*/
static ARM_DRIVER_VERSION ARM_USBH_GetVersion (void) { return usbh_driver_version; }

/**
  \fn          ARM_USBH_CAPABILITIES ARM_USBH_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBH_CAPABILITIES
*/
static ARM_USBH_CAPABILITIES ARM_USBH_GetCapabilities (void) { return usbh_driver_capabilities; }

/**
  \fn          int32_t ARM_USBH_Initialize (ARM_USBH_SignalPortEvent_t cb_port_event,
                                            ARM_USBH_SignalPipeEvent_t cb_pipe_event)
  \brief       Initialize USB Host Interface.
  \param[in]   cb_port_event  Pointer to \ref ARM_USBH_SignalPortEvent
  \param[in]   cb_pipe_event  Pointer to \ref ARM_USBH_SignalPipeEvent
  \return      \ref execution_status
*/
static int32_t ARM_USBH_Initialize (ARM_USBH_SignalPortEvent_t cb_port_event, ARM_USBH_SignalPipeEvent_t cb_pipe_event) {

  if (drv_state & USBH_DRIVER_INITIALIZED) { return ARM_DRIVER_OK; }
  if (drv_state)                           { return ARM_DRIVER_ERROR; }                                                                

  // Initialize event callbacks
  signal_port_event = cb_port_event;
  signal_pipe_event = cb_pipe_event;
  
  // Initialize device state
  drv_state = USBH_DRIVER_INITIALIZED;
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_Uninitialize (void)
  \brief       De-initialize USB Host Interface.
  \return      \ref execution_status
*/
static int32_t ARM_USBH_Uninitialize (void) {

  if (!(drv_state & USBH_DRIVER_INITIALIZED)) { return ARM_DRIVER_OK; }

  // Clear device state
  drv_state = 0U;  
  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Host Interface Power.
  \param[in]   state  Power state
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PowerControl (ARM_POWER_STATE state) {
 
  switch (state) {
    case ARM_POWER_OFF:
      if (!(drv_state & USBH_DRIVER_POWERED)) { return ARM_DRIVER_OK; }      
      // Disable USB
      USBH_DisableUsb();     
      drv_state  &= ~USBH_DRIVER_POWERED;
      break;

    case ARM_POWER_FULL:
      if (!(drv_state & USBH_DRIVER_INITIALIZED)) { return ARM_DRIVER_ERROR; }
      if (drv_state & USBH_DRIVER_POWERED) { return ARM_DRIVER_OK; }      
      // Enable USB Host
      USBH_EnableUsbHost();      
      drv_state |= USBH_DRIVER_POWERED;
      break;
    
    default:
      return ARM_DRIVER_ERROR_UNSUPPORTED;
  }      
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PortVbusOnOff (uint8_t port, bool vbus)
  \brief       Root HUB Port VBUS on/off.
  \param[in]   port  Root HUB Port Number
  \param[in]   vbus
                - \b false VBUS off
                - \b true  VBUS on
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PortVbusOnOff (uint8_t port, bool vbus) {

  if (!(drv_state & USBH_DRIVER_POWERED)) return ARM_DRIVER_ERROR;
  if (port)                               return ARM_DRIVER_ERROR_PARAMETER;

  if (vbus) {  
    // VBUS power on     
    USBHS->USBHS_SFR |= USBHS_SFR_VBUSRQS;
    PIO_Clear(USBH_VBUS_PORT);
  } else {  
    // VBUS power off 
    USBHS->USBHS_SCR |= (0x1U << 9U); // USBHS_SCR_VBUSRQC
    PIO_Set(USBH_VBUS_PORT);
  }
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PortReset (uint8_t port)
  \brief       Do Root HUB Port Reset.
  \param[in]   port  Root HUB Port Number
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PortReset (uint8_t port) {
  controlpipe_enabled = USBHS_IsHostPipeEnable(USBHS, 0);
  // Send reset signal
  USBHS_Reset();
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PortSuspend (uint8_t port)
  \brief       Suspend Root HUB Port (stop generating SOFs).
  \param[in]   port  Root HUB Port Number
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PortSuspend (uint8_t port) {
  uint8_t pipe;
  
  for (pipe = 1; pipe < USBHS_EPT_NUM; pipe++) {
    USBHS_HostEnablePipeIntType(USBHS, pipe, USBHS_HSTPIPIER_PFREEZES);
  }
  USBHS_DisableSOF();  
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PortResume (uint8_t port)
  \brief       Resume Root HUB Port (start generating SOFs).
  \param[in]   port  Root HUB Port Number
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PortResume (uint8_t port) {
  uint8_t pipe;
  
  // Restore pipes unfreezed
  for (pipe = 1; pipe < USBHS_EPT_NUM; pipe++) {
    USBHS_HostDisablePipeIntType(USBHS, pipe, USBHS_HSTPIPIDR_PFREEZEC);
  }  
  // Check USB clock ready after a potential sleep mode < IDLE
  while (!USBHS_ISUsableClock(USBHS));
  USBHS_UnFreezeClock(USBHS);
  USBHS_EnableSOF(USBHS);
  USBHS_Resume();  
  // Send event
  signal_port_event(0, ARM_USBH_EVENT_RESUME);  
  return ARM_DRIVER_OK;
}

/**
  \fn          ARM_USBH_PORT_STATE ARM_USBH_PortGetState (uint8_t port)
  \brief       Get current Root HUB Port State.
  \param[in]   port  Root HUB Port Number
  \return      Port State \ref ARM_USBH_PORT_STATE
*/
static ARM_USBH_PORT_STATE ARM_USBH_PortGetState (uint8_t port) {
ARM_USBH_PORT_STATE state = {0U}; 

  if (!(drv_state & USBH_DRIVER_POWERED)) return state;
  if (port != 0U)                         return state;  

  port_state.overcurrent = 0U;
  switch (USBHS_GetUsbSpeed(USBHS)) {
    case USBHS_SR_SPEED_HIGH_SPEED: 
     // High speed                         
     port_state.speed = ARM_USB_SPEED_HIGH;
     break;
    case USBHS_SR_SPEED_FULL_SPEED: 
     // Full speed                               
     port_state.speed = ARM_USB_SPEED_FULL;
     break;
    case USBHS_SR_SPEED_LOW_SPEED:                       
    default:
     // Low speed                         
     port_state.speed = ARM_USB_SPEED_LOW;
     break;
  }    
  return port_state;
}

/**
  \fn          ARM_USBH_PIPE_HANDLE ARM_USBH_PipeCreate (uint8_t  dev_addr,
                                                         uint8_t  dev_speed,
                                                         uint8_t  hub_addr,
                                                         uint8_t  hub_port,
                                                         uint8_t  ep_addr,
                                                         uint8_t  ep_type,
                                                         uint16_t ep_max_packet_size,
                                                         uint8_t  ep_interval)
  \brief       Create Pipe in System.
  \param[in]   dev_addr   Device Address
  \param[in]   dev_speed  Device Speed
  \param[in]   hub_addr   Hub Address
  \param[in]   hub_port   Hub Port
  \param[in]   ep_addr    Endpoint Address
                - ep_addr.0..3: Address
                - ep_addr.7:    Direction
  \param[in]   ep_type    Endpoint Type (ARM_USB_ENDPOINT_xxx)
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \param[in]   ep_interval        Endpoint Polling Interval
  \return      Pipe Handle \ref ARM_USBH_PIPE_HANDLE
*/
static ARM_USBH_PIPE_HANDLE ARM_USBH_PipeCreate (uint8_t dev_addr, 
                                                 uint8_t dev_speed, 
                                                 uint8_t hub_addr, 
                                                 uint8_t hub_port, 
                                                 uint8_t ep_addr, 
                                                 uint8_t ep_type, 
                                                 uint16_t ep_max_packet_size, 
                                                 uint8_t  ep_interval) {
  ARM_USBH_PIPE_HANDLE pipe_hndl;

  // find first available pipe
  for (pipe_hndl = 0; pipe_hndl < USBHS_EPT_NUM; pipe_hndl++) {
    if (USBHS_IsHostPipeEnable(USBHS, pipe_hndl)) {
      continue;
    }
    USBHS_HostPipeEnable(USBHS, pipe_hndl);
    pipe[pipe_hndl].ep_dir = (ep_addr & 0x80)?(USBHS_HSTPIPCFG_PTOKEN_IN):(USBHS_HSTPIPCFG_PTOKEN_OUT);
    pipe[pipe_hndl].ep_addr = ep_addr;
    pipe[pipe_hndl].dev_addr = dev_addr;
    pipe[pipe_hndl].ep_interval = ep_interval;
    
    // Bank choice
    switch(ep_type) {
      case ARM_USB_ENDPOINT_ISOCHRONOUS:
        pipe[pipe_hndl].bank = USBHS_HSTPIPCFG_PBK_2_BANK;
        pipe[pipe_hndl].ep_type = USBHS_HSTPIPCFG_PTYPE_ISO;
        break;
      case ARM_USB_ENDPOINT_INTERRUPT:
        pipe[pipe_hndl].bank = USBHS_HSTPIPCFG_PBK_1_BANK;
        pipe[pipe_hndl].ep_type = USBHS_HSTPIPCFG_PTYPE_INTRPT;
        break;
      case ARM_USB_ENDPOINT_BULK:
        pipe[pipe_hndl].bank = USBHS_HSTPIPCFG_PBK_2_BANK;
        pipe[pipe_hndl].ep_type = USBHS_HSTPIPCFG_PTYPE_BLK;
        pipe[pipe_hndl].ep_interval = 0;
        break;
      case ARM_USB_ENDPOINT_CONTROL:
        pipe[pipe_hndl].bank = USBHS_HSTPIPCFG_PBK_1_BANK;
        pipe[pipe_hndl].ep_type = USBHS_HSTPIPCFG_PTYPE_CTRL;
        pipe[pipe_hndl].ep_dir = USBHS_HSTPIPCFG_PTYPE_CTRL;
        break;
    }

    // Configure pack size 
    USBH_ConfigurePacketSize(pipe_hndl, ep_max_packet_size);

    // Configure pipe    
    USBH_ConfigurePipe(pipe_hndl);     
    
    return pipe_hndl|USBH_PIPE_HANDLE;
  }
  
  // No pipe available
  return NULL;
}

/**
  \fn          int32_t ARM_USBH_PipeModify (ARM_USBH_PIPE_HANDLE pipe_hndl,
                                            uint8_t              dev_addr,
                                            uint8_t              dev_speed,
                                            uint8_t              hub_addr,
                                            uint8_t              hub_port,
                                            uint16_t             ep_max_packet_size)
  \brief       Modify Pipe in System.
  \param[in]   pipe_hndl  Pipe Handle
  \param[in]   dev_addr   Device Address
  \param[in]   dev_speed  Device Speed
  \param[in]   hub_addr   Hub Address
  \param[in]   hub_port   Hub Port
  \param[in]   ep_max_packet_size Endpoint Maximum Packet Size
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PipeModify (ARM_USBH_PIPE_HANDLE pipe_hndl, 
                                    uint8_t dev_addr, 
                                    uint8_t dev_speed, 
                                    uint8_t hub_addr, 
                                    uint8_t hub_port, 
                                    uint16_t ep_max_packet_size) {
  pipe_hndl &= ~USBH_PIPE_HANDLE;                                      
  // Set device address
  pipe[pipe_hndl].dev_addr = dev_addr;  
  // Configure pack size 
  USBH_ConfigurePacketSize(pipe_hndl, ep_max_packet_size);
  // Configure pipe   
  return USBH_ConfigurePipe(pipe_hndl);
}

/**
  \fn          int32_t ARM_USBH_PipeDelete (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Delete Pipe from System.
  \param[in]   pipe_hndl  Pipe Handle
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PipeDelete (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  pipe_hndl &= ~USBH_PIPE_HANDLE;
  // Disable pipe   
  USBHS_HostPipeDisable(USBHS, pipe_hndl);
  // Unallocate memory
  USBHS_HostFreeMem(USBHS, pipe_hndl);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PipeReset (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Reset Pipe.
  \param[in]   pipe_hndl  Pipe Handle
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PipeReset (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  pipe_hndl &= ~USBH_PIPE_HANDLE;
  // Reset pipe
  USBHS_HostPipeReset(USBHS,pipe_hndl);
  return ARM_DRIVER_OK;
}

/**
  \fn          int32_t ARM_USBH_PipeTransfer (ARM_USBH_PIPE_HANDLE pipe_hndl,
                                              uint32_t             packet,
                                              uint8_t             *data,
                                              uint32_t             num)
  \brief       Transfer packets through USB Pipe.
  \param[in]   pipe_hndl  Pipe Handle
  \param[in]   packet     Packet information
  \param[in]   data       Pointer to buffer with data to send or for data to receive
  \param[in]   num        Number of data bytes to transfer
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PipeTransfer (ARM_USBH_PIPE_HANDLE pipe_hndl, uint32_t packet, uint8_t *data, uint32_t num) {

  // Clear transfer related info
  pipe_hndl &= ~USBH_PIPE_HANDLE;
  pipe[pipe_hndl].num = 0;
  pipe[pipe_hndl].buf = data;
  pipe[pipe_hndl].buf_size = num;
  
  // Control pipe
  if (pipe_hndl==0) {
    switch (packet&ARM_USBH_PACKET_TOKEN_Msk) {    
      case ARM_USBH_PACKET_SETUP:
        USBHS_HostSetToken(USBHS, pipe_hndl, USBHS_HSTPIPCFG_PTOKEN_SETUP);        
        USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_TXSTPIC);         
        USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_TXSTPES);
        pipe[pipe_hndl].num = num;
        break;     
      case ARM_USBH_PACKET_OUT:   
        USBHS_HostSetToken(USBHS, pipe_hndl, USBHS_HSTPIPCFG_PTOKEN_OUT);      
        USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_TXOUTIC);        
        USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_TXOUTES);
        pipe[pipe_hndl].num = num; 
        break;     
      case ARM_USBH_PACKET_IN:   
        USBHS_HostSetToken(USBHS, pipe_hndl, USBHS_HSTPIPCFG_PTOKEN_IN);     
        USBHS_HostAckPipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPICR_RXINIC | USBHS_HSTPIPICR_SHORTPACKETIC) );
        USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_RXINES);  
        break;      
    }  
    // Fill and acknowledge FIFO, unfreeze pipe
    USBH_FillFifo(data, num); 
    USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPIDR_FIFOCONC | USBHS_HSTPIPIDR_PFREEZEC) );  
    return ARM_DRIVER_OK;  
  }

  // Other pipes
  switch (packet&ARM_USBH_PACKET_TOKEN_Msk) {
    case ARM_USBH_PACKET_OUT:   
      // Packet OUT
      USBHS_HostSetToken(USBHS, pipe_hndl, USBHS_HSTPIPCFG_PTOKEN_OUT);  
      // Zero length packet   
      if (num==0) {
        USBHS_HostAckPipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPICR_TXOUTIC);
        USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIER_TXOUTES);
        USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPIDR_FIFOCONC | USBHS_HSTPIPIDR_PFREEZEC)); 
        return ARM_DRIVER_OK;
      }
      // Interrupt or ISO pipe
      if (USBHS_HostGetPipeType(USBHS, pipe_hndl) & (USBHS_HSTPIPCFG_PTYPE_ISO | USBHS_HSTPIPCFG_PTYPE_INTRPT))
        pipe[pipe_hndl].periodic_start = true; 
      break;  
    case ARM_USBH_PACKET_IN:   
      // Packet IN
      USBHS_HostSetToken(USBHS, pipe_hndl, USBHS_HSTPIPCFG_PTOKEN_IN);     
      break;
  }
  
  // Request first transfer
  USBH_PipeDmaTransfer(pipe_hndl);

  return ARM_DRIVER_OK;
}

/**
  \fn          uint32_t ARM_USBH_PipeTransferGetResult (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Get result of USB Pipe transfer.
  \param[in]   pipe_hndl  Pipe Handle
  \return      number of successfully transferred data bytes
*/
static uint32_t ARM_USBH_PipeTransferGetResult (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  if (!pipe_hndl) return 0;
  pipe_hndl &= ~USBH_PIPE_HANDLE;
  return pipe[pipe_hndl].num;
}

/**
  \fn          int32_t ARM_USBH_PipeTransferAbort (ARM_USBH_PIPE_HANDLE pipe_hndl)
  \brief       Abort current USB Pipe transfer.
  \param[in]   pipe_hndl  Pipe Handle
  \return      \ref execution_status
*/
static int32_t ARM_USBH_PipeTransferAbort (ARM_USBH_PIPE_HANDLE pipe_hndl) {
  pipe_hndl &= ~USBH_PIPE_HANDLE;
  USBHS_HostPipeReset(USBHS, pipe_hndl);
  // Autoswitch bank and interrupts has been reseted, then re-enable it
  USBHS_HostEnableAutoSw(USBHS, pipe_hndl);
  USBHS_HostEnablePipeIntType(USBHS, pipe_hndl, (USBHS_HSTPIPIER_RXSTALLDES | USBHS_HSTPIPIER_PERRES) );
  USBHS_HostDisablePipeIntType(USBHS, pipe_hndl, USBHS_HSTPIPIDR_TXOUTEC);
  USBHS_HostConfigureDma(&USBHS->USBHS_HSTDMA[pipe_hndl-1], 0); 
  return ARM_DRIVER_OK;
}

/**
  \fn          uint16_t ARM_USBH_GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t ARM_USBH_GetFrameNumber (void) {
  return ((USBHS->USBHS_DEVFNUM & USBHS_DEVFNUM_FNUM_Msk) >> USBHS_DEVFNUM_FNUM_Pos);
}


ARM_DRIVER_USBH Driver_USBH0 = {
  ARM_USBH_GetVersion,
  ARM_USBH_GetCapabilities,
  ARM_USBH_Initialize,
  ARM_USBH_Uninitialize,
  ARM_USBH_PowerControl,
  ARM_USBH_PortVbusOnOff,
  ARM_USBH_PortReset,
  ARM_USBH_PortSuspend,
  ARM_USBH_PortResume,
  ARM_USBH_PortGetState,
  ARM_USBH_PipeCreate,
  ARM_USBH_PipeModify,
  ARM_USBH_PipeDelete,
  ARM_USBH_PipeReset,
  ARM_USBH_PipeTransfer,
  ARM_USBH_PipeTransferGetResult,
  ARM_USBH_PipeTransferAbort,
  ARM_USBH_GetFrameNumber
};
